/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent;

import de.nachtlicht_media.epoint.redcent.exception.ForbiddenExistingException;

public class AuthorizedIssuer {
	
	public static enum Operations {
		updatePublicKey,
		updateNotaryStatus,
		createAuthorizedIssuer,
		updateEmissionLimits
	}
	
	// The authorized issuer's master key fingerprint in hex, no spaces all UPPERCASE
	private String fingerprint;
	
	// The AI's public key block (including signatures of at least <n> notaries)
	private String publicKey;

	// A comma separated list of all non-revoked keys in the AI's public key ring, all HEX, uppercase and padded to 16 Digits
	private String publicKeyIdList;
	
	// The AI's emission boundaries or unlimited if <null>, init to 0 here for security reasons
	private Long emissionLimitLowerBound = 0L;
	private Long emissionLimitUpperBound = 0L;
	
	// The AI's current balance with this principal issuer
	private Long balance;
	
	// Whether this AI has notary rights, and if so, which notaries signed for this AI
	// An AI having notary rights may (toghether with <n> other notaries) introduce new AI's to the guild
	private boolean hasNotaryRights = false;
	// The latest valid serial number at the time of becoming a notary
	private Long notarySinceLSN;
	
	// The last operation performed on this AI, e.g. raise/lower limits, the corresponding authorization and the
	// LSN when this operation was performed
	private String lastAction;
	private String lastActionsAuthorization;
	private Long lastActionsLSN;
	
	public AuthorizedIssuer() {
	}

	public AuthorizedIssuer(String fingerprint, String publicKey, String publicKeyIdList,
			Long emissionLimitLowerBound, Long emissionLimitUpperBound,
			Long balance, boolean hasNotaryRights, Long notarySinceLSN,
			String lastAction, String lastActionsAuthorization,
			Long lastActionsLSN) {
		super();
		this.fingerprint = fingerprint;
		this.publicKey = publicKey;
		this.publicKeyIdList = publicKeyIdList;
		this.emissionLimitLowerBound = emissionLimitLowerBound;
		this.emissionLimitUpperBound = emissionLimitUpperBound;
		this.balance = balance;
		this.hasNotaryRights = hasNotaryRights;
		this.notarySinceLSN = notarySinceLSN;
		this.lastAction = lastAction;
		this.lastActionsAuthorization = lastActionsAuthorization;
		this.lastActionsLSN = lastActionsLSN;
	}

	public boolean mayEmitEpointAmount(long amount) {
		if (getEmissionLimitUpperBound() != null && (getBalance() + Math.abs(amount) > getEmissionLimitUpperBound().longValue()))
			return false;
		
		return true;
	}
	
	public void checkEmitEpointAmount(long amount) throws ForbiddenExistingException {
		// if there is a limit, check it and throw exception in case the limit would be violated
		if (!mayEmitEpointAmount(amount))
			throw new ForbiddenExistingException("Transaction would violate emission limit for " + getFingerprint());
	}
	
	public boolean mayDestroyEpointAmount(long amount) {
		if (getEmissionLimitLowerBound() != null && (getBalance() - Math.abs(amount) < getEmissionLimitLowerBound().longValue()))
			return false;
		return true;
	}
	
	public void checkDestroyEpointAmount(long amount) throws ForbiddenExistingException {
		// if there is a limit, check it and throw exception in case the limit would be violated
		if (!mayDestroyEpointAmount(amount))
			throw new ForbiddenExistingException("Transaction would violate emission limit for " + getFingerprint());
	}
	
	public String getFingerprint() {
		if (fingerprint == null)
			return null;
		return fingerprint.replaceAll(" ", "").trim().toUpperCase();
	}

	public void setFingerprint(String fingerprint) {
		if (fingerprint != null) {
			this.fingerprint = fingerprint.replaceAll(" ", "").trim().toUpperCase();
		} else
			this.fingerprint = null;
	}
	
	public String getPublicKey() {
		return publicKey;
	}

	public Long getEmissionLimitLowerBound() {
		return emissionLimitLowerBound;
	}

	public Long getEmissionLimitUpperBound() {
		return emissionLimitUpperBound;
	}

	public Long getBalance() {
		return balance;
	}

	public boolean hasNotaryRights() {
		return hasNotaryRights;
	}

	public Long getNotarySinceLSN() {
		return notarySinceLSN;
	}

	public String getLastAction() {
		return lastAction;
	}

	public String getLastActionsAuthorization() {
		return lastActionsAuthorization;
	}

	public Long getLastActionsLSN() {
		return lastActionsLSN;
	}

	public void setPublicKey(String publicKey) {
		this.publicKey = publicKey;
	}

	public void setEmissionLimitLowerBound(Long emissionLimitLowerBound) {
		this.emissionLimitLowerBound = emissionLimitLowerBound;
	}

	public void setEmissionLimitUpperBound(Long emissionLimitUpperBound) {
		this.emissionLimitUpperBound = emissionLimitUpperBound;
	}

	public String getPublicKeyIdList() {
		return publicKeyIdList;
	}

	public void setPublicKeyIdList(String publicKeyIdList) {
		this.publicKeyIdList = publicKeyIdList;
	}
	
}
