package de.nachtlicht_media.epoint.redcent;

import java.io.File;

import org.springframework.beans.factory.annotation.Autowired;

import de.nachtlicht_media.epoint.redcent.exception.ConflictException;
import de.nachtlicht_media.epoint.redcent.interfaces.AuthorizedIssuerStore;
import de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageEncryptor;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageHandler;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageSignatureCreator;
import de.nachtlicht_media.epoint.redcent.interfaces.RedCentConfig;
import de.nachtlicht_media.epoint.redcent.interfaces.TemplateManager;

public abstract class EpointService implements AuthorizedIssuerStore
{
	protected final RedCentConfig appConfig;

	static final String na = "N/A";
	public static final String DOCUMENT_REALM_ISSUER_PERSISTENT_DOCUMENT = "DOCUMENT_REALM_ISSUER_PERSISTENT_DOCUMENT";

	@Autowired
	public EpointService(RedCentConfig appConfig)
	{
		this.appConfig = appConfig;
	}

	public abstract boolean testDataSource() throws Exception;
	
	public String getEnvVar(String env)
	{
	    return System.getProperty(env, na);
	}

	/**
	 * Dependency injection style configuration function for TemplateRealPath.
	 * @return The absolute path to the webapplication's template directory on
	 * the server's file system. In this directory certificates for example
	 * for epoint certificates have to be located.
	 */
	public String getTemplateRealPath()
	{
		return appConfig.getTemplateRealPath();
	}

	/**
	 * Dependency injection style configuration function for TrustedPublicKeyRingPath.
	 * @return The absolute path to the webapplication's TrustedPublicKeyRing on
	 * the server's file system. All keys in this key ring file can be used to
	 * perform special actions such as emission and invalidation of ePoints.
	 */
	public String getTrustedPublicKeyRingPath()
	{
		return new File(appConfig.getTrustedPublicKeyRingPath()).getAbsolutePath();
	}

	/**
	 * Dependency injection style configuration function for IssuersPublicKeyRingPath.
	 * @return The absolute path to the webapplication's IssuersPublicKeyRing on
	 * the server's file system. Used to verify signatures. Resolving keys from the
	 * local storage is much faster than download them from a key server.
	 */
	public String getIssuersPublicKeyRingPath()
	{
		return new File(appConfig.getIssuersPublicKeyRingPath()).getAbsolutePath();
	}

	/**
	 * Dependency injection style configuration function for IssuersPrivateKeyRingPath.
	 * @return The absolute path to the webapplication's IssuersPrivateKeyRing on
	 * the server's file system. This file must contain the server's private key used to
	 * sign documents.
	 */
	public String getIssuersPrivateKeyRingPath()
	{
		return new File(appConfig.getIssuersPrivateKeyRingPath()).getAbsolutePath();
	}
	
	/**
	 * @return the password required to extract the issuers private key from his private keyring
	 */
	protected String getIssuersKeyPassword()
	{
		return appConfig.getIssuersPrivateKeyPassword();
	}

	/**
	 * @return the id of the trusted key to be used as the master key
	 */
	protected String getMasterKeyId() {
		return appConfig.getMasterKeyId();
	}
	
	/**
	 * Dependency injection style configuration function for IssuerURL.
	 * @return The webapplication's URL on the
	 * external interface. Used to replace variable $ISSUER_URL in templates.
	 */
	public String getIssuerURL()
	{
		return appConfig.getIssuerURL();
	}
	
	/**
	 * Dependency injection style configuration function for IssuerName.
	 * @return The issuer's name, used to replace variable $ISSUER_NAME in templates.
	 */
	public String getIssuerName()
	{
		return appConfig.getIssuerName();
	}
	
	/**
	 * Dependency injection style configuration function for SSL_KEY.
	 * @return The issuer's ssl key, used to replace variable $SSL_KEY in templates.
	 */
	public String getSSLKey()
	{
		return appConfig.getSSLKey();
	}
	
	public String getIssuersEmailAddress()
	{
		return appConfig.getIssuerEmail();
	}
	
	/**
	 * 
	 * @return the number of digits ePoint certificate's serial numbers are padded to.
	 */
	public int getFixedSerialNumberlength()
	{
		return 16;
	}	

	/**
	 * @return TemplateManager
	 */
	public TemplateManager getTemplateManager() throws ConflictException
	{
		try
		{
			if (templateManager == null)
				initTemplateManager();
		}
		catch(Exception ex)
		{
			throw new ConflictException(ex);
		}
		
		return templateManager;
	}

	@Autowired
	protected TemplateManager templateManager;
	protected void initTemplateManager() throws Exception
	{
		templateManager.init(this);
	}

	@Autowired
	protected DocumentStore documentStore;
	public DocumentStore getDocumentStore() throws ConflictException
	{
		return documentStore;
	}
	
	@Autowired
	private MessageHandler messageHandler = null;
	protected MessageHandler getMessageHandler()
	{
		return messageHandler;
	}

	@Autowired
	protected MessageSignatureCreator messageSignatureCreator;
	protected MessageSignatureCreator getMessageSignatureCreator()
	{
		return messageSignatureCreator;
	}
	
	@Autowired
	protected MessageEncryptor messageEncryptor;
	protected MessageEncryptor getMessageEncryptor()
	{
		return messageEncryptor;
	}

	protected byte[] signMessage(String message) throws ConflictException
	{
		return getMessageSignatureCreator().signMessage(message);
	}

	protected byte[] signBinaryContent(byte[] content, boolean armor) throws ConflictException
	{    
		return getMessageSignatureCreator().signBinaryContent(content, armor);
	}

	protected byte[] signKey(byte[] publicKeyData, boolean armor) throws ConflictException
	{
		return getMessageSignatureCreator().signKey(publicKeyData, armor);
	}
	
	protected byte[] encryptMessage(String title, byte[] message, byte[] publicKeyData, long[] encryptionIds, boolean armor) throws ConflictException
	{
		return getMessageEncryptor().encryptMessage(title, message, publicKeyData, encryptionIds, armor);
	}

	public String getPaddedSerialNumber(Long serial) {
		return getLeftPaddedNumber(serial, '0');
	}
	
	public String getLeftPaddedNumber(Long serial, char pad)
	{
		StringBuffer val = new StringBuffer(String.valueOf(serial));
		
		int padAmount = getFixedSerialNumberlength() - val.length();
		if (padAmount > 0)
		{
			for(int i = padAmount; i > 0; i--)
				val.insert(0, pad);
		}
			
		return val.toString();
	}
	
}