/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.List;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;

import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.openpgp.PGPPublicKeyRingCollection;
import org.bouncycastle.openpgp.PGPSecretKeyRingCollection;
import org.bouncycastle.openpgp.PGPUtil;
import org.springframework.beans.factory.annotation.Autowired;

import de.nachtlicht_media.epoint.redcent.crypto.BouncycastleMessageDecryptionVerificationFactory;
import de.nachtlicht_media.epoint.redcent.exception.BadSerialNumberException;
import de.nachtlicht_media.epoint.redcent.exception.BadsignatureException;
import de.nachtlicht_media.epoint.redcent.exception.ConflictException;
import de.nachtlicht_media.epoint.redcent.exception.DatabaseErrorException;
import de.nachtlicht_media.epoint.redcent.exception.ForbiddenExistingException;
import de.nachtlicht_media.epoint.redcent.exception.InvalidPGPMessageException;
import de.nachtlicht_media.epoint.redcent.exception.MalformedRequestException;
import de.nachtlicht_media.epoint.redcent.exception.NotFoundException;
import de.nachtlicht_media.epoint.redcent.exception.RedcentProcessingException;
import de.nachtlicht_media.epoint.redcent.interfaces.AuthorizedIssuerStore;
import de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore;
import de.nachtlicht_media.epoint.redcent.interfaces.EpointCertificateTypes;
import de.nachtlicht_media.epoint.redcent.interfaces.EpointRecordConverter;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageDecryptionVerificationHandler;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageEncryptor;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageHandler;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageSignatureCreator;
import de.nachtlicht_media.epoint.redcent.interfaces.RedCentConfig;
import de.nachtlicht_media.epoint.redcent.interfaces.TemplateManager;
import de.nachtlicht_media.epoint.redcent.messaging.DefaultTemplateManager;
import de.nachtlicht_media.epoint.redcent.messaging.EpointRecordToXLogConverter;
import de.nachtlicht_media.epoint.redcent.messaging.StringDataMap;
import de.nachtlicht_media.epoint.redcent.util.Splitter;
import de.nachtlicht_media.epoint.redcent.util.StringUtils;

/**
 * RedCent is the common super-class for all implementations of the issuer application. It is instanciated once,
 * for example by a servlet-application and handles all ePoint transactions and signing processes.
 */
public abstract class RedCent extends EpointService
{
	public static final boolean ENABLE_OBLIGATION_AND_DRAFT_PROCESSING = true;
	
	public static final int RETURN_RESULTING_CERT = 0;
	public static final int RETURN_TRANSACTION_RECEIPT = 1;
	public static final int RETURN_TRANSACTIONVALUE_STRING_AS_BYTEARRAY_UTF8 = 2;
	
	private static final Log log = LogFactory.getLog(RedCent.class);
	protected Log xLog;
	
	public static final String DOCUMENT_REALM_EXCHANGE_RATE_NOTIFICATIONS = "DOCUMENT_REALM_EXCHANGE_RATE_NOTIFICATIONS";

	@Autowired
	public RedCent(RedCentConfig config)
	{
		super(config);
	}
	
	@PostConstruct
	public void init() throws Exception
	{
		testDataSource();
		
		String xLogName = StringUtils.replace(StringUtils.replace(getIssuerURL(), "/" , "_"), ":", "_") + ".TransactionLog";
		xLog = LogFactory.getLog(xLogName);

		BouncycastleMessageDecryptionVerificationFactory.init(
				this,
				getMasterKeyId(),
				new PGPPublicKeyRingCollection(PGPUtil.getDecoderStream(new FileInputStream(getIssuersPublicKeyRingPath()))), 
				new PGPSecretKeyRingCollection(PGPUtil.getDecoderStream(new FileInputStream(getIssuersPrivateKeyRingPath()))), 
				getIssuersKeyPassword(), new PGPPublicKeyRingCollection(PGPUtil.getDecoderStream(new FileInputStream(getTrustedPublicKeyRingPath()))));
		
		log.info("Using default emission limits " + appConfig.getDefaultEmissionLimitLowerBound() + " and " + appConfig.getDefaultEmissionLimitUpperBound());
		log.info(appConfig.getNumberOfNotarySignaturesRequiredForDefaultAIRegistration() + " notary signatures are required to add a new AI with default settings to this issuer.");
		
		initTemplateManager();
		
		initServiceContract();
		
		log.info("Initialized issuer at " + getIssuerURL());
		log.info("Initialized transaction logger '"+xLogName+"'");
		xLog.info("INIT " + getIssuerURL());
	}

	@PreDestroy
	public void destroy() throws Exception
	{
		
	}
	
	protected void initServiceContract() throws ConflictException
	{
		// verify that the service contract located at /$WEBROOT/service is up to date
		// if not update it from the current template
		try
		{
			String currentServiceContractTemplate = getTemplateManager().getTemplate(getTemplateManager().getHashValueForCurrentVersion(EpointCertificateTypes.SERVER_SERVICE_CONTRACT));
		
			byte[] deployedServiceContractData = getDocumentStore().getDocument(DOCUMENT_REALM_ISSUER_PERSISTENT_DOCUMENT, EpointCertificateTypes.SERVER_SERVICE_CONTRACT);
		
			MessageDecryptionVerificationHandler messageSignatureChecker = null;

			boolean replaceServiceContract = false;

			if (deployedServiceContractData != null)
			{
				String deployedServiceContract = getTemplateManager().readTemplate(new ByteArrayInputStream(deployedServiceContractData));
				messageSignatureChecker =  BouncycastleMessageDecryptionVerificationFactory.getMessageDecryptionVerificationHandler();
				messageSignatureChecker.setMessage(deployedServiceContract.getBytes("UTF-8"));

				try{
					messageSignatureChecker.analyse();
				}catch (InvalidPGPMessageException e) {
					e.printStackTrace();
					replaceServiceContract = true;
				}
			}

			if (	messageSignatureChecker != null 
					&& 	messageSignatureChecker.isValid() != null && messageSignatureChecker.isValid().booleanValue() 
					&& 	messageSignatureChecker.isClearText() != null && messageSignatureChecker.isClearText().booleanValue())
			{
				// Message is properly signed and in cleartext, check contents
				log.info("Found a service contract at webroot, comparing with a current version from the template archive...");
				
				String deployedServiceContractClearText = messageSignatureChecker.getClearText();
//				deployedServiceContractClearText = deployedServiceContractClearText.substring(deployedServiceContractClearText.indexOf("\n\n") + 2);
				
				if (getTemplateManager().calcDigest(createServerServiceContractContents()).equals(getTemplateManager().calcDigest(deployedServiceContractClearText)))
				{
					log.info("... service contract is up to date and will not be touched.");
					replaceServiceContract = false;
				}
				else
				{
					log.info("... service contract is not up to date and will be replaced.");
					log.debug("currentServiceContractTemplate is '"+currentServiceContractTemplate+"'");
					log.debug("deployedServiceContractTemplate is '"+deployedServiceContractClearText+"'");
					replaceServiceContract = true;
				}
			}
			else
				replaceServiceContract = true;
			
			if (replaceServiceContract)
			{
				// The deployed service contract is not signed or not cleartext or nonexisting...
				// get a signed current version and write it to the documentStore
				log.info("Updating service contract with a current version from the template archive...");
				getDocumentStore().storeOrUpdateDocument(DOCUMENT_REALM_ISSUER_PERSISTENT_DOCUMENT, EpointCertificateTypes.SERVER_SERVICE_CONTRACT, createSignedServerServiceContract());
			}
		}
		catch(Exception ex)
		{
			throw new ConflictException(ex);
		}
	}

	private byte[] createSignedServerServiceContract() throws DatabaseErrorException, ConflictException
	{
		return signMessage(createServerServiceContractContents());
	}

	private String createServerServiceContractContents() throws ConflictException
	{
		StringDataMap data = new StringDataMap();
		
		data.put("ISSUER_URL", getIssuerURL());
		data.put("ISSUER_NAME", getIssuerName());
		data.put("ISSUER_EMAIL", getIssuersEmailAddress());
		data.put("ISSUER_FINGERPRINT", getIssuerFingerPrintText());
		data.put("SSL_KEY", getSSLKey());

		return getMessageHandler().createMessage(EpointCertificateTypes.SERVER_SERVICE_CONTRACT, data);
	}

	private EpointRecordConverter xLogConverter = null;
	protected EpointRecordConverter getXLogConverter()
	{
		if (xLogConverter == null)
			xLogConverter = new EpointRecordToXLogConverter();
		
		return xLogConverter;
	}
	
	public String getIssuerFingerPrintText() throws ConflictException
	{
		return getMessageSignatureCreator().getIssuerFingerPrintText();
	}

	public String getIssuerFingerPrint() throws ConflictException
	{
		return getIssuerFingerPrintText().replaceAll(" ", "").toUpperCase();
	}

	protected String createCertMessage(long amount, EpointRecord rec, String issuerNotes, String userNotes) throws ConflictException
	{
		StringDataMap data = new StringDataMap();
		
		data.put("VALUE", String.valueOf(amount));
		data.put("ID", Splitter.group(rec.getId(),4,5));
		data.put("SN", getPaddedSerialNumber(new Long(rec.getSerialNumber())));
		
		data.put("EMAIL", getIssuersEmailAddress());
		data.put("ISSUER_NOTES",(issuerNotes != null ? issuerNotes + "\n": "N/A\n"));
		data.put("USER_NOTES",(userNotes != null ? userNotes + "\n": "N/A\n"));
	
		data.put("ISSUER", getIssuerFingerPrintText());
		data.put("URL", getIssuerURL());
		data.put("MD_ALGO", "SHA1");
		data.put("LICENSORS", "N/A");
		data.put("DATE", rec.getTxDate());
		
		return getMessageHandler().createMessage(EpointCertificateTypes.EPOINT_CERTIFICATE, data);
	}

	protected String createInvalidationMessage(EpointRecord rec, String idToInvalidate, String issuerNotes, String userNotes) throws ConflictException
	{
		StringDataMap data = new StringDataMap();
		
		data.put("VALUE", String.valueOf(rec.getValue()));
		data.put("INVALID_ID", Splitter.group(idToInvalidate,4,5));
		data.put("ID", Splitter.group(rec.getId(),4,5));
		data.put("SN", getPaddedSerialNumber(new Long(rec.getSerialNumber())));
		
		data.put("EMAIL", getIssuersEmailAddress());
		data.put("ISSUER_NOTES",(issuerNotes != null ? issuerNotes + "\n": "N/A\n"));
		data.put("USER_NOTES",(userNotes != null ? userNotes + "\n": "N/A\n"));
	
		data.put("ISSUER", getIssuerFingerPrintText());
		data.put("URL", getIssuerURL());
		data.put("MD_ALGO", "SHA1");
		data.put("LICENSORS", "N/A");
		data.put("DATE", rec.getTxDate());
		
		return getMessageHandler().createMessage(EpointCertificateTypes.INVALIDATION_CERTIFICATE, data);
	}

	protected String createObligationCertMessage(long amount, EpointRecord obligationRec, long difference, String issuerNotes, String userNotes) throws ConflictException
	{
		StringDataMap data = new StringDataMap();

		data.put("VALUE", String.valueOf(amount));
		data.put("ID", Splitter.group(obligationRec.getId(),4,5));
		data.put("SN", getPaddedSerialNumber(new Long(obligationRec.getSerialNumber())));

		data.put("EMAIL", getIssuersEmailAddress());
		data.put("ISSUER_NOTES", (issuerNotes != null ? issuerNotes + "\n": "N/A\n"));
		data.put("USER_NOTES", (userNotes != null ? userNotes + "\n": "N/A\n"));
		data.put("DIFFERENCE", String.valueOf(difference));

		data.put("ISSUER", getIssuerFingerPrintText());
		data.put("URL", getIssuerURL());
		data.put("MD_ALGO", "SHA1");
		data.put("LICENSORS", "N/A");
		data.put("DATE", obligationRec.getTxDate());
		data.put("SEC", new Integer(0));

		return getMessageHandler().createMessage(EpointCertificateTypes.OBLIGATION_CERTIFICATE, data);
	}

	protected byte[] createTransactionReceipt(long value, String targetDigestOrKey) throws ConflictException
	{
		if (value >= Integer.MAX_VALUE)
			throw new ConflictException("Cannot create transactionReceipt for values larger than MAX_INT");
		
		try
		{
			byte[] digestOrKey = Hex.decodeHex(targetDigestOrKey.toCharArray());

			ByteArrayOutputStream out = new ByteArrayOutputStream();

			// write the integer out MSB first
			DataOutputStream valOut = new DataOutputStream(out);
			valOut.writeInt((int)value);
			valOut.flush();
			valOut.close();

			out.write(digestOrKey);

			out.flush();
			out.close();

			return out.toByteArray();
		}
		catch(IOException ioEx)
		{
			throw new ConflictException("IO Error while creating transaction receipt", ioEx);
		}
		catch (DecoderException decEx)
		{
			throw new ConflictException("Error while decoding Hex representation of Digest or Key", decEx);
		}
	}
	
	/**
	 * Returns whether the transaction that resulted in issuing the given
	 * record is vulnerable to replay attacks. Emission and
	 * draft endorsements (split obligations) are the typical cases.
	 * @param record
	 * @return
	 */
	public boolean isTransActionVulnerableToReplayAttacks(EpointRecord record)
	{
		return (record.getIssuingCause() == EpointRecord.EMIT || record.getIssuingCause() == EpointRecord.OBLIGATIONS_SPLIT);
	}

	/* All abstract methods beneath have to be implemented by specific implementations, for example an SQL-based solution*/
	public abstract byte[] getServerStatus() throws RedcentProcessingException;
	public abstract byte[] getServerBalance(String key) throws RedcentProcessingException;
	public abstract byte[] getSignedLedger(String id, Long start, Long end) throws Exception;

	public abstract byte[] getExchangeRateInformation() throws RedcentProcessingException;
	
	public byte[] getDeployedServerServiceContract() throws Exception
	{
		return getDocumentStore().getDocument(DOCUMENT_REALM_ISSUER_PERSISTENT_DOCUMENT, EpointCertificateTypes.SERVER_SERVICE_CONTRACT);
	}
	
	/**
	 * Returns the published ExchangeRateNotification corresponding to the given xrnSerial from the
	 * document storage.
	 * 
	 * @param xrnSerial
	 * @return byte-encoded encrypted, clearsigned ExchangeRateNotification
	 * @throws Exception
	 */
	public byte[] getEncryptedExchangeRateNotification(long xrnSerial) throws Exception
	{
		// get document from store
		
		byte[] clearSignedXRN = getDocumentStore().getDocument(DOCUMENT_REALM_EXCHANGE_RATE_NOTIFICATIONS, String.valueOf(xrnSerial));
		
		if (clearSignedXRN == null)
			throw new NotFoundException("An exchange rate notification document with serial number " + xrnSerial + " could not be found!");

		// encrypt it
		return encryptMessage("XRN-"+xrnSerial, clearSignedXRN, BouncycastleMessageDecryptionVerificationFactory.getEncodedTrustedPublicKeys(), BouncycastleMessageDecryptionVerificationFactory.getValidTrustedIdsForEncryption(), true); //TODO: remove direct bouncycastle dependency
	}
	
	public String publishExchangeRateNotification(byte[] xrn) throws Exception
	{
		// Verify xrn again
		MessageDecryptionVerificationHandler messageSignatureChecker =  BouncycastleMessageDecryptionVerificationFactory.getMessageDecryptionVerificationHandler();
		messageSignatureChecker.setMessage(xrn);
		messageSignatureChecker.analyse();

		if (	messageSignatureChecker.isValid() == null 
				|| !messageSignatureChecker.isValid().booleanValue() 
				|| messageSignatureChecker.isClearText() == null 
				|| !messageSignatureChecker.isClearText().booleanValue())
			throw new BadsignatureException("The submitted exchange rate notification is not a valid clear-signed document.");

		if (!Boolean.TRUE.equals(messageSignatureChecker.signerIsIssuersMasterKey()))
			throw new BadsignatureException("The submitted exchange rate notification is not signed by the issuer's master key.");

		// Parse incoming message and enforce it to be created from the most recent XRN template.
		StringDataMap data = getMessageHandler().parseMessage(messageSignatureChecker.getClearText(), getTemplateManager().getHashValueForCurrentVersion(EpointCertificateTypes.EXCHANGE_RATE_NOTIFICATION));

		log.debug("Retrieved data: " + data);
		
		// Verify contents:
		Object sn = data.restoreNumber("XRNSN");
		if (sn != null && sn instanceof Number)
		{
			long xrnSerial = ((Number)sn).longValue();
			long mostRecent = StringUtils.parsePrimitiveLongFromString(getDocumentStore().getMostRecentDocumentId(DOCUMENT_REALM_EXCHANGE_RATE_NOTIFICATIONS));

			if (xrnSerial != (mostRecent + 1))
				throw new BadSerialNumberException("Serial numbers for exchange rate notifications must be a sequence. The most recent published serial number is " + mostRecent);

			byte[] existing = getDocumentStore().getDocument(DOCUMENT_REALM_EXCHANGE_RATE_NOTIFICATIONS, String.valueOf(xrnSerial));

			if (existing != null)
				throw new ForbiddenExistingException("An exchange rate notification with the serial number provided already exists.");
			else
			{
				// Verify that the XRN is valid for the current issuer.
				Object issuer = data.get("ISSUER");
				if (issuer == null || !StringUtils.replace(String.valueOf(issuer), " ", "").equals(StringUtils.replace(getIssuerFingerPrintText(), " ", "")))
					throw new MalformedRequestException("The submitted exchange rate notification document's issuer id ('"+issuer+"') does not match the one of the addressed instance ('"+getIssuerFingerPrintText()+"').");

				//TODO: more verification?

				getDocumentStore().storeOrUpdateDocument(DOCUMENT_REALM_EXCHANGE_RATE_NOTIFICATIONS, String.valueOf(xrnSerial), xrn);
				return String.valueOf(xrnSerial);
			}
		}
		else
			throw new BadSerialNumberException("The serial number could not be extracted from the submitted exchange rate notification.");
	}
	
	public abstract byte[] exchange(int resultType, String rand, String targetID, String userNotes) throws RedcentProcessingException;
	public abstract byte[] emit(String targetID, long amount, String signature, String authorizedBy, String targetIssuer, String userNotes) throws RedcentProcessingException;
	public abstract byte[] invalidate(String rand, String signature, String authorizedBy, String targetIssuer, String userNotes) throws RedcentProcessingException;
	public abstract byte[] split(String rand, String targetID, long amount, String restID, String userNotes) throws RedcentProcessingException;
	public abstract byte[] mergeObligation(int resultType, String rand, String key, String userNotes) throws RedcentProcessingException;
	public abstract byte[] processDraftMessage(MessageDecryptionVerificationHandler signedMessage) throws RedcentProcessingException;
	public abstract byte[] splitObligation(String targetID, long amount, String key, byte[] signedDraft, String userNotes) throws RedcentProcessingException;
	public abstract EpointRecord findValidRecordByID(String id) throws RedcentProcessingException;
	public abstract EpointRecord findBySerial(long serial) throws RedcentProcessingException;
	public abstract AuthorizedIssuer findAuthorizedIssuerByFingerprint(String authorizedIssuer) throws RedcentProcessingException;
	public abstract AuthorizedIssuer findAuthorizedIssuerByKeyID(long keyID) throws RedcentProcessingException;
	public abstract AuthorizedIssuer updateAuthorizedIssuer(String authorizedIssuerFingerPrint, String authorizedIssuerPublicKeyBlock, String emissionLimitLowerBoundary, String emissionLimitHigherBoundary, MessageDecryptionVerificationHandler signedMessage, String targetIssuer) throws RedcentProcessingException;

	
	protected String checkTargetDigest(String digestName, String targetID) throws MalformedRequestException
	{
		if (targetID == null || targetID.trim().length() == 0)
			throw new MalformedRequestException(digestName + " digest must not be empty.");
		else
		{
			try
			{
				// Digests are always uppercase whithout spaces
				targetID = targetID.trim().toUpperCase().replaceAll(" ","");

				if (targetID.length() != 40)
					throw new MalformedRequestException(digestName + " digest must be HEX-decodable and have a length of 40 chars.");			

				Hex.decodeHex(targetID.toLowerCase().toCharArray());

				return targetID;
			}
			catch(DecoderException ex)
			{
				throw new MalformedRequestException(digestName + " digest must be HEX-decodable.");
			}
		}
	}

	protected String checkTargetIssuer(String targetIssuer) throws RedcentProcessingException
	{
		if(targetIssuer == null)
			throw new MalformedRequestException("No parameter targetIssuer present containing the issuer's fingerprint.");
		else
		{
			targetIssuer = targetIssuer.toUpperCase().replaceAll(" ","");
			String thisIssuer = getIssuerFingerPrintText().toUpperCase().replaceAll(" ", "");

			if (!targetIssuer.equals(thisIssuer))
				throw new MalformedRequestException("The submitted issuers fingerprint("+targetIssuer+") does not match the fingerprint of this issuer ("+thisIssuer+").");
		}

		return targetIssuer;
	}
	
	protected String checkKeyFingerprint(String fingerprint) throws MalformedRequestException
	{
		if (fingerprint == null || fingerprint.trim().length() == 0)
			throw new MalformedRequestException("Key fingerprint must not be empty.");
		else
		{
			try
			{
				// Digests are always uppercase whithout spaces
				fingerprint = fingerprint.trim().toUpperCase().replaceAll(" ","");
				
				if (fingerprint.length() != 40)
					throw new MalformedRequestException("Key fingerprint must be HEX-decodable and have a length of 40 chars.");			

				Hex.decodeHex(fingerprint.toLowerCase().toCharArray());
				
				return fingerprint;
			}
			catch(DecoderException ex)
			{
				throw new MalformedRequestException("Key fingerprint must be HEX-decodable.");
			}
		}
	}

}
