/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.crypto;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;

import javax.annotation.PostConstruct;

import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.openpgp.PGPPublicKey;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPPublicKeyRingCollection;
import org.bouncycastle.openpgp.PGPSecretKeyRingCollection;
import org.springframework.beans.factory.annotation.Autowired;

import de.nachtlicht_media.epoint.redcent.AuthorizedIssuer;
import de.nachtlicht_media.epoint.redcent.RedCent;
import de.nachtlicht_media.epoint.redcent.interfaces.AuthorizedIssuerStore;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageDecryptionVerificationHandler;

public class BouncycastleMessageDecryptionVerificationFactory
{
	private static final Log log = LogFactory.getLog(BouncycastleMessageDecryptionVerificationFactory.class);
	
	/** 
	 * Don't allow instances of this class. Only calls to static methods are allowed.
	 *
	 */
	private BouncycastleMessageDecryptionVerificationFactory()
	{
	}

	private static PGPPublicKeyRing masterKey;
	private static PGPPublicKeyRingCollection issuersPublicKey;
	private static PGPSecretKeyRingCollection issuersSecretKey;
	private static String issuersPrivateKeyPassword;
	private static PGPPublicKeyRingCollection trustedPublicKeys;
	
	private static AuthorizedIssuerStore authStore;

	public static void init(AuthorizedIssuerStore authStore, String _masterKeyId, final PGPPublicKeyRingCollection _issuersPublicKey, final PGPSecretKeyRingCollection _issuersSecretKey, final String _issuersPrivateKeyPassword, final PGPPublicKeyRingCollection _trustedPublicKeys) throws Exception
	{
		log.debug("Init BouncycastleMessageDecryptionVerificationFactory");
		
		issuersPublicKey = _issuersPublicKey;
		issuersSecretKey = _issuersSecretKey;
		issuersPrivateKeyPassword = _issuersPrivateKeyPassword;
		trustedPublicKeys = _trustedPublicKeys;
		
		if (_masterKeyId == null)
			throw new IllegalArgumentException("Could not initialize BouncycastleMessageDecryptionVerificationFactory without a defined masterKey");
		
		_masterKeyId.toLowerCase();
		
		if (_masterKeyId.startsWith("0x"))
			_masterKeyId = _masterKeyId.substring(2);
		
		masterKey = trustedPublicKeys.getPublicKeyRing(BouncycastleUtil.convertKeyHexIDToLong(_masterKeyId));
		if (masterKey == null)
			throw new IllegalArgumentException("Could not initialize BouncycastleMessageDecryptionVerificationFactory: no trusted public key with id " + _masterKeyId + " could be found as a master key!");
		
		BouncycastleMessageDecryptionVerificationFactory.authStore = authStore;
		if(authStore == null)
			throw new IllegalArgumentException("Could not initialize BouncycastleMessageDecryptionVerificationFactory: no authorizedIssuersStore configured!");
	}

	public static synchronized MessageDecryptionVerificationHandler getMessageDecryptionVerificationHandler() throws Exception
	{
		return new BouncycastleMessageDecryptionVerificationHandler(authStore, issuersPublicKey, issuersSecretKey, issuersPrivateKeyPassword, masterKey, trustedPublicKeys);
	}
	
	public static byte[] getEncodedTrustedPublicKeys() throws IOException
	{
		return trustedPublicKeys.getEncoded();
	}

	/**
	 * @return
	 */
	private static long[] trustedIds;
	public synchronized static long[] getValidTrustedIdsForEncryption()
	{
		if (trustedIds == null)
		{
			ArrayList list = new ArrayList();

			Iterator iter = trustedPublicKeys.getKeyRings();
			while(iter.hasNext())
			{
				PGPPublicKeyRing ring = ((PGPPublicKeyRing)iter.next());

				PGPPublicKey mostRecent = null;

				Iterator subIter = ring.getPublicKeys();
				while (subIter.hasNext())
				{
					PGPPublicKey key = (PGPPublicKey) subIter.next();
					
			        log.trace("Checking key: " + key + (key != null ? " id: 0x" + String.valueOf(Hex.encodeHex(key.getFingerprint())) + " isMaster: " + key.isMasterKey() + " isEncryptionKey: " + key.isEncryptionKey() + " isSuitableForEncryption: " + BouncycastleUtil.keyIsSuitableForEncryption(key) + " isRevoked: " + key.isRevoked(): ""));

					if (BouncycastleUtil.keyIsSuitableForEncryption(key) && !key.isRevoked() && !BouncycastleUtil.keyHasExpired(key))
					{
						if (mostRecent == null)
							mostRecent = key;
						else
						{
							if (key.getCreationTime().getTime() > mostRecent.getCreationTime().getTime())
								mostRecent = key;
						}
					}
				}

				if (mostRecent != null)
					list.add(new Long(mostRecent.getKeyID()));
			}
			
			trustedIds = new long[list.size()];
			for (int i = 0; i < trustedIds.length; i++)
				trustedIds[i] = ((Long)list.get(i)).longValue();
		}
		
		return trustedIds;
	}
}
