/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.crypto;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.Iterator;

import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.bcpg.ArmoredInputStream;
import org.bouncycastle.bcpg.ArmoredOutputStream;
import org.bouncycastle.openpgp.PGPCompressedData;
import org.bouncycastle.openpgp.PGPEncryptedDataList;
import org.bouncycastle.openpgp.PGPException;
import org.bouncycastle.openpgp.PGPLiteralData;
import org.bouncycastle.openpgp.PGPObjectFactory;
import org.bouncycastle.openpgp.PGPOnePassSignature;
import org.bouncycastle.openpgp.PGPOnePassSignatureList;
import org.bouncycastle.openpgp.PGPPrivateKey;
import org.bouncycastle.openpgp.PGPPublicKey;
import org.bouncycastle.openpgp.PGPPublicKeyEncryptedData;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPPublicKeyRingCollection;
import org.bouncycastle.openpgp.PGPSecretKey;
import org.bouncycastle.openpgp.PGPSecretKeyRingCollection;
import org.bouncycastle.openpgp.PGPSignature;
import org.bouncycastle.openpgp.PGPSignatureList;
import org.bouncycastle.openpgp.PGPUtil;

import de.nachtlicht_media.epoint.redcent.AuthorizedIssuer;
import de.nachtlicht_media.epoint.redcent.exception.InvalidPGPMessageException;
import de.nachtlicht_media.epoint.redcent.exception.NotFoundException;
import de.nachtlicht_media.epoint.redcent.exception.RedcentProcessingException;
import de.nachtlicht_media.epoint.redcent.interfaces.AuthorizedIssuerStore;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageDecryptionVerificationHandler;

public class BouncycastleMessageDecryptionVerificationHandler implements MessageDecryptionVerificationHandler
{
	AuthorizedIssuerStore authorizedIssuerStore;
	
	public static final String THE_KEY_THAT_WAS_USED_TO_SIGN_THE_MESSAGE_IS_NO_MORE_VALID = "The key that was used to sign the message is no more valid.";
	public static final String THE_KEY_THAT_WAS_USED_TO_SIGN_THE_MESSAGE_IS_REVOKED = "The key that was used to sign the message is revoked.";

	private static final Log log = LogFactory.getLog(BouncycastleMessageDecryptionVerificationHandler.class);

	private byte[] message;
	
	private String clearText = null;
	private String encryptedTo = null;
	private String signersFingerprint = null;
	private String signersPrimaryKeyFingerprint = null;
	private Long keyId = null;
	
	private Boolean signerIsNotary = null;
	private Boolean signerIsAuthorizedIssuer = null;
	private Boolean signerIsIssuersMasterKey= null;
	
	private Boolean isValid = null;
	private Boolean isEncrypted = null;
	private Boolean isClearText = null;
	
	private PGPPublicKeyRingCollection pubRing;
	private PGPSecretKeyRingCollection secRing;
	private String issuersSecretKeyPassword;
	private PGPPublicKeyRingCollection embeddedNotaryPubRing;
	private PGPPublicKeyRing issuerMasterPubRing;
	
	private PGPPublicKeyRingCollection tempPubRing;

	public BouncycastleMessageDecryptionVerificationHandler(AuthorizedIssuerStore authorizedIssuerStore, PGPPublicKeyRingCollection pubRing, PGPSecretKeyRingCollection secRing, String issuersKeyPassword, PGPPublicKeyRing issuerMasterPubRing, PGPPublicKeyRingCollection trustedPubRing) throws Exception
	{
		this.authorizedIssuerStore = authorizedIssuerStore;
		this.pubRing = pubRing;
		this.secRing = secRing;
		this.embeddedNotaryPubRing = trustedPubRing;
		this.issuersSecretKeyPassword = issuersKeyPassword;
		this.issuerMasterPubRing = issuerMasterPubRing;
	}
	
	public synchronized void setMessage(byte[] message)
	{
		if (this.message == null)
			this.message = message;
		else
			throw new IllegalArgumentException("Multiple calls to setMessage() are not permitted.");
	}
	
	public byte[] getMessage()
	{
		return message;
	}

	public void analyse() throws InvalidPGPMessageException
	{
		try
		{
			byte[] decodedMessage = getMessage();//URLDecoder.decode(getMessage(), "UTF-8");
			log.debug("Decoded Message from UTF-8:\n" + new String(decodedMessage, "UTF-8"));
			
			// switch message type: clearSigned or encrypted/signed
			ArmoredInputStream aIn = new ArmoredInputStream(new ByteArrayInputStream(decodedMessage));
			
			if (aIn.isClearText())
			{
				log.debug("Processing clearsigned message...");
				// BC does not strip trailing spaces
				stripTrailingSpaces();
				verifyClearSignedMessage(new ByteArrayInputStream(decodedMessage));
				isClearText = Boolean.TRUE;
			}
			else
			{
				log.debug("Processing signed/encrypted message...");
				decrypt(new ByteArrayInputStream(decodedMessage));
				isClearText = Boolean.FALSE;
			}
		}
		catch(Exception ex)
		{
			throw new InvalidPGPMessageException(ex);
		}
	}

//	private void debugDumpKeyRings(PGPSecretKeyRingCollection secRing, PGPPublicKeyRingCollection pubRing)
//	{
//		Iterator outIter = pubRing.getKeyRings();
//		while(outIter.hasNext())
//		{
//			PGPPublicKeyRing ring = (PGPPublicKeyRing)outIter.next();
//			Iterator iter = ring.getPublicKeys();
//			while(iter.hasNext())
//			{
//				PGPPublicKey publicKey = ((PGPPublicKey)iter.next());
//				log.trace("Got public key: " + publicKey.getKeyID() + " algo: " + publicKey.getAlgorithm() + " FingerPrint: " + new String(Hex.encodeHex(publicKey.getFingerprint()))
//				+ " daysvalid " + publicKey.getValidDays());
//			}
//		}
//		
//		outIter = secRing.getKeyRings();
//		while(outIter.hasNext())
//		{
//			PGPSecretKeyRing ring = (PGPSecretKeyRing)outIter.next();
//			
//			Iterator iter = ring.getSecretKeys();
//			while(iter.hasNext())
//			{
//				PGPSecretKey secretKey = ((PGPSecretKey)iter.next());
//				log.trace("Got secret key: " + secretKey.getKeyID() + secretKey.isMasterKey() + secretKey.isSigningKey());
//			}
//		}
//	}

	public String getClearText()
	{
		return clearText;
	}

	public String getEncryptedTo()
	{
		return encryptedTo;
	}

	public String getSignersFingerprint()
	{
		return signersFingerprint;
	}

	public String getSignersPrimaryKeyFingerprint()
	{
		return signersPrimaryKeyFingerprint;
	}
	
	public Boolean signerIsAuthorizedIssuer()
	{
		return signerIsAuthorizedIssuer;
	}
	
	public Boolean signerIsNotary()
	{
		return signerIsNotary;
	}
	
	public Boolean signerIsIssuersMasterKey()
	{
		return signerIsIssuersMasterKey;
	}
	
	public Boolean isClearText()
	{
		return isClearText;
	}

	public Boolean isEncrypted()
	{
		return isEncrypted;
	}

	public Boolean isValid()
	{
		return isValid;
	}

	private void decrypt(InputStream in) throws Exception
	{
		// messages are armored...
		ArmoredInputStream aIn = new ArmoredInputStream(in);

		PGPObjectFactory fac = new PGPObjectFactory(aIn);

		Object o = null;

		if((o = fac.nextObject()) != null)
		{
			log.trace(o);

			if (o instanceof PGPEncryptedDataList)
			{
				PGPEncryptedDataList dl = (PGPEncryptedDataList)o;
				Iterator iter = dl.getEncryptedDataObjects();
				while(iter.hasNext())
				{
					Object dataItem = iter.next();

					log.debug("Read next dataItem from PGPEncryptedDataList: " + dataItem);
					
					if (dataItem instanceof PGPPublicKeyEncryptedData)
					{
						PGPPublicKeyEncryptedData pubEncData = (PGPPublicKeyEncryptedData)dataItem;
						log.debug("Data was encrypted using public key:  " + pubEncData.getKeyID());

						isEncrypted = Boolean.TRUE;

						PGPSecretKey secKey = getSecRing().getSecretKey(pubEncData.getKeyID());
						if (secKey == null)
						{
							log.debug("Could not locate private key with id " + pubEncData.getKeyID());
							continue;
						}
						else
							log.debug("Secret key is: " + secKey);

						PGPPublicKey pubKey = getPubRing().getPublicKey(pubEncData.getKeyID());
						if (pubKey == null)
						{
							log.debug("Could not locate public key with id " + pubEncData.getKeyID());
							continue;
						}

						log.debug("Public key data was encrypted to found: " + pubKey + " using algorithm " + pubKey.getAlgorithm());

						encryptedTo = String.valueOf(pubKey.getFingerprint());

						PGPPrivateKey privKey = secKey.extractPrivateKey(issuersSecretKeyPassword.toCharArray(), "BC");
						log.debug("Private key is: " + privKey);

						log.debug("Reading data... ");
						InputStream dataStream = pubEncData.getDataStream(privKey, "BC");

						PGPObjectFactory plainFac = new PGPObjectFactory(dataStream);

						Object decO = null;
						if((decO = plainFac.nextObject()) != null)
						{
							log.trace("Found decrypted Object: " + decO);
							if (decO instanceof PGPLiteralData)
							{
								PGPLiteralData lit = (PGPLiteralData)decO;
								clearText = extractClearTextFromPGPLiteral(lit);
								log.debug("Extracted cleartext:" + clearText);
								isValid = applyValidityChecks(Boolean.TRUE, pubKey);
								
								return;
							}
							else if(decO instanceof PGPCompressedData)
							{
								PGPCompressedData data = (PGPCompressedData)decO;
								PGPObjectFactory dataFac = new PGPObjectFactory(data.getDataStream());

								Object decompO = null;
								while((decompO = dataFac.nextObject()) != null)
								{
									log.trace("Decrypted decompressed data is: " + decompO);
									if (decompO instanceof PGPLiteralData)
									{
										PGPLiteralData lit = (PGPLiteralData)decompO;
										clearText = extractClearTextFromPGPLiteral(lit);
										log.debug("Extracted cleartext:" + clearText);
										isValid = applyValidityChecks(Boolean.TRUE, pubKey);
										
										return;
									}
									else if (decompO instanceof PGPOnePassSignatureList)
									{
										PGPOnePassSignatureList sigLst = (PGPOnePassSignatureList)decompO;

										for (int i = 0; i < sigLst.size(); i++)
										{
											PGPOnePassSignature sig = sigLst.get(i);
											log.debug("Found signature created with key " + sig.getKeyID() + " type " + sig.getSignatureType());

											PGPPublicKey sigVKey = lookupSigningKey(sig.getKeyID());

											if (sigVKey == null) {
												throw new NotFoundException("Could not lookup signing key for id " + BouncycastleUtil.convertKeyIDToHex(sig.getKeyID()));
											}
											
											keyId = new Long(sigVKey.getKeyID());

											PGPLiteralData p2 = (PGPLiteralData)dataFac.nextObject();
											InputStream                 dIn = p2.getInputStream();
											ByteArrayOutputStream out = new ByteArrayOutputStream();
											int                         ch;

											sig.initVerify(sigVKey, "BC");

											while ((ch = dIn.read()) >= 0)
											{
												sig.update((byte)ch);
												out.write(ch);
											}
											out.close();
											PGPSignatureList            p3 = (PGPSignatureList)dataFac.nextObject();

											if (sig.verify(p3.get(0)))
											{
												log.debug("signature verified: " + out.toString("UTF-8"));
												clearText = out.toString("UTF-8");
												isValid = Boolean.TRUE;
											}
											else
											{
												log.debug("signature verification failed.");
												isValid = Boolean.FALSE;
									            signerIsAuthorizedIssuer = null;
									            signerIsIssuersMasterKey = null;
									            signerIsNotary = null;
											}
										}
									}
								}
							}
						}
					}
				}
			}
			else
				throw new Exception("Found unsupported Data: " + o);
		}
		else
		{
			isValid = Boolean.FALSE;
			throw new Exception("Message contains no valid object or is not a valid PGP Message");
		}
	}
	
	private Boolean applyValidityChecks(Boolean valueToSet, PGPPublicKey pubKey) throws InvalidPGPMessageException
	{
		log.debug("The key "+pubKey.getKeyID()+" (Version "+pubKey.getVersion()+") that was used to sign the message is "+(pubKey.isMasterKey()?"":" not ")+"a master key.");
		log.debug("The key "+pubKey.getKeyID()+" that was used to sign the message is valid for " + pubKey.getValidSeconds() + " seconds.");
		log.debug("The key "+pubKey.getKeyID()+" that was used to sign the message is valid for " + pubKey.getValidDays() + " days.");
		
		if (Boolean.TRUE.equals(valueToSet))
		{
			if (pubKey.isRevoked())
				throw new InvalidPGPMessageException(THE_KEY_THAT_WAS_USED_TO_SIGN_THE_MESSAGE_IS_REVOKED);
			else if (pubKey.getValidSeconds() != 0 && BouncycastleUtil.keyHasExpired(pubKey))
			{
				throw new InvalidPGPMessageException(THE_KEY_THAT_WAS_USED_TO_SIGN_THE_MESSAGE_IS_NO_MORE_VALID);
			}
			else
				return Boolean.TRUE;
		}
		else
			return Boolean.FALSE;
	}

	private PGPPublicKey findMasterKey(PGPPublicKeyRingCollection publicKeyRingCollection, long keyID) throws PGPException
	{
		PGPPublicKeyRing masterKeyRing = publicKeyRingCollection.getPublicKeyRing(keyID);

		if (masterKeyRing != null)
		{
			return extractMasterKeyFromRing(masterKeyRing);
		}
		
		return null;
	}

	private static PGPPublicKey extractMasterKeyFromRing(PGPPublicKeyRing searchRing)
	{
		Iterator<PGPPublicKey> iter = searchRing.getPublicKeys();
		while(iter.hasNext())
		{
			PGPPublicKey key = iter.next();
			if (key.isMasterKey())
				return key;
		}
		
		return null;
	}

	private URI extractURIDataFromPGPLiteral(PGPLiteralData lit) throws IOException, URISyntaxException
	{
		log.trace("Decrypted decompressed literal's format is: " + (char)lit.getFormat());
		BufferedReader reader = new BufferedReader(new InputStreamReader(lit.getDataStream()));

		String request = reader.readLine();
		URI uri = new URI(request);
		
		log.trace("URI: " + uri);
		log.trace("Path: " + uri.getPath());
		log.trace("Query: " + uri.getQuery());
		
		return uri;
	}

	private String extractClearTextFromPGPLiteral(PGPLiteralData lit) throws IOException, URISyntaxException
	{
		log.trace("Decrypted decompressed literal's format is: " + (char)lit.getFormat());
		BufferedReader reader = new BufferedReader(new InputStreamReader(lit.getDataStream(), "UTF8"));
		
		StringBuffer clearTextBuffer = new StringBuffer();
		
		String line;
		while((line = reader.readLine()) != null)
		{
			log.debug("Read line: " + line);
			clearTextBuffer.append(line + "\n");
		}
		
		return clearTextBuffer.toString();
	}

	private void verifyClearSignedMessage(InputStream in) throws Exception
	{
        ArmoredInputStream    aIn = new ArmoredInputStream(in);

        //
        // read the input, making sure we ingore the last newline.
        //
        int                   ch, lastCh;
        boolean               newLine = false;
        ByteArrayOutputStream bOut = new ByteArrayOutputStream();

        lastCh = 0;
        
        while ((ch = aIn.read()) >= 0 && aIn.isClearText())
        {
            if (lastCh == '\r' && ch == '\n')
            {
                continue;
            }
            
            if (newLine)
            {
                bOut.write(lastCh);
                newLine = false;
            }
            
            if (ch == '\r' || ch == '\n')
            {
                lastCh = ch;
                newLine = true;
                continue;
            }

            bOut.write((byte)ch);
            lastCh = ch;
        }
        
        PGPObjectFactory           pgpFact = new PGPObjectFactory(aIn);
        PGPSignatureList           p3 = (PGPSignatureList)pgpFact.nextObject();
        PGPSignature               sig = p3.get(0);

        PGPPublicKey signKey = lookupSigningKey(sig.getKeyID());
        
        if (signKey != null) {
        	log.debug("The key used to sign is: " + signKey + (signKey != null ? " id: 0x" + String.valueOf(Hex.encodeHex(signKey.getFingerprint())) + " isMaster: " + signKey.isMasterKey() + " isNotary: " + signerIsNotary + " isAuthorizedIssuer: " + signerIsAuthorizedIssuer + " signerIsIssuersMasterKey: " + signerIsIssuersMasterKey : ""));

        	log.debug("Signature created " + sig.getCreationTime() + " hash algorithm " + sig.getHashAlgorithm() + " key algorithm: " + sig.getKeyAlgorithm() +" signatureType: " + sig.getSignatureType() + " signatureTrailer: '" + BouncycastleUtil.dumpPacketStructure(sig.getSignatureTrailer()) + "' keyID: " + BouncycastleUtil.convertKeyIDToHex(sig.getKeyID()));

        	sig.initVerify(signKey, "BC");

        	sig.update(bOut.toByteArray());

        	if (sig.verify())
        	{
        		log.debug("signature verified.");
        		isValid = applyValidityChecks(Boolean.TRUE, signKey);
        		clearText = new String(bOut.toByteArray(), "UTF-8") + "\n";
        		keyId = new Long(sig.getKeyID());
        	}
        	else
        	{
        		log.debug("signature verification failed.");
        		isValid = Boolean.FALSE;
        		signerIsAuthorizedIssuer = null;
        		signerIsIssuersMasterKey = null;
        		signerIsNotary = null;
        		signersFingerprint = null;
				signersPrimaryKeyFingerprint = null;
				keyId = null;
        	}
        } else {
        	log.warn("The key " + sig.getKeyID() + " used to sign the message could not be looked up.");
        }
    }

	//	private void debugDumpKeyRings(PGPSecretKeyRingCollection secRing, PGPPublicKeyRingCollection pubRing)
	//	{
	//		Iterator outIter = pubRing.getKeyRings();
	//		while(outIter.hasNext())
	//		{
	//			PGPPublicKeyRing ring = (PGPPublicKeyRing)outIter.next();
	//			Iterator iter = ring.getPublicKeys();
	//			while(iter.hasNext())
	//			{
	//				PGPPublicKey publicKey = ((PGPPublicKey)iter.next());
	//				log.trace("Got public key: " + publicKey.getKeyID() + " algo: " + publicKey.getAlgorithm() + " FingerPrint: " + new String(Hex.encodeHex(publicKey.getFingerprint()))
	//				+ " daysvalid " + publicKey.getValidDays());
	//			}
	//		}
	//		
	//		outIter = secRing.getKeyRings();
	//		while(outIter.hasNext())
	//		{
	//			PGPSecretKeyRing ring = (PGPSecretKeyRing)outIter.next();
	//			
	//			Iterator iter = ring.getSecretKeys();
	//			while(iter.hasNext())
	//			{
	//				PGPSecretKey secretKey = ((PGPSecretKey)iter.next());
	//				log.trace("Got secret key: " + secretKey.getKeyID() + secretKey.isMasterKey() + secretKey.isSigningKey());
	//			}
	//		}
	//	}
	
	private void stripTrailingSpaces() throws UnsupportedEncodingException {
		
		String messageToStrip = new String(message, "UTF-8");
		
		messageToStrip = messageToStrip.replaceAll("\\s+$", "");
		
		message = messageToStrip.getBytes("UTF-8");
	}

	private PGPPublicKey lookupSigningKey(long signingKeyId)
			throws PGPException, Exception, RedcentProcessingException,
			IOException, UnsupportedEncodingException {
		
		PGPPublicKeyRing signRing = null;
		PGPPublicKey signKey = null; 
		if (getEmbeddedNotaryPubRing()!=null)
			signKey = getEmbeddedNotaryPubRing().getPublicKey(signingKeyId);
		
        if (signKey == null)
        {
        	AuthorizedIssuer ai = null; 
        	if (authorizedIssuerStore!=null)
        		ai = authorizedIssuerStore.findAuthorizedIssuerByKeyID(signingKeyId);

        	if (ai != null) {
        		ArmoredInputStream aiKeyIn = new ArmoredInputStream(new ByteArrayInputStream(ai.getPublicKey().getBytes("UTF-8")));
        		PGPPublicKeyRing aiPublicKeyRing = new PGPPublicKeyRing(aiKeyIn);
				signKey = aiPublicKeyRing.getPublicKey(signingKeyId);
        		aiKeyIn.close();

        		if (signKey != null) {
        			log.debug("Public key for signature verification is " + signKey);
        			signerIsNotary = ai.hasNotaryRights();
        			signerIsAuthorizedIssuer = Boolean.TRUE;
        			signRing = aiPublicKeyRing;
        		} else {
        			log.warn("AI retrieved from DB does not contain requested key " + signingKeyId);
            		signerIsAuthorizedIssuer = Boolean.FALSE;
        		}
        	} else {
        		log.debug("Cannot resolve AI key for id " + signingKeyId);
        		signerIsAuthorizedIssuer = Boolean.FALSE;
        	}
        	
        	if (signKey == null) {
        		// Lookup key in public key rings
        		signKey = getPubRing().getPublicKey(signingKeyId);

        		if (signKey != null) {
        			log.debug("Public key for clearsigned message verification is contained in the issuer's collection of public keys" + signKey);
        			signRing = getPubRing().getPublicKeyRing(signingKeyId);
        		} else {
            		// Lookup key in public key rings
        			if (getTemporaryPubRing()!=null)
        				signKey = getTemporaryPubRing().getPublicKey(signingKeyId);
            		
            		if (signKey == null)
            			log.debug("Encountered unknown public key + " + signingKeyId);
            		else {
            			log.debug("Found temporary public key " + signKey);
            			signRing = getTemporaryPubRing().getPublicKeyRing(signingKeyId);
            		}
        		}
        	}
        }
        else
        {
			log.debug("Public key for clearsigned message verification is an embedded NOTARY key " + signKey);
			signerIsNotary = Boolean.TRUE;
			signRing = getEmbeddedNotaryPubRing().getPublicKeyRing(signingKeyId);
			
			AuthorizedIssuer ai = authorizedIssuerStore.findAuthorizedIssuerByKeyID(signingKeyId);
			if (ai != null) {
				if (!ai.hasNotaryRights())
					log.warn("Encoutered embedded notary " + ai.getFingerprint() + " who does does not have notaryRights granted in authorizedIssuerStore!");
				
				signerIsAuthorizedIssuer = Boolean.TRUE;
			} else {
        		signerIsAuthorizedIssuer = Boolean.FALSE;
			}
        }
        
        // lookup and set fingerprint and primary fingerprint
        if (signKey != null && signRing != null) {
    		signersFingerprint = new String(Hex.encodeHex(signKey.getFingerprint())).toUpperCase();

    		if (signKey.isMasterKey()) {
    			signersPrimaryKeyFingerprint = new String(Hex.encodeHex(signKey.getFingerprint())).toUpperCase();
    		} else {
    			PGPPublicKey masterKey = extractMasterKeyFromRing(signRing);

    			if (masterKey != null)
    				signersPrimaryKeyFingerprint = new String(Hex.encodeHex(masterKey.getFingerprint())).toUpperCase();
    			else
    				signersPrimaryKeyFingerprint = null;
    		}
        } else {
        	signersFingerprint = null;
        	signersPrimaryKeyFingerprint = null;
        }
        
        if (issuerMasterPubRing != null) {
        	// verify if this signature was created by the issuer's master key
        	PGPPublicKey sigIssuerMasterKey = issuerMasterPubRing.getPublicKey(signingKeyId);
        	signerIsIssuersMasterKey = (sigIssuerMasterKey != null);
        }

        return signKey;
	}

	public synchronized void addTemporaryPublicKey(String asciiArmoredPublicKey) throws Exception
	{
		PGPPublicKeyRingCollection pCol = new PGPPublicKeyRingCollection(PGPUtil.getDecoderStream(new ByteArrayInputStream(asciiArmoredPublicKey.getBytes())));
		if (pCol.size() == 0)
			throw new Exception("No valid public key could be extracted from data: " + asciiArmoredPublicKey);
		else
		{
			
			if (tempPubRing == null)
				tempPubRing = pCol;
			else {
				Iterator<PGPPublicKeyRing> iter = pCol.getKeyRings();
				while(iter.hasNext()) {
					PGPPublicKeyRing currentKeyRing = iter.next();
					if (!tempPubRing.contains(currentKeyRing.getPublicKey().getKeyID()))
						tempPubRing = PGPPublicKeyRingCollection.addPublicKeyRing(tempPubRing, currentKeyRing);
				}
			}
		}
	}
	
	public String getSignersPublicKey() throws Exception
	{
		if (isValid() != null && Boolean.TRUE.equals(isValid()) && getSignersFingerprint() != null && keyId != null)
		{
			// Look for public key first
			PGPPublicKey key = getPubRing().getPublicKey(keyId.longValue());
			
			// If not found, the message might be signed with a trusted public key
			if (key == null)
				getEmbeddedNotaryPubRing().getPublicKey(keyId.longValue());
			
			if (key != null)
			{
				ByteArrayOutputStream bOut = new ByteArrayOutputStream();
				ArmoredOutputStream aOut = new ArmoredOutputStream(bOut);
				aOut.write(key.getEncoded());
				
				// flush and close the stream
				aOut.flush();
				aOut.close();
				
				// return the ascii armored public key
				return bOut.toString();
			}
		}
		
		return null;
	}

	private PGPPublicKeyRingCollection getPubRing() throws Exception
	{
		return pubRing;
	}

	private PGPSecretKeyRingCollection getSecRing() throws Exception
	{
		return secRing;
	}

	private PGPPublicKeyRingCollection getEmbeddedNotaryPubRing() throws Exception
	{
		return embeddedNotaryPubRing;
	}
	
	private PGPPublicKeyRingCollection getTemporaryPubRing() throws Exception 
	{
		return tempPubRing;
	}
}
