/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.crypto;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.Security;
import java.security.SignatureException;
import java.util.Date;

import javax.annotation.PostConstruct;

import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.bcpg.ArmoredOutputStream;
import org.bouncycastle.bcpg.BCPGOutputStream;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.openpgp.PGPException;
import org.bouncycastle.openpgp.PGPLiteralData;
import org.bouncycastle.openpgp.PGPLiteralDataGenerator;
import org.bouncycastle.openpgp.PGPPrivateKey;
import org.bouncycastle.openpgp.PGPPublicKey;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPSecretKey;
import org.bouncycastle.openpgp.PGPSecretKeyRing;
import org.bouncycastle.openpgp.PGPSignature;
import org.bouncycastle.openpgp.PGPSignatureGenerator;
import org.bouncycastle.openpgp.PGPSignatureSubpacketGenerator;
import org.bouncycastle.openpgp.PGPUtil;

import de.nachtlicht_media.epoint.redcent.exception.ConflictException;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageSignatureCreator;
import de.nachtlicht_media.epoint.redcent.interfaces.RedCentConfig;
import de.nachtlicht_media.epoint.redcent.util.Splitter;

public class BouncycastleMessageSignatureCreator implements MessageSignatureCreator
{
	private static final Log log = LogFactory.getLog(BouncycastleMessageSignatureCreator.class);
	static
	{
        Security.addProvider(new BouncyCastleProvider());
	}
		
	private String issuersPrivateKeyRingPath = null;
	private String issuersKeyPassword = null;
	private int textSignatureVersion = -1;
	private int binarySignatureVersion = -1;

	private PGPSecretKeyRing issuersSecretKeyRing;

	private PGPSecretKey issuersSecretKey;

	private PGPPrivateKey issuersPgpPrivateKey;
	
	public BouncycastleMessageSignatureCreator(RedCentConfig config, BouncycastleCryptoConfig cryptoConfig)
	{
		this.issuersPrivateKeyRingPath = config.getIssuersPrivateKeyRingPath();
		this.issuersKeyPassword = config.getIssuersPrivateKeyPassword();
		this.textSignatureVersion = cryptoConfig.getTextSignatureVersion();
		this.binarySignatureVersion = cryptoConfig.getBinarySignatureVersion();
	}
	
	protected int getBinarySignatureVersion()
	{
		return binarySignatureVersion;
	}

	protected String getIssuersKeyPassword()
	{
		return issuersKeyPassword;
	}

	protected String getIssuersPrivateKeyRingPath()
	{
		return issuersPrivateKeyRingPath;
	}

	protected int getTextSignatureVersion()
	{
		return textSignatureVersion;
	}

	@PostConstruct
	public void init() throws ConflictException
	{
		try
		{
			log.trace("Loading PGP-Keyrings...");
			
			issuersSecretKeyRing = new PGPSecretKeyRing(PGPUtil.getDecoderStream(new FileInputStream(getIssuersPrivateKeyRingPath())));
			issuersSecretKey = issuersSecretKeyRing.getSecretKey();
			issuersPgpPrivateKey = issuersSecretKey.extractPrivateKey(getIssuersKeyPassword().toCharArray(), "BC");
			
			log.trace("PGP-Keyrings loaded.");
		}
		catch(PGPException pgpEx)
		{
			throw new ConflictException("Error reading secret keyring from storage.", pgpEx);
		}
		catch(IOException ioEx)
		{
			throw new ConflictException("IO Error while reading secret key data.", ioEx);
		}
		catch(NoSuchProviderException nsPex)
		{
			throw new ConflictException("Could not find suitable crypto provider", nsPex);
		}
	}
	
	public byte[] signMessage(String message) throws ConflictException
	{
		return signMessage(message, getTextSignatureVersion(), issuersSecretKey, issuersPgpPrivateKey);
		/*
		//TODO: check if message header is well-formed?
		try
		{
			V3V4PGPSignatureGenerator sGen = new V3V4PGPSignatureGenerator(getTextSignatureVersion(), issuersSecretKey.getPublicKey().getAlgorithm(), PGPUtil.SHA1, "BC");
			sGen.initSign(PGPSignature.CANONICAL_TEXT_DOCUMENT, issuersPgpPrivateKey);

			int ch = 0;
			int lastCh = 0;

			ByteArrayOutputStream  bOut = new ByteArrayOutputStream();
			ArmoredOutputStream    aOut = new ArmoredOutputStream(bOut);
			ByteArrayInputStream   bIn = new ByteArrayInputStream(message.getBytes("UTF-8"));

			aOut.beginClearText(PGPUtil.SHA1);

			boolean newLine = false;

			//
			// note the last \n in the file is ignored
			//
			while ((ch = bIn.read()) >= 0)
			{
				aOut.write(ch);

				if (lastCh == '\r' && ch == '\n')
				{
					continue;
				}

				if (newLine)
				{
					sGen.update((byte)lastCh);
					newLine = false;
				}

				if (ch == '\r' || ch == '\n')
				{
					lastCh = ch;
					newLine = true;
					continue;
				}

				sGen.update((byte)ch);
				lastCh = ch;
			}

			aOut.endClearText();

			BCPGOutputStream            bcpgOut = new BCPGOutputStream(aOut);

			sGen.generate().encode(bcpgOut);

			aOut.close();

			return bOut.toByteArray();
		}
		catch(PGPException pgpEx)
		{
			throw new ConflictException("Error reading secret keyring from storage.", pgpEx);
		}
		catch(IOException ioEx)
		{
			throw new ConflictException("IO Error while reading secret key data.", ioEx);
		}
		catch(NoSuchProviderException nsPex)
		{
			throw new ConflictException("Could not find suitable crypto provider", nsPex);
		}
		catch(SignatureException sEx)
		{
			throw new ConflictException("Could not create signature", sEx);
		}
		catch(NoSuchAlgorithmException sEx)
		{
			throw new ConflictException("Could not find suitable algorithm", sEx);
		}
		*/
	}
	
	public static byte[] signMessage(String message, int signatureVersion, PGPSecretKey issuersSecretKey, PGPPrivateKey privateKey) throws ConflictException
	{
		try
		{
			V3V4PGPSignatureGenerator sGen = new V3V4PGPSignatureGenerator(signatureVersion, issuersSecretKey.getPublicKey().getAlgorithm(), PGPUtil.SHA1, "BC");
			sGen.initSign(PGPSignature.CANONICAL_TEXT_DOCUMENT, privateKey);

			int ch = 0;
			int lastCh = 0;

			ByteArrayOutputStream  bOut = new ByteArrayOutputStream();
			ArmoredOutputStream    aOut = new ArmoredOutputStream(bOut);
			ByteArrayInputStream   bIn = new ByteArrayInputStream(message.getBytes("UTF-8"));

			aOut.beginClearText(PGPUtil.SHA1);

			boolean newLine = false;

			//
			// note the last \n in the file is ignored
			//
			while ((ch = bIn.read()) >= 0)
			{
				aOut.write(ch);

				if (lastCh == '\r' && ch == '\n')
				{
					continue;
				}

				if (newLine)
				{
					sGen.update((byte)lastCh);
					newLine = false;
				}

				if (ch == '\r' || ch == '\n')
				{
					lastCh = ch;
					newLine = true;
					continue;
				}

				sGen.update((byte)ch);
				lastCh = ch;
			}

			aOut.endClearText();

			BCPGOutputStream            bcpgOut = new BCPGOutputStream(aOut);

			sGen.generate().encode(bcpgOut);

			aOut.close();

			return bOut.toByteArray();
		}
		catch(PGPException pgpEx)
		{
			throw new ConflictException("Error reading secret keyring from storage.", pgpEx);
		}
		catch(IOException ioEx)
		{
			throw new ConflictException("IO Error while reading secret key data.", ioEx);
		}
		catch(NoSuchProviderException nsPex)
		{
			throw new ConflictException("Could not find suitable crypto provider", nsPex);
		}
		catch(SignatureException sEx)
		{
			throw new ConflictException("Could not create signature", sEx);
		}
		catch(NoSuchAlgorithmException sEx)
		{
			throw new ConflictException("Could not find suitable algorithm", sEx);
		}
	}

	public byte[] signBinaryContent(byte[] content, boolean armor) throws ConflictException
	{    
		try
		{
			OutputStream out = new ByteArrayOutputStream();
			final ByteArrayOutputStream dataOut = (ByteArrayOutputStream)out;
	
			if (armor)
			{
				out = new ArmoredOutputStream(out);
			}
	
			V3V4PGPSignatureGenerator       sGen = new V3V4PGPSignatureGenerator(getBinarySignatureVersion(), issuersSecretKey.getPublicKey().getAlgorithm(), PGPUtil.SHA1, "BC");
	
			sGen.initSign(PGPSignature.BINARY_DOCUMENT, issuersPgpPrivateKey);
	
			BCPGOutputStream            bOut = new BCPGOutputStream(out);
	
			sGen.generateOnePassVersion(false).encode(bOut);
	
			PGPLiteralDataGenerator     lGen = new PGPLiteralDataGenerator(true);
			OutputStream                lOut = lGen.open(bOut, PGPLiteralData.BINARY, "", content.length, new Date());
			ByteArrayInputStream		fIn = new ByteArrayInputStream(content);
			int                         ch = 0;
	
			int count = 0;
			while ((ch = fIn.read()) >= 0)
			{
				lOut.write(ch);
				sGen.update((byte)ch);
				count++;
			}
	
			sGen.generate().encode(bOut);
			bOut.flush();
			lGen.close();
	
			out.flush();
			out.close();
	
			dataOut.flush();
			dataOut.close();
	
			return dataOut.toByteArray();
		}
		catch(PGPException pgpEx)
		{
			throw new ConflictException("Error reading secret keyring from storage.", pgpEx);
		}
		catch(IOException ioEx)
		{
			throw new ConflictException("IO Error while reading secret key data.", ioEx);
		}
		catch(NoSuchProviderException nsPex)
		{
			throw new ConflictException("Could not find suitable crypto provider", nsPex);
		}
		catch(SignatureException sEx)
		{
			throw new ConflictException("Could not create signature", sEx);
		}
		catch(NoSuchAlgorithmException sEx)
		{
			throw new ConflictException("Could not find suitable algorithm", sEx);
		}
	}

	public byte[] signKey(byte[] publicKeyData, boolean armor) throws ConflictException
	{    
		try
		{
			PGPPublicKeyRing pubRing = new PGPPublicKeyRing(new ByteArrayInputStream(publicKeyData));
			
			OutputStream out = new ByteArrayOutputStream();
	
			if (armor)
			{
				out = new ArmoredOutputStream(out);
			}
	
			PGPSignatureGenerator       sGen = new PGPSignatureGenerator(issuersSecretKey.getPublicKey().getAlgorithm(), PGPUtil.SHA1, "BC");
	
			sGen.initSign(PGPSignature.DIRECT_KEY, issuersPgpPrivateKey);
	
			BCPGOutputStream            bOut = new BCPGOutputStream(out);
	
			sGen.generateOnePassVersion(false).encode(bOut);
	
			PGPSignatureSubpacketGenerator spGen = new PGPSignatureSubpacketGenerator();

			spGen.setNotationData(true, true, "@epointSystem.org.attachment", "This guy is a really bad one...");
			sGen.setHashedSubpackets(spGen.generate());
			
			bOut.flush();

			return PGPPublicKey.addCertification(pubRing.getPublicKey(), sGen.generate()).getEncoded();
		}
		catch(PGPException pgpEx)
		{
			throw new ConflictException("Error reading secret keyring from storage.", pgpEx);
		}
		catch(IOException ioEx)
		{
			throw new ConflictException("IO Error while reading secret key data.", ioEx);
		}
		catch(NoSuchProviderException nsPex)
		{
			throw new ConflictException("Could not find suitable crypto provider", nsPex);
		}
		catch(SignatureException sEx)
		{
			throw new ConflictException("Could not create signature", sEx);
		}
		catch(NoSuchAlgorithmException sEx)
		{
			throw new ConflictException("Could not find suitable algorithm", sEx);
		}
	}
	
	private String issuerFingerPrintText = null;
	public String getIssuerFingerPrintText()
	{
		if (issuerFingerPrintText == null)
		{
				byte[] fp=issuersSecretKey.getPublicKey().getFingerprint();
				if(fp==null) return "";

				if(fp.length==16) return Splitter.group(new String(Hex.encodeHex(fp)).toUpperCase(),2,8);
				issuerFingerPrintText = Splitter.group(new String(Hex.encodeHex(fp)).toUpperCase(),4,5);
		}
		
		return issuerFingerPrintText;
	}
}
