/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.crypto;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.MutableTreeNode;
import javax.swing.tree.TreeNode;

import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.bcpg.ArmoredInputStream;
import org.bouncycastle.bcpg.ArmoredOutputStream;
import org.bouncycastle.bcpg.PublicKeyAlgorithmTags;
import org.bouncycastle.bcpg.sig.NotationData;
import org.bouncycastle.openpgp.PGPCompressedData;
import org.bouncycastle.openpgp.PGPEncryptedDataList;
import org.bouncycastle.openpgp.PGPException;
import org.bouncycastle.openpgp.PGPLiteralData;
import org.bouncycastle.openpgp.PGPMarker;
import org.bouncycastle.openpgp.PGPObjectFactory;
import org.bouncycastle.openpgp.PGPOnePassSignature;
import org.bouncycastle.openpgp.PGPOnePassSignatureList;
import org.bouncycastle.openpgp.PGPPBEEncryptedData;
import org.bouncycastle.openpgp.PGPPublicKey;
import org.bouncycastle.openpgp.PGPPublicKeyEncryptedData;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPPublicKeyRingCollection;
import org.bouncycastle.openpgp.PGPSecretKey;
import org.bouncycastle.openpgp.PGPSecretKeyRing;
import org.bouncycastle.openpgp.PGPSignature;
import org.bouncycastle.openpgp.PGPSignatureList;
import org.bouncycastle.openpgp.PGPSignatureSubpacketVector;
import org.bouncycastle.openpgp.PGPUserAttributeSubpacketVector;

import de.nachtlicht_media.epoint.redcent.AuthorizedIssuer;
import de.nachtlicht_media.epoint.redcent.exception.RedcentProcessingException;
import de.nachtlicht_media.epoint.redcent.interfaces.AuthorizedIssuerStore;

public class BouncycastleUtil
{
	private static final String AI_SIGNATURE_ISSUER_BINDING_NOTATION_NAME = "issuer@epointsystem.org";
	private static final Log log = LogFactory.getLog(BouncycastleUtil.class);
	
	public static boolean keyContainsValidSigningKey(PGPPublicKeyRing pubRing) {
		
		Iterator<PGPPublicKey> iter = pubRing.getPublicKeys();
		while(iter.hasNext())
			if (keyIsSuitableForSigning(iter.next()))
				return true;
		
		return false;
	}

	public static boolean keyIsSuitableForSigning(PGPPublicKey pubKey) {
		
		if (pubKey != null) {
			
			int algorithm = pubKey.getAlgorithm();
			
			return !pubKey.isRevoked() && ((algorithm == PublicKeyAlgorithmTags.RSA_GENERAL) || (algorithm == PublicKeyAlgorithmTags.RSA_SIGN)
					|| (algorithm == PublicKeyAlgorithmTags.DSA));
		}
		
		return false;
	}

	public static boolean keyIsSuitableForEncryption(PGPPublicKey pubKey)
	{
		PGPSignature mostRecentFlagSignature = null;
		
		Iterator iter = pubKey.getSignatures();
		while(iter.hasNext())
		{
			PGPSignature sig = (PGPSignature)iter.next();
			
			if (
					(	pubKey.isMasterKey() && sig.getKeyID() == pubKey.getKeyID()
						&& 
						(
							sig.getSignatureType() == PGPSignature.POSITIVE_CERTIFICATION 
							|| sig.getSignatureType() == PGPSignature.DIRECT_KEY
						)
					)
				||	(!pubKey.isMasterKey() && sig.getSignatureType() == PGPSignature.SUBKEY_BINDING))
			{
				mostRecentFlagSignature = sig;
			}
		}
		
		if (mostRecentFlagSignature != null)
		{
			PGPSignatureSubpacketVector hsp = mostRecentFlagSignature.getHashedSubPackets();
	
			if (hsp != null)
			{
				int flags = hsp.getKeyFlags();
				
				if (flags != 0)
				{
					if (((flags & 0x04) == 0x04) || ((flags & 0x08) == 0x08))
						return true;
					else
						return false;
				}
			}
		}
	
		int algorithm = pubKey.getAlgorithm();
	
		return ((algorithm == PublicKeyAlgorithmTags.RSA_GENERAL) || (algorithm == PublicKeyAlgorithmTags.RSA_ENCRYPT)
				|| (algorithm == PublicKeyAlgorithmTags.ELGAMAL_ENCRYPT) || (algorithm == PublicKeyAlgorithmTags.ELGAMAL_GENERAL));
	}
	
	public static long[] getValidEncryptionKeys(PGPPublicKeyRingCollection collection) {

		ArrayList<Long> list = new ArrayList<Long>();

		Iterator<PGPPublicKeyRing> iter = collection.getKeyRings();
		while(iter.hasNext())
		{
			PGPPublicKeyRing ring = iter.next();

			PGPPublicKey mostRecent = null;

			Iterator<PGPPublicKey> subIter = ring.getPublicKeys();
			while (subIter.hasNext())
			{
				PGPPublicKey key = subIter.next();

				log.trace("Checking key: " + key + (key != null ? " id: 0x" + String.valueOf(Hex.encodeHex(key.getFingerprint())) + " isMaster: " + key.isMasterKey() + " isEncryptionKey: " + key.isEncryptionKey() + " isSuitableForEncryption: " + BouncycastleUtil.keyIsSuitableForEncryption(key) + " isRevoked: " + key.isRevoked(): ""));

				if (BouncycastleUtil.keyIsSuitableForEncryption(key) && !key.isRevoked() && !BouncycastleUtil.keyHasExpired(key))
				{
					if (mostRecent == null)
						mostRecent = key;
					else
					{
						if (key.getCreationTime().getTime() > mostRecent.getCreationTime().getTime())
							mostRecent = key;
					}
				}
			}

			if (mostRecent != null)
				list.add(new Long(mostRecent.getKeyID()));
		}

		long[] foundIds = new long[list.size()];
		for (int i = 0; i < foundIds.length; i++)
			foundIds[i] = ((Long)list.get(i)).longValue();
		
		return foundIds;
	}
	
	public static PGPPublicKeyRing readArmoredPublicKeyRing(byte[] pubKeyRing) throws IOException {
		ArmoredInputStream aIn = new ArmoredInputStream(new ByteArrayInputStream(pubKeyRing));
		PGPPublicKeyRing ring = new PGPPublicKeyRing(aIn);
		aIn.close();
		return ring;
	}
	
	public static String writeArmoredPublicKeyRing(PGPPublicKeyRing keyRing) throws IOException {
		ByteArrayOutputStream bOut = new ByteArrayOutputStream();
		ArmoredOutputStream aOut = new ArmoredOutputStream(bOut);
		aOut.write(keyRing.getEncoded());
		aOut.flush();
		aOut.close();
		bOut.flush();
		bOut.close();
		return new String(bOut.toByteArray(), "UTF8");
	}
	
	public static String getPublicKeyIDList(byte[] pubKeyRing) throws IOException {
		return getKeyIDList(readArmoredPublicKeyRing(pubKeyRing));
	}

	public static String getKeyIDList(PGPPublicKeyRing pubKeyRing) {
		if (pubKeyRing == null)
			return null;
		
		StringBuffer buf = new StringBuffer();
		
		Iterator<PGPPublicKey> keyIter = pubKeyRing.getPublicKeys();
		while(keyIter.hasNext())
		{
			PGPPublicKey key = keyIter.next();
			if (!key.isRevoked())
			{
				if (buf.length() > 0)
					buf.append(",");
				buf.append(convertKeyIDToHex(key.getKeyID()));
			}
		}
		
		return buf.toString();
	}
	
	public static String getFingerprintList(Set<AuthorizedIssuer> ais) {
		if (ais == null)
			return null;
		
		StringBuffer buf = new StringBuffer();
		Iterator<AuthorizedIssuer> aiIter = ais.iterator();
		while (aiIter.hasNext()) {
			AuthorizedIssuer ai = aiIter.next();
			if (buf.length() > 0)
				buf.append(",");
			buf.append(ai.getFingerprint());
		}
		return buf.toString();
	}
	
	public static PGPPublicKey getPrimaryKey(PGPPublicKeyRing ring) {
		Iterator<PGPPublicKey> iter = ring.getPublicKeys();
		while(iter.hasNext()) {
			PGPPublicKey key = iter.next();
			if (key.isMasterKey())
				return key;
		}
		
		return null;
	}
	
	public static String convertKeyIDToHex(long keyID) {
		
		final char lpad = '0';
		StringBuffer buf = new StringBuffer(Long.toHexString(keyID));
		
		while(buf.length() < 16)
			buf.insert(0,lpad);
		
		return buf.toString().toUpperCase();
	}
	
	public static long convertKeyHexIDToLong(String _fingerprint) {
		
		String fingerprint = _fingerprint;
		
		fingerprint.toLowerCase();
		
		if (fingerprint.startsWith("0x"))
			fingerprint = fingerprint.substring(2);
		try {
			return Long.parseLong(fingerprint, 16);
		} catch (NumberFormatException nfe) {
			// Java seems to have problems to parse some hex values. Signed/Unsigned. Brute-Force as fallback
			return longFromHex(fingerprint);
		}
	}

	private static long longFromHex(String hex) {
		long val = 0;

		for(int i = 0; i < hex.length(); i++) {
			val <<= 4;

			if(Character.isDigit(hex.charAt(i)))
				val += (hex.charAt(i) - '0');
			else
				val += (hex.charAt(i) - 'A' + 10);
		}

		return val;
	}
	
	
	public static String convertKeyFingerprintToHex(byte[] fingerprint) {
		return String.valueOf(Hex.encodeHex(fingerprint)).toUpperCase();
	}

	public static boolean keyHasExpired(PGPPublicKey pubKey)
	{
		if (pubKey.getValidSeconds() == 0)
		{
			log.trace("Checking whether key has expired. Key was created: " + pubKey.getCreationTime() + ", key will never expire!");
			return false;
		}
		
		GregorianCalendar now = new GregorianCalendar();
		 
		GregorianCalendar keyExpiration = new GregorianCalendar();
		keyExpiration.setTime(pubKey.getCreationTime());
		log.trace("Checking whether key has expired. Now: " + now.getTime() + ", key was created: " + keyExpiration.getTime());
	
		keyExpiration.setTimeInMillis(keyExpiration.getTimeInMillis() + (pubKey.getValidSeconds()*1000l));
		log.trace("Key will expire: " + keyExpiration.getTime());
		
		return keyExpiration.before(now);
	}
	
	/**
	 * Extract a Map of valid, non-revoked POSITIVE_CERTIFICATION from Authorized Issuers with notary rights for the given
	 * public key
	 * @param aiStore, the store to retrieve AI data from
	 * @param key, the key to check
	 * @return
	 * @throws RedcentProcessingException
	 */
	public static Map<AuthorizedIssuer, PGPSignature> getNotarySignatures(String issuersFingerprint, AuthorizedIssuerStore aiStore, PGPPublicKey key) throws RedcentProcessingException {
		
		if (key != null) {
			
			Hashtable<String, PGPSignature> signatures = new Hashtable<String, PGPSignature>();
			
			// read all positive certifications, then remove revoked ones
			Iterator<PGPSignature> positiveIter = key.getSignaturesOfType(SignatureType.POSITIVE_CERTIFICATION.getId());
			Iterator<PGPSignature> revocationIter = key.getSignaturesOfType(SignatureType.CERTIFICATION_REVOCATION.getId());
			
			while (positiveIter.hasNext()) {
				PGPSignature sig = positiveIter.next();
				signatures.put(convertKeyIDToHex(sig.getKeyID()), sig);
			}
			
			while(revocationIter.hasNext()) {
				PGPSignature sig = revocationIter.next();
				signatures.remove(convertKeyIDToHex(sig.getKeyID()));
			}
			
			Map<AuthorizedIssuer, PGPSignature> validSignatures = new HashMap<AuthorizedIssuer, PGPSignature>();
			
			Iterator<PGPSignature> remainingIter = signatures.values().iterator();
			while(remainingIter.hasNext()) {
				PGPSignature sig = remainingIter.next();

				if (sig.hasSubpackets()) {
					NotationData[] notationDataOccurences = sig.getHashedSubPackets().getNotationDataOccurences();
					if (notationDataOccurences != null) {
						
						boolean foundAnyOccurencesOfIssuerBinding = false;
						boolean foundAnyOccurenceValidForThisIssuer = false;
						
						for (NotationData data : notationDataOccurences) {
							if (AI_SIGNATURE_ISSUER_BINDING_NOTATION_NAME.equals(data.getNotationName())) {
								foundAnyOccurencesOfIssuerBinding = true;
								
								if (!issuersFingerprint.equals(String.valueOf(data.getNotationValue()).replaceAll(" ", "").trim().toUpperCase())) {
									log.debug("Found AI issuer signature binding for " + data.getNotationValue() + " not valid for this issuer.");
								} else
									foundAnyOccurenceValidForThisIssuer = true;
							}
						}
						
						// if there were bindings present, there has to be one for us, otherwise the key does not qualify
						if (foundAnyOccurencesOfIssuerBinding) {
							log.debug("Found at least one AI to issuer signature binding as a notation subpacket with name " + AI_SIGNATURE_ISSUER_BINDING_NOTATION_NAME);
							if (!foundAnyOccurenceValidForThisIssuer) {
								log.debug("Signature " + sig.getKeyID() + " does not qualify for this issuer, it is skipped.");
								continue;
							}
						} else
							log.debug("Found no AI to issuer signature binding as a notation subpacket with name " + AI_SIGNATURE_ISSUER_BINDING_NOTATION_NAME);
					}
				}
				
				AuthorizedIssuer ai = aiStore.findAuthorizedIssuerByKeyID(sig.getKeyID());
				if (ai != null) {
					if (ai.hasNotaryRights()) {
						validSignatures.put(ai, sig);
					}
				}
			}
			
			return validSignatures;
		}
		
		return null;
	}

	public static PGPPublicKeyRing mergePublicKeys(PGPPublicKeyRing existingRing, PGPPublicKeyRing ringToMerge) throws PGPException, IOException {
		
		if (existingRing == null && ringToMerge != null)
			return ringToMerge;
		else if (existingRing != null && ringToMerge == null)
			return existingRing;
		else {
			
			PGPPublicKeyRing result = existingRing;
			
			Iterator<PGPPublicKey> mergeKeyIterator = ringToMerge.getPublicKeys();
			while(mergeKeyIterator.hasNext()) {
				
				PGPPublicKey keyToMerge = mergeKeyIterator.next();
				
				if (result.getPublicKey(keyToMerge.getKeyID()) == null) {
					result = PGPPublicKeyRing.insertPublicKey(result, keyToMerge);
				} else {
					
					// the existing ring already contains the key to merge
					// perform merge on subobjects
					PGPPublicKey existingKey = result.getPublicKey(keyToMerge.getKeyID());
					
					Iterator<PGPSignature> signaturesToMerge = keyToMerge.getSignatures();
					
					while(signaturesToMerge.hasNext()) {
						PGPSignature sigToMerge = signaturesToMerge.next();
						
						boolean exists = false;
						Iterator<PGPSignature> existingSigFinder = existingKey.getSignatures();
						while(existingSigFinder.hasNext()) {
							PGPSignature foundSig = existingSigFinder.next();
							if (Arrays.equals(foundSig.getEncoded(), sigToMerge.getEncoded()))
								exists = true;
						}
						
						if (!exists)
							existingKey = PGPPublicKey.addCertification(existingKey, sigToMerge);
					}
					result = PGPPublicKeyRing.removePublicKey(result, existingKey);
					result = PGPPublicKeyRing.insertPublicKey(result, existingKey);
				}
			}
			
			return result;
		}
	}
	
	/**
	 * Check whether key has a non-revoked POSITIVE_CERTIFICATION from the issuer master rights for the given
	 * public key
	 * @param key, the key to check
	 * @return
	 * @throws RedcentProcessingException
	 */
	public static boolean hasKeyNotaryRights(String issuersFingerprint, String issuerMasterKeyID, PGPPublicKey key) throws RedcentProcessingException {
		
		if (key != null) {
			
			Hashtable<String, PGPSignature> signatures = new Hashtable<String, PGPSignature>();
			
			// read all positive certifications, then remove revoked ones
			Iterator<PGPSignature> positiveIter = key.getSignaturesOfType(SignatureType.POSITIVE_CERTIFICATION.getId());
			Iterator<PGPSignature> revocationIter = key.getSignaturesOfType(SignatureType.CERTIFICATION_REVOCATION.getId());
			
			while (positiveIter.hasNext()) {
				PGPSignature sig = positiveIter.next();
				signatures.put(convertKeyIDToHex(sig.getKeyID()), sig);
			}
			
			while(revocationIter.hasNext()) {
				PGPSignature sig = revocationIter.next();
				signatures.remove(convertKeyIDToHex(sig.getKeyID()));
			}
			
			Map<Long, PGPSignature> validSignatures = new HashMap<Long, PGPSignature>();
			
			Iterator<PGPSignature> remainingIter = signatures.values().iterator();
			while(remainingIter.hasNext()) {
				PGPSignature sig = remainingIter.next();

				if (sig.hasSubpackets()) {
					NotationData[] notationDataOccurences = sig.getHashedSubPackets().getNotationDataOccurences();
					if (notationDataOccurences != null) {
						
						boolean foundAnyOccurencesOfIssuerBinding = false;
						boolean foundAnyOccurenceValidForThisIssuer = false;
						
						for (NotationData data : notationDataOccurences) {
							if (AI_SIGNATURE_ISSUER_BINDING_NOTATION_NAME.equals(data.getNotationName())) {
								foundAnyOccurencesOfIssuerBinding = true;
								
								if (!issuersFingerprint.equals(String.valueOf(data.getNotationValue()).replaceAll(" ", "").trim().toUpperCase())) {
									log.debug("Found AI issuer signature binding for " + data.getNotationValue() + " not valid for this issuer.");
								} else
									foundAnyOccurenceValidForThisIssuer = true;
							}
						}
						
						// if there were bindings present, there has to be one for us, otherwise the key does not qualify
						if (foundAnyOccurencesOfIssuerBinding) {
							log.debug("Found at least one AI to issuer signature binding as a notation subpacket with name " + AI_SIGNATURE_ISSUER_BINDING_NOTATION_NAME);
							if (!foundAnyOccurenceValidForThisIssuer) {
								log.debug("Signature " + sig.getKeyID() + " does not qualify for this issuer, it is skipped.");
								continue;
							}
						} else
							log.debug("Found no AI to issuer signature binding as a notation subpacket with name " + AI_SIGNATURE_ISSUER_BINDING_NOTATION_NAME);
					}
				}
				
				if (sig.getKeyID() == convertKeyHexIDToLong(issuerMasterKeyID)) {
					validSignatures.put(new Long(sig.getKeyID()), sig);
				}
			}
			
			return validSignatures.size() > 0;
		}
		
		return false;
	}

	public static String dumpPacketStructure(byte[] implData)
	{
		try
		{
			return convertToString(extractStructure(implData));
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
		
		return "< -- error while dumping structure -- >";
	}

	public static String dumpPGPPublicKey(PGPPublicKey implData)
	{
		try
		{
			return convertToString((DefaultMutableTreeNode)extractStructure(implData));
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
		
		return "< -- error while dumping structure -- >";
	}

	public static String dumpPGPPublicKeyRing(PGPPublicKeyRing implData)
	{
		try
		{
			return convertToString((DefaultMutableTreeNode)extractStructure(implData));
		}
		catch(Exception ex)
		{
			ex.printStackTrace();
		}
		
		return "< -- error while dumping structure -- >";
	}
	
	private static String convertToString(DefaultMutableTreeNode node) throws Exception
	{
		StringBuffer buf = new StringBuffer("<-\n");
		
		buf.append(dumpNode(node, 0));
		
		buf.append("->");
		return buf.toString();
	}

	private static String dumpNode(TreeNode node, int indent)
	{
		String res = "";
		
		for (int i = 0; i < indent; i++)
			res += "\t";
		
		res += node.toString();
		
		if (node.getChildCount() > 0)
		{
			for (int i = 0; i < indent; i++)
				res += "\t";

			res +="{\n";

			for (int i = 0; i < node.getChildCount(); i++)
			{
				res += dumpNode(node.getChildAt(i), indent+1);

				if (i < (node.getChildCount()-1))
					res += ",\n";
			}

			for (int i = 0; i < indent; i++)
				res += "\t";
			res += "}\n";
		}
		
		return res;
	}

	public static DefaultMutableTreeNode extractStructure(byte[] message) throws Exception
	{
		ArmoredInputStream aIn = new ArmoredInputStream(new ByteArrayInputStream(message));
		PGPObjectFactory fac = new PGPObjectFactory(aIn);
		
		DefaultMutableTreeNode root = new DefaultMutableTreeNode(new String(message, "UTF8"));
		
		Object o = null;

		try {
			while ((o = fac.nextObject()) != null)
			{
				extractStructure(root, o);
			}
		} catch (IOException ex) {
			root.add(new DefaultMutableTreeNode("Unable to extract structure: " + ex.getMessage()));
		}
		
		return root;
	}

	private static void extractStructure(DefaultMutableTreeNode root, Object o) throws Exception
	{
		if (o instanceof PGPSignatureList)
		{
			PGPSignatureList p = (PGPSignatureList)o;
			root.add(extractStructure(p));
		}
		else if (o instanceof PGPPublicKeyRing)
		{
			PGPPublicKeyRing p = (PGPPublicKeyRing)o;
			root.add(extractStructure(p));
		}
		else if (o instanceof PGPSecretKeyRing)
		{
			PGPSecretKeyRing p = (PGPSecretKeyRing)o;
			root.add(extractStructure(p));
		}
		else if (o instanceof PGPCompressedData)
		{
			PGPCompressedData p = (PGPCompressedData)o;
			root.add(extractStructure(p));
		}
		else if (o instanceof PGPLiteralData)
		{
			PGPLiteralData p = (PGPLiteralData)o;
			root.add(extractStructure(p));
		}
		else if (o instanceof PGPEncryptedDataList)
		{
			PGPEncryptedDataList p = (PGPEncryptedDataList)o;
			root.add(extractStructure(p));
		}
		else if (o instanceof PGPOnePassSignatureList)
		{
			PGPOnePassSignatureList p = (PGPOnePassSignatureList)o;
			root.add(extractStructure(p));
		}
		else if (o instanceof PGPMarker)
		{
			PGPMarker p = (PGPMarker)o;
			root.add(extractStructure(p));
		}
		else
			root.add(new DefaultMutableTreeNode("object of unknown type: " + o));
	}

	private static MutableTreeNode extractStructure(PGPMarker p)
	{
		DefaultMutableTreeNode node = new DefaultMutableTreeNode("PGPMarker");
		return node;
	}

	private static MutableTreeNode extractStructure(PGPOnePassSignatureList p)
	{
		DefaultMutableTreeNode node = new DefaultMutableTreeNode("PGPOnePassSignatureList");
		
		for (int i = 0; i < p.size(); i++)
		{
			PGPOnePassSignature sig = p.get(i);
			node.add(extractStructure(sig));
		}
		
		return node;
	}

	private static MutableTreeNode extractStructure(PGPOnePassSignature sig)
	{
		return new DefaultMutableTreeNode(sig);
	}

	private static MutableTreeNode extractStructure(PGPEncryptedDataList p)
	{
		DefaultMutableTreeNode node = new DefaultMutableTreeNode("PGPEncryptedDataList");
		
		Iterator iter = p.getEncryptedDataObjects();
		while(iter.hasNext())
		{
			Object encObj = iter.next();
			
			if (encObj instanceof PGPPBEEncryptedData)
			{
				node.add(new DefaultMutableTreeNode("Content of type PGPPBEEncryptedData: " + encObj));				
			}
			else if (encObj instanceof PGPPublicKeyEncryptedData)
			{
				DefaultMutableTreeNode subNode = new DefaultMutableTreeNode("Content of type PGPPublicKeyEncryptedData: " + encObj);
				node.add(subNode);
				
				PGPPublicKeyEncryptedData pgpEncObj = (PGPPublicKeyEncryptedData)encObj;
				subNode.add(new DefaultMutableTreeNode("KeyID: " + pgpEncObj.getKeyID()));
			}
			else
				node.add(new DefaultMutableTreeNode("Unexpected content of type: " + encObj));
		}
		
		return node;
	}

	private static MutableTreeNode extractStructure(PGPLiteralData p)
	{
		DefaultMutableTreeNode node = new DefaultMutableTreeNode("PGPLiteralData");
		
		try
		{
			BufferedReader reader = new BufferedReader(new InputStreamReader(p.getDataStream(), "UTF8"));
			StringBuffer clearTextBuffer = new StringBuffer();

			String line;
			while((line = reader.readLine()) != null)
				clearTextBuffer.append(line + "\n");

			node.add(new DefaultMutableTreeNode(clearTextBuffer.toString()));
		}
		catch (IOException e)
		{
			node.add(new DefaultMutableTreeNode(e.getMessage()));
			e.printStackTrace();
		}
		
		return node;
	}

	private static MutableTreeNode extractStructure(PGPCompressedData p)
	{
		DefaultMutableTreeNode node = new DefaultMutableTreeNode("PGPCompressedData");
		
		try
		{
			PGPObjectFactory fac = new PGPObjectFactory(p.getDataStream());
			
			Object o;
			try
			{
				while ((o = fac.nextObject()) != null)
				{
					extractStructure(node, o);
				}
			}
			catch (IOException ex)
			{
				ex.printStackTrace();
			}
			catch (Exception ex)
			{
				node.add(new DefaultMutableTreeNode(ex.getMessage()));
				ex.printStackTrace();
			}
			
		}
		catch (PGPException ex)
		{
			node.add(new DefaultMutableTreeNode(ex.getMessage()));
			ex.printStackTrace();
		}
		
		
		return node;
	}

	private static MutableTreeNode extractStructure(PGPPublicKeyRing p)
	{
		DefaultMutableTreeNode node = new DefaultMutableTreeNode("PGPPublicKeyRing");
		
		Iterator iter = p.getPublicKeys();
		while(iter.hasNext())
		{
			node.add(extractStructure((PGPPublicKey)iter.next()));
		}
		
		return node;
	}

	private static MutableTreeNode extractStructure(PGPSecretKeyRing p)
	{
		DefaultMutableTreeNode node = new DefaultMutableTreeNode("PGPSecretKeyRing");
		
		Iterator<PGPSecretKey> iter = p.getSecretKeys();
		while(iter.hasNext())
		{
			node.add(extractStructure(iter.next()));
		}
		
		return node;
	}

	private static MutableTreeNode extractStructure(PGPPublicKey key)
	{
		DefaultMutableTreeNode node = new DefaultMutableTreeNode((key.isMasterKey() ? "MASTER -- " : "") + Long.toHexString(key.getKeyID()).toUpperCase() + "(ID: "+key.getKeyID()+")" + " created: " + key.getCreationTime() + (key.isRevoked() ? " REVOKED on " + ((PGPSignature)key.getSignaturesOfType((key.isMasterKey() ? SignatureType.KEY_REVOCATION.getId() : SignatureType.SUBKEY_REVOCATION.getId())).next()).getCreationTime() : "") + " Fingerprint: 0x" + new String(Hex.encodeHex(key.getFingerprint())).toUpperCase() + " SHA1-Checksum: " + calcDigest(key) + " ");
		
		DefaultMutableTreeNode userIds = new DefaultMutableTreeNode("User-IDs");
		node.add(userIds);
		Iterator<String> userIdIter = key.getUserIDs();
		while(userIdIter.hasNext())
			userIds.add(new DefaultMutableTreeNode(userIdIter.next()));
		
		
		DefaultMutableTreeNode signatures = new DefaultMutableTreeNode("Signatures");
		node.add(signatures);
		
		Iterator<PGPSignature> signatureIter = key.getSignatures();
		while(signatureIter.hasNext())
			signatures.add(extractStructure(signatureIter.next()));
		
		DefaultMutableTreeNode userAttributes = new DefaultMutableTreeNode("User-Attributes");
		node.add(userAttributes);
		Iterator<PGPUserAttributeSubpacketVector> userAttributeIter = key.getUserAttributes();
		while(userAttributeIter.hasNext())
			userAttributes.add(extractStructure(userAttributeIter.next()));
		
		return node;
	}

	private static MutableTreeNode extractStructure(PGPUserAttributeSubpacketVector p)
	{
		DefaultMutableTreeNode node = new DefaultMutableTreeNode("PGPUserAttributeSubpacketVector " + p.getImageAttribute().getType());

		return node;
	}

	private static MutableTreeNode extractStructure(PGPSecretKey key)
	{
		DefaultMutableTreeNode node = new DefaultMutableTreeNode(" 0x" + Long.toHexString(key.getKeyID()));
		
		return node;
	}

	private static MutableTreeNode extractStructure(PGPSignatureList p)
	{
		DefaultMutableTreeNode node = new DefaultMutableTreeNode("PGPSignatureList, size=" + p.size());
		
		for (int i = 0; i < p.size(); i++)
		{
			PGPSignature sig = p.get(i);
			node.add(extractStructure(sig));
		}
		
		return node;
	}

	enum SignatureType {
		
		BINARY_DOCUMENT(0),
		CANONICAL_TEXT_DOCUMENT(1),
		STAND_ALONE(2),
		DEFAULT_CERTIFICATION(16),
		NO_CERTIFICATION(17),
		CASUAL_CERTIFICATION(18),
		POSITIVE_CERTIFICATION(19),
		SUBKEY_BINDING(24),
		DIRECT_KEY(31),
		KEY_REVOCATION(32),
		SUBKEY_REVOCATION(40),
		CERTIFICATION_REVOCATION(48),
		TIMESTAMP(64);

		private final int representation;
		
		SignatureType(int representation) {
			this.representation = representation;
		}
		
		public int getId() {
			return representation;
		}
		
		public static SignatureType get(int id) {
			for (int i = 0; i < values().length; i++)
				if (values()[i].getId() == id)
					return values()[i];
			
			return null;
		}
		
	}
	
	private static MutableTreeNode extractStructure(PGPSignature sig)
	{
		return new DefaultMutableTreeNode(sig + " " + SignatureType.get(sig.getSignatureType()) + " by 0x" + Long.toHexString(sig.getKeyID()).toUpperCase());
	}
	
	public static String calcDigest(String text)
	{
		try {
			return calcDigest(text.getBytes("UTF-8"));
		} catch (Exception ex) {
			log.error("Error calculating digest for text: " + text, ex);
			return null;
		}
	}
	
	private static String calcDigest(PGPPublicKey key)
	{
		try {
			return calcDigest(key.getEncoded());
		} catch (Exception ex) {
			log.error("Error calculating digest for PGPPublicKey: " + key, ex);
			return null;
		}
	}
	
	public static String calcDigest(byte[] data) throws NoSuchAlgorithmException
	{
		MessageDigest md = MessageDigest.getInstance("SHA1");
		String digest = new String(Hex.encodeHex(md.digest(data))).toUpperCase();
		return digest;
	}
	
}
