/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.crypto;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.SecureRandom;
import java.util.Date;
import java.util.Iterator;

import javax.annotation.PostConstruct;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.bcpg.ArmoredOutputStream;
import org.bouncycastle.openpgp.PGPCompressedDataGenerator;
import org.bouncycastle.openpgp.PGPEncryptedDataGenerator;
import org.bouncycastle.openpgp.PGPException;
import org.bouncycastle.openpgp.PGPLiteralData;
import org.bouncycastle.openpgp.PGPLiteralDataGenerator;
import org.bouncycastle.openpgp.PGPPublicKey;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPPublicKeyRingCollection;
import org.bouncycastle.openpgp.PGPUtil;

import de.nachtlicht_media.epoint.redcent.exception.ConflictException;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageEncryptor;

public class BouncycastlemessageEncryptor implements MessageEncryptor
{
	private static final Log log = LogFactory.getLog(BouncycastlemessageEncryptor.class);
	
	private SecureRandom secRnd;
	
	public BouncycastlemessageEncryptor()
	{
	}

	@PostConstruct
	public void init() throws ConflictException
	{
		try
		{
			secRnd = SecureRandom.getInstance("SHA1PRNG", "SUN");
			
			if (new File("/dev/urandom").exists())
			{
				try
				{
					log.trace("Salting secure random from /dev/urandom");
					byte[] salt = new byte[8192];
					new FileInputStream("/dev/urandom").read(salt);
					secRnd.setSeed(salt);
					for (int i = 0; i < salt.length; i++) salt[i]=0;
					salt = null;
					log.trace("Salting secure random done.");
				}
				catch (Exception ex)
				{
					log.error("Salting secure random failed: " + ex.getMessage(), ex);
				}
			}
		}
		catch (NoSuchAlgorithmException e)
		{
			throw new ConflictException("Could not initialize secure random generator: no such algorithm.", e);
		}
		catch (NoSuchProviderException e)
		{
			throw new ConflictException("Could not initialize secure random generator: no such provider.", e);
		}
	}

	public byte[] encryptMessage(String title, byte[] message, byte[] publicKeyData, long[] encryptionIds, boolean armor) throws ConflictException
	{
		PGPPublicKeyRingCollection publicKeys;
		
		try
		{
			publicKeys = unwrapPublicKeys(publicKeyData);
		}
		catch(PGPException pgpEx)
		{
			throw new ConflictException("Error reading public keyring from storage.", pgpEx);
		}
		catch(IOException ioEx)
		{
			throw new ConflictException("IO Error while public keyring data.", ioEx);
		}
		
		try
		{
			return encryptData(title, message, publicKeys, encryptionIds, armor);
		}
		catch(PGPException pgpEx)
		{
			throw new ConflictException("PGP error creating encrypted message.", pgpEx);
		}
		catch(NoSuchProviderException provEx)
		{
			throw new ConflictException("Provider error creating encrypted message.", provEx);
		}
		catch(IOException ioEx)
		{
			throw new ConflictException("IO Error while encrypting message.", ioEx);
		}
	}

	private PGPPublicKeyRingCollection unwrapPublicKeys(byte[] publicKeyData) throws IOException, PGPException
	{
		//TODO: cache unwrapping
		return new PGPPublicKeyRingCollection(PGPUtil.getDecoderStream(new ByteArrayInputStream(publicKeyData)));
	}

	private byte[] encryptData(String title, byte[] message, PGPPublicKeyRingCollection publicKeys, long[] encryptionIds, boolean armor) throws IOException, PGPException, NoSuchProviderException
	{
		ByteArrayOutputStream data = new ByteArrayOutputStream();
		
		OutputStream out;
		if (armor)
			out = new ArmoredOutputStream(data);
		else
			out = data;

		ByteArrayOutputStream bOut = new ByteArrayOutputStream();

		log.trace("Compressing data...");

		// get the data from the original file 
		PGPCompressedDataGenerator comData = new PGPCompressedDataGenerator(PGPCompressedDataGenerator.ZIP);
		PGPLiteralDataGenerator lData = new PGPLiteralDataGenerator();
		OutputStream            pOut = lData.open(comData.open(bOut), PGPLiteralData.BINARY, title, message.length, new Date());
		InputStream 			in = new ByteArrayInputStream(message);
		byte[]                  buf = new byte[4096];
		int                     len;
		
		while ((len = in.read(buf)) > 0)
		{
		    pOut.write(buf, 0, len);
		}
		
		lData.close();
		in.close();
		comData.close();

		log.trace("... compression done.");

		log.trace("Encrypting data using PGPEncryptedDataGenerator...");

		// object that encrypts the data
		PGPEncryptedDataGenerator cPk = new PGPEncryptedDataGenerator(PGPEncryptedDataGenerator.AES_128, true, secRnd, "BC");

		Iterator iter = publicKeys.getKeyRings();
		while(iter.hasNext())
		{
			PGPPublicKeyRing ring = ((PGPPublicKeyRing)iter.next());

			Iterator subIter = ring.getPublicKeys();
			while (subIter.hasNext())
			{
				PGPPublicKey key = (PGPPublicKey) subIter.next();

				if (BouncycastleUtil.keyIsSuitableForEncryption(key) && (encryptionIds == null || containsKeyId(encryptionIds, key.getKeyID())))
					cPk.addMethod(key);
			}
		}

		log.trace("... encryption done.");

		// take the outputstream of the original file and turn it into a byte array
		byte[] bytes = bOut.toByteArray();

		// write the plain text bytes to the armored outputstream
		OutputStream cOut = cPk.open(out, bytes.length);
		cOut.write(bytes);

		// cOut.close();
		cPk.close();
		out.close();
		
		data.close();
		return data.toByteArray();
	}

	/**
	 * @param encryptionIds
	 * @param keyID
	 * @return
	 */
	private boolean containsKeyId(long[] encryptionIds, long keyID)
	{
		for (int i = 0; i < encryptionIds.length; i++)
			if (keyID == encryptionIds[i])
				return true;
		
		return false;
	}
}
