/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.crypto;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;

import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.openpgp.PGPException;
import org.bouncycastle.openpgp.PGPPublicKey;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPPublicKeyRingCollection;
import org.bouncycastle.openpgp.PGPUtil;

public class HttpKeyServer
{
	private static final Log log = LogFactory.getLog(HttpKeyServer.class);
	
	private String hostName;
	private URL baseUrl;
	
	public HttpKeyServer(String hostName) throws MalformedURLException
	{
		this.hostName = hostName;
		baseUrl = new URL("http", hostName, 11371, "/pks/");
	}
	
	
	public PGPPublicKeyRingCollection getKeyRings(String userID, boolean matchPartial) throws PGPException, MalformedURLException, IOException
	{
		// URL to retrieve keys identified by userID
		URL getUrl = new URL(baseUrl, "lookup?op=get&search="+userID);

		PGPPublicKeyRingCollection col = retriveKeyRingsFromURL(getUrl);
		
		if (matchPartial)
			return col;
		else
		{
			// exact match required
			PGPPublicKeyRingCollection retCol = new PGPPublicKeyRingCollection(new ArrayList());
			
			Iterator iter = col.getKeyRings(userID, matchPartial);
			
			while(iter.hasNext())
				PGPPublicKeyRingCollection.addPublicKeyRing(retCol, (PGPPublicKeyRing)iter.next());
			
			return retCol;
		}
	}

	public PGPPublicKeyRing getPublicKeyRing(long keyID) throws PGPException, MalformedURLException, IOException
	{
		// URL to retrieve keys identified by userID
		URL getUrl = new URL(baseUrl, "lookup?op=get&search=0x"+Long.toHexString(keyID));
		
		// retrieve collection from server
		PGPPublicKeyRingCollection col = retriveKeyRingsFromURL(getUrl);
		
		// return the key identified by keyID
		return col.getPublicKeyRing(keyID);
	}
	
	protected PGPPublicKeyRingCollection retriveKeyRingsFromURL(URL getUrl) throws MalformedURLException, IOException, PGPException
	{
		long startTime = System.currentTimeMillis();
		HttpURLConnection con = (HttpURLConnection)getUrl.openConnection();
		log.trace("Connecting to " + con);
		
		con.connect();
		
		log.trace("Connecting to keyserver took: " + (System.currentTimeMillis()- startTime) + " ms.");
		
		if (con.getResponseCode() != HttpURLConnection.HTTP_OK)
		{
			log.warn("Server " + getUrl + " did not return 200");
			return null;
		}
		else
		{
			startTime = System.currentTimeMillis();
			BufferedReader result = new BufferedReader(new InputStreamReader(con.getInputStream()));
			
			String data = "";
			String line;
			
			while((line = result.readLine()) != null)
				data += line + "\n";
			
			log.trace("Retrieving data took: " + (System.currentTimeMillis()- startTime) + " ms.");

			int startIndex = -1;
			if ((startIndex = data.indexOf("-----BEGIN PGP PUBLIC KEY BLOCK")) != -1)
			{
				data = data.substring(startIndex);
				return readKeysFromInputStream(new ByteArrayInputStream(data.getBytes()));
			}
			else
			{
				log.warn("No valid keydata found in response from " + getUrl + ": " + data);
				return null;
			}
		}
	}
	
	private PGPPublicKeyRingCollection readKeysFromInputStream(InputStream in) throws IOException, PGPException
	{
		InputStream decIn = PGPUtil.getDecoderStream(in);
		PGPPublicKeyRingCollection pCol = new PGPPublicKeyRingCollection(decIn);
		
		if(log.isTraceEnabled())
		{
			log.trace("Found " + pCol.size() + " keys.");
			for (Iterator iter = pCol.getKeyRings(); iter.hasNext(); )
			{
				PGPPublicKeyRing pkr = (PGPPublicKeyRing)iter.next();

				for (Iterator keyIter = pkr.getPublicKeys(); keyIter.hasNext();)
				{
					PGPPublicKey key = (PGPPublicKey)keyIter.next();

					log.trace("----");
					log.trace(key.getKeyID() + " : " + new String(Hex.encodeHex(key.getFingerprint())) + " : " + key.getBitStrength() + " : " + key.isRevoked());
				}
			}
		}
		return pCol;
	}
}
