/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.interfaces;

import de.nachtlicht_media.epoint.redcent.exception.InvalidPGPMessageException;

public interface MessageDecryptionVerificationHandler
{
	/**
	 * Set the message to handle. Invalidates all cached states like isValid(). @see isValid()
	 * Calling this method must be allowed at most one time by the implementation.
	 * @param message The message to handle. Set message first, then invoke analyse(), ...
	 */
	public void setMessage(byte[] message);
	
	/**
	 * @return Returns the message that was analysed.
	 */
	public byte[] getMessage();
	
	/** 
	 * Perform message analysis. 
	 */
	public void analyse() throws InvalidPGPMessageException;
	
	/**
	 * @return Returns <code>true</code> if the message is a valid PGP Message either signed or encrypted
	 * and <code>false</code> if not.
	 */
	public Boolean isValid();
	
	/**
	 * @return Returns <code>true</code> if the message is valid, signed and not encrypted.
	 */
	public Boolean isClearText();
	
	/**
	 * @return Returns <code>true</code> if the message is encrypted and <code>false</code> if not.
	 */
	public Boolean isEncrypted();
	
	/**
	 * @return Returns messages signer's Fingerprint or <code>null</code>
	 * if message is not signed.
	 */
	public String getSignersFingerprint();

	/**
	 * @return Returns messages signer's masterkey's Fingerprint or <code>null</code>
	 * if message is not signed.
	 */
	public String getSignersPrimaryKeyFingerprint();

	/**
	 * @return The signer's public key in an ascii armored form.
	 */
	public String getSignersPublicKey() throws Exception;
	
	/**
	 * @return Returns <code>true</code> if the key used for signing is registered
	 * with notary permissions.
	 */
	public Boolean signerIsNotary();	
	
	/**
	 * @return Returns <code>true</code> if the key used for signing is registered
	 * with AI permissions.
	 */
	public Boolean signerIsAuthorizedIssuer();	
	
	/**
	 * @return Returns <code>true</code> if the key used for signing is registered
	 * in an extra keystore containing keys allowing special permissions and specified 
	 * in the issuers config to be the master key
	 */
	public Boolean signerIsIssuersMasterKey();
	
	/**
	 * @return Returns the public key fingerprint, the message was encrypted to if
	 * the message was encrypted and <code>null</code> if the message was not encrypted.
	 */
	public String getEncryptedTo();
	
	/**
	 * @return Returns message's cleartext if the message is encrypted or signed.
	 */
	public String getClearText();
	
	
	/**
	 * Method to add a public key to verify signatures.
	 * @param asciiArmoredPublicKey
	 * @throws InvalidPGPMessageException
	 */
	public void addTemporaryPublicKey(String asciiArmoredPublicKey) throws Exception;
}
