/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.messaging;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;

import de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore;

/**
 * @author js
 *
 */
public class DefaultDocumentStore implements DocumentStore
{
	private String baseLocation;
	private File baseDir;
	
	public DefaultDocumentStore(String baseLocation)
	{
		this.baseLocation = baseLocation;
	}

	/**
	 * @see de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore#getDocument(java.lang.String, java.lang.String)
	 */
	public synchronized byte[] getDocument(String realm, String documentId) throws Exception
	{
		if (baseDir == null)
			throw new Exception("Store is uninitialized.");

		checkPathName(realm);
		checkPathName(documentId);
		
		File documentDir = new File(baseDir, realm);
		
		if (!documentDir.exists())
			return null;
		else
		{
			File document = new File(documentDir, documentId);
			return readDocumentFromStorage(document);
		}
	}

	/**
	 * @param documentId
	 * @throws Exception 
	 */
	private void checkPathName(String id) throws Exception
	{
		if (id.indexOf(File.pathSeparatorChar) != -1)
			throw new Exception("Document realm or id may not contain pathSeparator!");
	}



	/**
	 * @param document
	 * @return
	 * @throws IOException
	 */
	private byte[] readDocumentFromStorage(File document) throws IOException
	{
		ByteArrayOutputStream out = new ByteArrayOutputStream();
		FileInputStream in = new FileInputStream(document);
		
		int b = -1;
		while ((b = in.read()) != -1)
			out.write(b);
		
		in.close();
		out.flush();
		out.close();
		
		return out.toByteArray();
	}


	/**
	 * @see de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore#storeOrUpdateDocument(java.lang.String, java.lang.String)
	 */
	public synchronized void storeOrUpdateDocument(String realm, String documentId, byte[] data)
			throws Exception
	{
		if (baseDir == null)
			throw new Exception("Store is uninitialized.");
		
		checkPathName(realm);
		checkPathName(documentId);

		File documentDir = new File(baseDir, realm);
		
		if (!documentDir.exists())
			documentDir.mkdirs();
		
		File document = new File(documentDir, documentId);
		storeDocumentToStorage(document, data);
	}


	/**
	 * @param document
	 * @param data
	 * @throws IOException 
	 */
	private void storeDocumentToStorage(File document, byte[] data) throws IOException
	{
		FileOutputStream out = new FileOutputStream(document);
		out.write(data);
		out.flush();
		out.close();
	}



	/**
	 * @see de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore#init()
	 */
	public void init() throws Exception
	{
		baseDir = new File(baseLocation);
		if (!baseDir.exists())
		{
			baseDir.mkdirs();
		}
	}

	/* (non-Javadoc)
	 * @see de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore#getMostRecentDocumentId(java.lang.String)
	 */
	public String getMostRecentDocumentId(String realm) throws Exception
	{
		if (baseDir == null)
			throw new Exception("Store is uninitialized.");

		checkPathName(realm);
		
		File documentDir = new File(baseDir, realm);
		
		if (!documentDir.exists())
			return null;
		else
		{
			File[] files = documentDir.listFiles();
			
			File mostRecent = null;
			for (int i = 0; i < files.length; i++)
			{
				if (mostRecent == null)
					mostRecent = files[i];
				else
				{
					if (files[i].lastModified() > mostRecent.lastModified())
						mostRecent = files[i];
				}
			}
			
			if (mostRecent != null)
				return mostRecent.getName();
			
			return null;
		}
	}

}
