/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.messaging;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.net.URI;
import java.net.URISyntaxException;
import java.text.DateFormat;
import java.text.Format;
import java.text.MessageFormat;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;
import java.util.TimeZone;
import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.nachtlicht_media.epoint.redcent.EpointService;
import de.nachtlicht_media.epoint.redcent.exception.ConflictException;
import de.nachtlicht_media.epoint.redcent.exception.MalformedRequestException;
import de.nachtlicht_media.epoint.redcent.interfaces.EpointCertificateTypes;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageHandler;
import de.nachtlicht_media.epoint.redcent.util.SimpleTemplateParser;
import de.nachtlicht_media.epoint.redcent.util.StringUtils;

public class DefaultMessageHandler implements MessageHandler
{
	private static final String DOCUMENT_TEMPLATE_HEADER = "Document-template: ";

	private static final String CONTENT_TYPE_HEADER = "Content-type: ";

	private static final Log log = LogFactory.getLog(DefaultMessageHandler.class);
	
	// Map with content type per certificate type
	public static final Map<String, String> contentTypes = new HashMap<String, String>();

	static {
		contentTypes.put(EpointCertificateTypes.EPOINT_CERTIFICATE, "text/plain.epoint.certificate");
		contentTypes.put(EpointCertificateTypes.INVALIDATION_CERTIFICATE, "text/plain.epoint.invalidation");
		contentTypes.put(EpointCertificateTypes.OBLIGATION_CERTIFICATE, "text/plain.epoint.obligation");
		contentTypes.put(EpointCertificateTypes.LEDGER, "text/plain.epoint.ledger");
		contentTypes.put(EpointCertificateTypes.DRAFT, "text/plain.epoint.non-negotiable-draft");
		contentTypes.put(EpointCertificateTypes.SERVER_STATUS, "text/plain.epoint.server-status");
		contentTypes.put(EpointCertificateTypes.SERVER_BALANCE, "text/plain.epoint.server-balance");
		contentTypes.put(EpointCertificateTypes.SERVER_SERVICE_CONTRACT, "text/plain.epoint.issuer-service-contract");
		contentTypes.put(EpointCertificateTypes.EXCHANGE_RATE_NOTIFICATION, "text/plain.epoint.rates");
		contentTypes.put(EpointCertificateTypes.EXCHANGE_RATE_INFORMATION, "text/plain.epoint.ratesinfo");
	}

	EpointService coreApp;
	public DefaultMessageHandler(EpointService coreApp)
	{
		this.coreApp = coreApp;
	}
	
	private class ParsedTemplate
	{
		String template;
		Map variables;
		public ParsedTemplate(String template, Map variables)
		{
			this.template = template;
			this.variables = variables;
		}
	}
	private Map parsedTemplateCache = new HashMap();
	/* (non-Javadoc)
	 * @see de.nachtlicht_media.epoint.redcent.messaging.MessageHandler#createMessage(java.lang.String, java.util.Map)
	 */
	public String createMessage(String message_type, StringDataMap data) throws ConflictException
	{
		// Possible race condition: templates may be parsed two times if the first two transactions
		// for the same message type are executed in concurrent threads. Since the result will be the same,
		// using synchronization to prevent double parsing will decrease performance.
		if (!parsedTemplateCache.containsKey(message_type))
		{
			String template = coreApp.getTemplateManager().getTemplate(message_type);
			SimpleTemplateParser parser = new SimpleTemplateParser(template);

			Map variables = parser.parseVariableNames();
			if (data.size() < variables.size())
				log.warn("The message template " + template + " ("+variables.size()+") does not match the available data ("+data.size()+")" + data);
			
			parsedTemplateCache.put(message_type, new ParsedTemplate(template, variables));
		}
		
		ParsedTemplate cachedTemplate = (ParsedTemplate)parsedTemplateCache.get(message_type);
		
		// create message from template by using variables and data
		String message = formatCertificateMessage(cachedTemplate.template, cachedTemplate.variables, data);
		
		message = addMessageHeader(message_type, message);
		
		return message;
	}
	
	public String addMessageHeader(String certificateType, String message) throws ConflictException
	{
		log.trace("Adding header for type " + certificateType);
		
		String header = null;
		
		header = CONTENT_TYPE_HEADER + DefaultMessageHandler.contentTypes.get(certificateType) + "; charset=utf-8" + "\n";
		header += DOCUMENT_TEMPLATE_HEADER + coreApp.getIssuerURL() + "/template?ID=" + coreApp.getTemplateManager().getHashValueForCurrentVersion(certificateType) + "\n";
		
		// Insert blank line to terminate header section
		header += "\n";
		
		return header + message;
	}
	
	private Map messageFormatCache = new HashMap();
	private String formatCertificateMessage(final String template, final Map variables, final Map data)
	{
		log.trace("Formatting message: " + template);
		
		// create a representation from the messageTemplate that is suitable to 
		// be processed by java.text.MessageFormat
		String messageTemplate = getMessageFormatRepresentation(template, variables);

		log.trace("Created messageFormat representation: " + messageTemplate);
		
		// keys in the order they appear in the document
		Vector variableNames = new Vector(variables.keySet());

		// produce the corresponding data array to be passed to the MessageFormat class
		Object[] formatData = new Object[variableNames.size()];
		
		Iterator iter = variableNames.iterator();
		for (int i = 0; iter.hasNext(); i++)
		{
			String key = (String)iter.next();
			
			if(key.startsWith("{") && key.endsWith("}"))
				key = key.substring(1, key.length() - 1);
				
			formatData[i] = data.get(key);
			
			if (formatData[i] == null)
				coreApp.getEnvVar(key);
		}
		
		// finally create the message formatter and create the 
		// message by merging template and data.
		MessageFormat messageFormat = new MessageFormat(messageTemplate);
		messageFormat.setLocale(Locale.US);
		
		// Override all DateFormats to be in UTC time and formatted according to standard
		Format[] formats = messageFormat.getFormatsByArgumentIndex();
		
		SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss z");
		formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
		
		for (int i = 0; i < formats.length; i++)
		{
			if (formats[i] instanceof DateFormat || formatData[i] instanceof Date)
			{
				messageFormat.setFormat(i, formatter);
			}
			
			if (formats[i] instanceof NumberFormat)
			{
				NumberFormat nf = ((NumberFormat)formats[i]);
				nf.setGroupingUsed(false);
				messageFormat.setFormat(i, nf);
			}
		}
		
		if (log.isTraceEnabled())
			log.trace("Creating Message with data: " + Arrays.asList(formatData) + " and Formats: " + Arrays.asList(messageFormat.getFormatsByArgumentIndex()));
		
		// fill the template with the data given
		String resultingMessage = messageFormat.format(formatData);
		
		log.trace("Resulting Message: '" + resultingMessage +"'");
		
		return resultingMessage;
	}

	public Map parseMessage(String incomingTextMessage) throws ConflictException
	{
		return parseMessage(incomingTextMessage, null);
	}

	/* (non-Javadoc)
	 * @see de.nachtlicht_media.epoint.redcent.messaging.MessageHandler#parseMessage(java.lang.String)
	 */
	public StringDataMap parseMessage(String incomingTextMessage, String forceTemplateId) throws ConflictException
	{
		try
		{
			log.trace("Parsing incoming message text: " + incomingTextMessage);

			EpointMessageFrame epMsg = parseEpointMessageHeaders(incomingTextMessage);
			
			if (epMsg.templateURI == null)
				throw new ConflictException("Message header does not contain 'Document-template:' section.");
						
			String templateID = epMsg.templateURI.toString().substring(epMsg.templateURI.toString().lastIndexOf("template?ID=") + "template?ID=".length()).trim().toUpperCase();
			
			if (forceTemplateId != null && !forceTemplateId.equals(templateID))
				throw new MalformedRequestException("The submitted message was created from template '"+templateID+"' but application requirements enforce creation from template '"+forceTemplateId+"'.");
			
			log.trace("Searching for template by id: " + templateID);
			
			//TODO: implement caching for raw templates analogous to (or combined with) the cache used in createMessage() 
			String rawTemplate = coreApp.getTemplateManager().getTemplate(templateID);
			
			log.trace("Using raw template:\n'" + rawTemplate + "'");
			
			if (rawTemplate == null)
				throw new ConflictException("Unable to locate template for ID " + templateID + ". Available IDs are: " + coreApp.getTemplateManager().getTemplateIdentifiers());
			
			SimpleTemplateParser parser = new SimpleTemplateParser(rawTemplate);
			Map variables = parser.parseVariableNames();
			Vector keys = new Vector(variables.keySet());
			
			String messageTemplate = getMessageFormatRepresentation(rawTemplate, variables);

			log.trace("Created messageTemplate:\n'" + messageTemplate + "'");
			
			MessageFormat messageFormat = new MessageFormat(messageTemplate);

			// Override all DateFormats to be in UTC time and formatted according to standard
			Format[] formats = messageFormat.getFormatsByArgumentIndex();
			
			SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss z");
			formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
			
			for (int i = 0; i < formats.length; i++)
			{
				if (formats[i] instanceof DateFormat)
				{
					messageFormat.setFormat(i, formatter);
				}
				
				if (formats[i] instanceof NumberFormat)
				{
					NumberFormat nf = ((NumberFormat)formats[i]);
					nf.setGroupingUsed(false);
					messageFormat.setFormat(i, nf);
				}
			}
			
			log.trace("Parsing body:\n'" + epMsg.body + "' with Formats: " + Arrays.asList(messageFormat.getFormatsByArgumentIndex()));
			Object data[] = messageFormat.parse(epMsg.body);
			
			StringDataMap returns = new StringDataMap();
			for (int i = 0; i < data.length; i++)
				returns.put(keys.get(i), data[i]);
			
			return returns;
		}
		catch(Exception ex)
		{
			throw new ConflictException(ex);
		}
	}

	public static class EpointMessageFrame
	{
		URI templateURI;
		String contentType;
		String charset;
		
		String header;
		String body;
		/**
		 * @return the templateURI
		 */
		public URI getTemplateURI() {
			return templateURI;
		}
		/**
		 * @return the contentType
		 */
		public String getContentType() {
			return contentType;
		}
		/**
		 * @return the charset
		 */
		public String getCharset() {
			return charset;
		}
		/**
		 * @return the header
		 */
		public String getHeader() {
			return header;
		}
		/**
		 * @return the body
		 */
		public String getBody() {
			return body;
		}
	}
	
	public static EpointMessageFrame parseEpointMessageHeaders(String incomingMessage) throws URISyntaxException, IOException
	{
		EpointMessageFrame msg = new EpointMessageFrame();
		
		BufferedReader lr = new BufferedReader(new StringReader(incomingMessage));
		
		String line ="";
		
		while(!"".equals(line = lr.readLine()) || line == null)
		{
			if (msg.header == null)
				msg.header = "";
			
			msg.header += line + "\n";
		}
		
		while((line = lr.readLine()) != null)
		{
			if (msg.body == null)
				msg.body = "";
			
			msg.body += line + "\n";
		}
		
		// parse templateuri, content-type and charset
		lr = new BufferedReader(new StringReader(msg.header));
		
		line = null;
		while((line = lr.readLine()) != null)
		{
			line = line.trim();
			
			if (line.startsWith(CONTENT_TYPE_HEADER))
			{
				// remove line header
				line = line.substring(CONTENT_TYPE_HEADER.length()).trim();
					
				// if line contains char-set information
				if (line.indexOf(';') != -1)
					line = line.substring(0, line.indexOf(';')).trim();
				
				// by default charset is UTF-8 for all ePointMessages
				msg.charset = "UTF-8";
				
				// line contains now only type
				msg.contentType = line.trim();
			}
			else if(line.startsWith(DOCUMENT_TEMPLATE_HEADER))
			{
				// remove line header
				line = line.substring(DOCUMENT_TEMPLATE_HEADER.length()).trim();
				
				msg.templateURI = new URI(line);
			}
			else
				log.warn("Ignored unknown header line '"+ line +"'in message: \n" + incomingMessage);
		}
		
		return msg;
	}

	/**
	 * Produces a java.lang.MessageFormat representation from a template containing
	 * $VARIABLE style named variables
	 * @param template
	 * @param data
	 * @return
	 */
	private String getMessageFormatRepresentation(String template, Map variables)
	{
		Vector variableNames = new Vector(variables.keySet());
		Vector variableNamesSortedByLength = (Vector)variableNames.clone();
		
		// sort variableNames by length to allow simple text replacements later on.
		// long variables have to be replaced first to prevent parts of variables
		// to be replaced by other variable names beeing only a part of the variable's
		// name
		Collections.sort(variableNamesSortedByLength, new Comparator()
		{
			public int compare(Object o1, Object o2)
			{
				return String.valueOf(o2).length() - String.valueOf(o1).length();
			}
		});
		
		// replace variable names in message by their index in variableNames, start with long variables first
		// Format of the resulting message MUST be java.lang.MessageFormat compatible like
		// This is a {1} message.
		for (int k = 0; k < variableNamesSortedByLength.size(); k++)
		{
			SimpleTemplateParser.Variable var = (SimpleTemplateParser.Variable)variables.get(variableNamesSortedByLength.get(k));
			template = StringUtils.replace(template, "$"+variableNamesSortedByLength.get(k), ((!var.isBraced())?"{":"")+variableNames.indexOf(variableNamesSortedByLength.get(k))+((!var.isBraced())?"}":""));
		}
		
		return template;
	}
}
