/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.messaging;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

import javax.annotation.PostConstruct;

import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.annotation.Autowired;

import de.nachtlicht_media.epoint.redcent.EpointService;
import de.nachtlicht_media.epoint.redcent.crypto.BouncycastleUtil;
import de.nachtlicht_media.epoint.redcent.interfaces.TemplateNames;
import de.nachtlicht_media.epoint.redcent.interfaces.RedCentConfig;
import de.nachtlicht_media.epoint.redcent.interfaces.TemplateManager;
import de.nachtlicht_media.epoint.redcent.util.StringUtils;

public class DefaultTemplateManager implements TemplateManager
{
	private static final Log log = LogFactory.getLog(DefaultTemplateManager.class);
	
	// Map with file names per certificate type
	private static Map<String, String> templateFileNames = null;
	
	@Autowired 
	protected TemplateNames templateNames = null;
	
	protected Map<String, String> templateCache;
	protected Map<String, String> currentHashCache;
	
	protected EpointService issuer;
	
	@Autowired 
	protected RedCentConfig redcentConfig = null;
	
	/* (non-Javadoc)
	 * @see de.nachtlicht_media.epoint.redcent.messaging.TemplateManager#init()
	 */
//	@PostConstruct
	public synchronized void init(EpointService issuer) throws Exception
	{
		this.issuer = issuer;

		templateFileNames = templateNames.templateFileNames;
		
		if (templateCache != null)
			throw new Exception("TemplateManager is already initialized.");
		
		log.info("Initializing template manager with templates from directory: " + getTemplatePath());
		
		File baseDir = new File(getTemplatePath());
		
		// Check if baseDir exists
		if (!baseDir.isDirectory() || baseDir.list().length == 0)
			throw new Exception("Invalid base directory specified: " + baseDir);
		
		// Make sure a place to store all versions of certificate templates exists
		ensureArchiveStorageExists();
		
		// All locations exists, read archive to memory first
		templateCache = new LinkedHashMap<String, String>();
		currentHashCache = new LinkedHashMap<String, String>();
		
		readArchivedTemplatesToCache();
		
		// Read template files from directory and archive them if modified
		Iterator<String> iter = templateFileNames.keySet().iterator();
		while(iter.hasNext())
		{
			String templateId = iter.next();
			String templateName = templateFileNames.get(templateId);
			String template = readTemplate(new FileInputStream(new File(baseDir, templateName)));
			String digest = calcDigest(template);
			currentHashCache.put(templateId, digest);
			
			// If the template was modified or an archived version does not exists, 
			// create an archived version now
			if (!templateCache.containsKey(digest))
			{
				// TODO: check templates on reparseability
				persistTemplateVersion(digest, template);
				templateCache.put(digest, template);
			}
			
			templateCache.put(templateId, template);
		}
	}

	protected void readArchivedTemplatesToCache() throws IOException, FileNotFoundException, NoSuchAlgorithmException, UnsupportedEncodingException, Exception
	{
		File archiveDir = getArchiveDir();
		
		File[] archiveFiles = archiveDir.listFiles();
		for (int i = 0; i < archiveFiles.length; i++)
		{
			// Read the file and compare the hash-value with the content's hash value. If they match,
			// add the template to the store.
			String hashValue = archiveFiles[i].getName();
			String template = readTemplate(new FileInputStream(archiveFiles[i]));
			
			String digest = calcDigest(template);
			
			if (hashValue.equalsIgnoreCase(digest))
				templateCache.put(digest, template);
			else
				throw new Exception("Ignored archived template file " + archiveFiles[i] + ": Filename does not match content's hash value '"+digest+"'");
		}
	}

	private File getArchiveDir()
	{
		return new File(new File(getTemplatePath()), "archive");
	}

	protected void ensureArchiveStorageExists() throws Exception
	{
		File archiveDir = getArchiveDir();
		// If archive dire does not exist, create it
		if (!archiveDir.exists())
		{
			if(archiveDir.mkdir())
				log.info("Created archive directory: " + archiveDir.getAbsolutePath());
			else
				throw new Exception("Could not create archive directory: " + archiveDir.getAbsolutePath());
		}
	}

	protected void persistTemplateVersion(String digest, String template) throws Exception
	{
		OutputStreamWriter outW = new OutputStreamWriter(new FileOutputStream(new File(getArchiveDir(), digest)), "UTF-8");
		outW.write(template);
		outW.flush();
		outW.close();
	}

	public String calcDigest(String template) throws NoSuchAlgorithmException, UnsupportedEncodingException
	{
		return BouncycastleUtil.calcDigest(template.getBytes("UTF-8"));
	}
	
	/**
	 * @param in InputStream to read template from. It is assumed that the template data is
	 * UTF-8 encoded.
	 * @return The template in a generic Java String, line ends trimmed and tabs replaced by 4 spaces.
	 * @throws IOException
	 */
	public String readTemplate(InputStream in) throws IOException
	{
		BufferedReader reader = new BufferedReader(new InputStreamReader(in,"UTF-8"));

		String message = "";
		String line = null;
		while((line = reader.readLine()) != null)
		{
			line = StringUtils.trimLineEnd(line);
			line = StringUtils.replace(line, "\t", "    ");
			
			message += line + "\n";
		}
		return message;
	}

	/* (non-Javadoc)
	 * @see de.nachtlicht_media.epoint.redcent.messaging.TemplateManager#getTemplate(java.lang.String)
	 */
	public String getTemplate(String certificateType)
	{
		return (String)templateCache.get(certificateType);
	}

	/* (non-Javadoc)
	 * @see de.nachtlicht_media.epoint.redcent.messaging.TemplateManager#getTemplateIdentifiers()
	 */
	public Collection<String> getTemplateIdentifiers()
	{
		return Collections.unmodifiableCollection(templateCache.keySet());
	}

	/* (non-Javadoc)
	 * @see de.nachtlicht_media.epoint.redcent.messaging.TemplateManager#getHashValueForCurrentVersion(java.lang.String)
	 */
	public String getHashValueForCurrentVersion(String certificateType)
	{
		return (String)currentHashCache.get(certificateType);
	}
	
	/**
	 * @return The location of the template directory
	 */
	protected String getTemplatePath() {
		return redcentConfig.getTemplateRealPath();
	}
}
