/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.servlet;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.net.URLDecoder;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Map;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.commons.codec.net.URLCodec;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.web.context.WebApplicationContext;
import org.springframework.web.context.support.WebApplicationContextUtils;

import de.nachtlicht_media.epoint.redcent.EpointRecord;
import de.nachtlicht_media.epoint.redcent.RedCent;
import de.nachtlicht_media.epoint.redcent.exception.BadsignatureException;
import de.nachtlicht_media.epoint.redcent.exception.ConflictException;
import de.nachtlicht_media.epoint.redcent.exception.MalformedRequestException;
import de.nachtlicht_media.epoint.redcent.exception.NotFoundException;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageDecryptionVerificationHandler;
import de.nachtlicht_media.epoint.redcent.servlet.crypto.ServletMessageDecryptionVerificationFactory;
import de.nachtlicht_media.epoint.redcent.servlet.exception.DefaultProcessingExceptionMapper;
import de.nachtlicht_media.epoint.redcent.servlet.exception.ProcessingException;
import de.nachtlicht_media.epoint.redcent.servlet.exception.ProcessingExceptionMapper;
import de.nachtlicht_media.epoint.redcent.servlet.filter.SecureMessageRedirectionFilter;

/**
 * The RedCentServlet is currently used to provide the http(s) interface to the RedCent application. During
 * its initialization phase creates the RedCentServlet the necessary instance of the RedCent application.
 * @author js
 *
 */
public class RedCentServlet extends HttpServlet
{
	private static final Log log = LogFactory.getLog(RedCentServlet.class);

	protected RedCent issuer;
	
	public RedCentServlet()
	{
	}

	/**
	 * Init method for servlet. If initialization of issuer reference fails, a ServletException is thrown
	 * to prevent request processing until the servlet is re-deployed and the issuer is correctly initialized.
	 * 
	 * @see javax.servlet.GenericServlet#init(javax.servlet.ServletConfig)
	 */
	public void init(final ServletConfig config) throws ServletException
	{
		super.init(config);
		
		log.info("initializing RedCentServlet");
		
		WebApplicationContext ctx = 
			WebApplicationContextUtils.getRequiredWebApplicationContext(
				this.getServletContext());
		issuer = (RedCent) ctx.getBean("redcent");

		log.info("...done.");
	}

	/**
	 * Redirect GET-request to doPost to be processed. If a certain URL shall not be available via
	 * GET-requests, a servlet filter has to be used to redirect these requests before entering the
	 * servlet's code.
	 */
	public void doGet(HttpServletRequest arg0, HttpServletResponse arg1) throws ServletException, IOException
	{
		doPost(arg0, arg1);
	}
	
	/**
	 * Process POST-requests: wraps http requests to method calls to the issuer application.
	 * @see javax.servlet.http.HttpServlet#doPost(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
	 */
	protected void doPost(HttpServletRequest req, HttpServletResponse resp) throws ServletException, IOException
	{
		if (log.isTraceEnabled())
			log.trace("Processing request " + req.getRequestURI() + " with parameters " + parameterMapToString(req.getParameterMap()));
		
		if ("/action".equals(req.getServletPath()))
		{
			doAction(req, resp);
		}
		else if ("/emit".equals(req.getServletPath()))
		{
			doEmit(req, resp);
		}
		else if ("/invalidate".equals(req.getServletPath()))
		{
			doInvalidate(req, resp);
		}
		else if ("/balance".equals(req.getServletPath()))
		{
			doBalance(req, resp);
		}
		else if ("/info".equals(req.getServletPath()))
		{
			doInfo(req, resp);
		}
		else if ("/rates".equals(req.getServletPath()))
		{
			doRates(req, resp);
		}
		else if ("/updateAI".equals(req.getServletPath()))
		{
			doUpdateAI(req, resp);
		}
		else if (req.getServletPath().startsWith("/CERT") && req.getServletPath().endsWith(".asc"))
		{
			doCertRequest(req, resp);
		}
		else if (req.getServletPath().equals("/service"))
		{
			doService(resp);
		}
		else if ("/template".equals(req.getServletPath()))
		{
			doTemplate(req, resp);
		}
		// a bit ugly workaround because tomcat does not support regular expressions in
		// servlet mappings.
		else if ("/pubkey.asc".equals(req.getServletPath()) || "/pubkey".equals(req.getServletPath()))
		{
			doPubkey(resp);
		}
		else
		{
			try
			{
				super.doPost(req, resp);
			}
			catch (Exception ex)
			{
				ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
				respondWithProcessingException(resp, pEx);
			}
		}
	}

	private void doPubkey(HttpServletResponse resp) throws FileNotFoundException, IOException
	{
		resp.setContentType("text/plain");
		resp.setStatus(200);
		
		BufferedReader keyReader = new BufferedReader(new FileReader(getServletContext().getRealPath("/pubkey.asc")));
		
		String line = null;
		while((line = keyReader.readLine()) != null)
			resp.getWriter().println(line);
		
		keyReader.close();
	}

	private void doTemplate(HttpServletRequest req, HttpServletResponse resp) throws IOException
	{
		if (req.getParameter("ID") != null)
		{
			String id = req.getParameter("ID").toUpperCase();

			log.trace("Getting template for id " + id);

			try
			{
				String template = issuer.getTemplateManager().getTemplate(id);

				if (template != null)
				{
					resp.setContentType("text/plain; ; charset=utf-8");
					resp.setStatus(200);
					resp.getOutputStream().write(template.getBytes("UTF-8"));
				}
				else
					throw new NotFoundException("A template identified by hash value " + id + " could not be found.");
			}
			catch(Exception ex)
			{
				// Aquire a ProcessingException for the thrown exception to be able
				// to return the correct response status and short text
				ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
				respondWithProcessingException(resp, pEx);
			}
		}
		else
			respondWithProcessingException(resp, getProcessingExceptionMapper().getMappedException(new MalformedRequestException("No parameter ID specified.")));
	}

	private void doService(HttpServletResponse resp) throws IOException
	{
		try
		{
			resp.setContentType("text/plain; ; charset=utf-8");
			resp.setStatus(200);

			resp.getOutputStream().write(issuer.getDeployedServerServiceContract());
		}
		catch(Exception ex)
		{
			// Aquire a ProcessingException for the thrown exception to be able
			// to return the correct response status and short text
			ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
			respondWithProcessingException(resp, pEx);
		}
	}

	private void doCertRequest(HttpServletRequest req, HttpServletResponse resp) throws IOException
	{
		// Return certificate no matter whether it's valid or invalid
		String numString = req.getServletPath().trim();
		numString = numString.substring(5).trim();
		numString = numString.substring(0, numString.length()-4).trim();
		
		try
		{
			EpointRecord result = issuer.findBySerial(new Long(numString).longValue());
			resp.getOutputStream().write(result.getCert());
		}
		catch (Exception ex)
		{
			ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
			respondWithProcessingException(resp, pEx);
		}
	}

	private void doRates(HttpServletRequest req, HttpServletResponse resp) throws IOException
	{
		if (req.getParameterMap().isEmpty())
		{
			// return ExchangeRateInformation, public information
			try
			{
				byte[] exchangeRateInformation = issuer.getExchangeRateInformation();

				resp.setContentType("text/plain");
				resp.setStatus(200);
				resp.getOutputStream().write(exchangeRateInformation);
			}
			catch(Exception ex)
			{
				// Acquire a ProcessingException for the thrown exception to be able
				// to return the correct response status and short text
				ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
				respondWithProcessingException(resp, pEx);
			}
		}
		else
		{
			// other functions like retrieving a specific ExchangeRateNotification (XRN) or publishing of a 
			// a new XRN by the PI
			try
			{
				if (req.getParameterMap().get("ID") != null)
				{
					String xrnSerial = req.getParameter("ID").trim();
					
					byte[] encryptedExchangeRateNotificaion = issuer.getEncryptedExchangeRateNotification(Long.parseLong(xrnSerial));
					
					if (encryptedExchangeRateNotificaion != null)
					{
						resp.getOutputStream().write(encryptedExchangeRateNotificaion);
					}
					else
						throw new NotFoundException("Not found");
				}
				else if (req.getParameterMap().get("XRN") != null)
				{
					byte[] xrn = req.getParameter("XRN").getBytes("UTF8");
					resp.getOutputStream().write(issuer.publishExchangeRateNotification(xrn).getBytes("UTF8"));
				}
				else
					throw new MalformedRequestException("Invalid request: either no parameter or one of 'ID', 'XRN' has to be specified.");
			}
			catch(Exception ex)
			{
				ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
				respondWithProcessingException(resp, pEx);
			}
		}
		
//			debug(req, resp);
	}

	private void doInfo(HttpServletRequest req, HttpServletResponse resp) throws IOException
	{
		if (req.getParameterMap().get("ID") == null && req.getParameterMap().get("KEY") == null  && req.getParameterMap().get("CERT") == null)
		{
			try
			{
				byte[] signedServerStatus = issuer.getServerStatus();

				resp.setContentType("text/plain");
				resp.setStatus(200);
				resp.getOutputStream().write(signedServerStatus);
			}
			catch(Exception ex)
			{
				// Aquire a ProcessingException for the thrown exception to be able
				// to return the correct response status and short text
				ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
				respondWithProcessingException(resp, pEx);
			}
		}
		else
		{
			try
			{
				if (req.getParameterMap().get("ID") != null || req.getParameterMap().get("KEY") != null )
				{
					String id = req.getParameter("ID");
					
					if (id == null)
						id = req.getParameter("KEY");
					
					id = id.trim().toUpperCase().replaceAll(" ", "");
					
					if (req.getParameterMap().containsKey("TYPE")) {

						String type = req.getParameterMap().containsKey("TYPE") ? req.getParameter("TYPE") : null;

						if ("ledger".equalsIgnoreCase(type)) {

							if (!req.getParameterMap().containsKey("KEY") || req.getParameterMap().containsKey("ID"))
								throw new MalformedRequestException("Requesting ledger reports is allowed for obligations only.");

							String start = req.getParameterMap().containsKey("START") ? req.getParameter("START") : null;
							String end = req.getParameterMap().containsKey("END") ? req.getParameter("END") : null;

							Long startTx;
							Long endTx;

							try {
								startTx = start != null ? new Long(Long.parseLong(start)) : null;
							} catch (NumberFormatException nf) {
								throw new MalformedRequestException("Invalid request: cannot parse transaction boundary START " + start + " not a numeric value!");
							}
							try {
								endTx = end != null ? new Long(Long.parseLong(end)) : null;
							} catch (NumberFormatException nf) {
								throw new MalformedRequestException("Invalid request: cannot parse transaction boundary END " + end + " not a numeric value!");
							}

							byte[] ledger = issuer.getSignedLedger(id, startTx, endTx);

							if (ledger != null)
							{
								resp.setContentType("text/plain");
								resp.setStatus(200);
								resp.getOutputStream().write(ledger);
							}
							else
								throw new NotFoundException("No datasets found for the KEY parameter you submitted: " + id);
						} else
							throw new MalformedRequestException("Invalid request: unknow report type '"+type+"'");


					} else {
						EpointRecord rec = issuer.findValidRecordByID(id.toUpperCase());

						if (rec != null)
						{
							resp.setContentType("text/plain");
							resp.setStatus(200);
							resp.getOutputStream().write(rec.getCert());
						}
						else
							throw new NotFoundException("Not found");
					}
				}
				else if (req.getParameterMap().get("CERT") != null)
				{
					String cert = req.getParameter("CERT");
					cert = cert.trim().toUpperCase().replaceAll(" ", "");

					EpointRecord result = issuer.findBySerial(new Long(cert).longValue());

					resp.setContentType("text/plain");
					resp.setStatus(200);
					resp.getOutputStream().write(result.getCert());
				}
				else
					throw new MalformedRequestException("Invalid request: either no parameter or exactly one of 'ID' or 'KEY' OR 'CERT' has to be specified.");
			}
			catch(Exception ex)
			{
				ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
				respondWithProcessingException(resp, pEx);
			}
		}
		
//			debug(req, resp);
	}

	private void doBalance(HttpServletRequest req, HttpServletResponse resp) throws IOException
	{
		try
		{
			String key = req.getParameter("KEY");
			if (key != null)
				key = key.toUpperCase();
			
			byte[] signedServerBalance = issuer.getServerBalance(key);

			resp.setContentType("text/plain");
			resp.setStatus(200);
			resp.getOutputStream().write(signedServerBalance);
		}
		catch(Exception ex)
		{
			// Aquire a ProcessingException for the thrown exception to be able
			// to return the correct response status and short text
			ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
			respondWithProcessingException(resp, pEx);
		}
	}

	private void doInvalidate(HttpServletRequest req, HttpServletResponse resp) throws IOException
	{
		// invalidation request
		try
		{
			if (req.getParameterMap().isEmpty())
			{
				throw new MalformedRequestException("Illegal call without arguments");
			}
			else
			{
				// This attribute has to be set by SecureMessageRedirectionFilter that validates/decrypts and redirects messages
				MessageDecryptionVerificationHandler msg = null;
				
				if (req.getAttribute(SecureMessageRedirectionFilter.ATTR_SECURE_MESSAGE_HANDLER) != null)
				{
					msg = (MessageDecryptionVerificationHandler)req.getAttribute(SecureMessageRedirectionFilter.ATTR_SECURE_MESSAGE_HANDLER);

					if (!msg.isValid().booleanValue())
						throw new MalformedRequestException("Illegal call to /invalidate with non-valid signed message.");
					else if (msg.getSignersFingerprint() == null)
						throw new BadsignatureException ("Illegal call: cannot resolve signer.");
					else if (!msg.isClearText().booleanValue())
						throw new BadsignatureException ("Illegal call: signed requests to /invalidate must be clearsigned.");

					// msg is valid and signed, proceed
				}
				
				String rand = req.getParameter("B");
				String authorizedIssuer = req.getParameter("A");
				String targetIssuer = req.getParameter("K");
				String signature = msg != null ? new String(msg.getMessage()) : null;

				if (msg != null) {
					// just for fun: check if redirect parameters are the same as in the messageHandler
					// might differ when the ServletContainer messes sessions up.
					String decodedMessage = URLDecoder.decode(msg.getClearText(), "UTF-8");
					if(decodedMessage.indexOf(String.valueOf(rand)) == -1
					)
					{
						throw new MalformedRequestException("Fatal Error: MessageHandler and request parameters do not match.");
					}
				}
				
				// verify that the request was originally created for this issuer instance
				targetIssuer = checkTargetIssuer(targetIssuer);
				
				// get cert for rest, make targetDist valid with amount
				byte[] newCert = issuer.invalidate(rand, signature, authorizedIssuer, targetIssuer, extractCommentFromRequest(req));

				resp.setStatus(200);
				resp.getOutputStream().write(newCert);
			}
		}
		catch(Exception ex)
		{
			// Aquire a ProcessingException for the thrown exception to be able
			// to return the correct response status and short text
			ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
			respondWithProcessingException(resp, pEx);
		}
	}

	private void doEmit(HttpServletRequest req, HttpServletResponse resp) throws IOException
	{
		// emission request
		try
		{
			if (req.getParameterMap().isEmpty())
			{
				throw new MalformedRequestException("Illegal call without arguments");
			}
			else if (req.getAttribute(SecureMessageRedirectionFilter.ATTR_SECURE_MESSAGE_HANDLER) == null)
			{
				throw new MalformedRequestException("Illegal call without secure message handler.");
			}
			else
			{
				// This attribute has to be set by SecureMessageRedirectionFilter that validates/decrypts and redirects messages
				MessageDecryptionVerificationHandler msg = (MessageDecryptionVerificationHandler)req.getAttribute(SecureMessageRedirectionFilter.ATTR_SECURE_MESSAGE_HANDLER);
				
				if (!msg.isValid().booleanValue())
					throw new MalformedRequestException("Illegal call to /emit with non-valid message.");
				else if (msg.getSignersFingerprint() == null)
					throw new MalformedRequestException ("Illegal call: requests to /emit have to be signed.");
				else if (!Boolean.TRUE.equals(msg.isClearText()))
					throw new BadsignatureException ("Illegal call: requests to /emit must be clearsigned.");
				else if (!Boolean.TRUE.equals(msg.signerIsAuthorizedIssuer()))
					throw new BadsignatureException ("Illegal call: requests to /emit must be signed by an authorized issuer's key.");
				else
				{
					// msg is valid and signed
					String targetDigest = req.getParameter("D");
					
					if (targetDigest == null || targetDigest.trim().length() == 0)
						throw new MalformedRequestException ("Illegal call: requests to /emit have to contain non-empty parameter D.");
					
					String amountText = req.getParameter("F");
					if (amountText == null || amountText.trim().length() == 0)
						throw new MalformedRequestException ("Illegal call: requests to /emit have to contain non-empty parameter F.");

					long amount = Long.parseLong(amountText);
					
					String signature = new String(msg.getMessage());
					
					String targetIssuer = req.getParameter("K");
					
					// just for fun: check if redirect parameters are the same as in the messageHandler
					// might differ when the ServletContainer messes sessions up.
					String decodedMessage = URLDecoder.decode(msg.getClearText(), "UTF-8");
					if(			decodedMessage.indexOf(targetDigest) == -1
							|| 	decodedMessage.indexOf(String.valueOf(amount)) == -1
					)
					{
						throw new MalformedRequestException("Fatal Error: MessageHandler and request parameters do not match.");
					}
					
					if(amount <= 0)
						throw new MalformedRequestException("Fatal Error: The amount of ePoints to emit must not be less or equal to zero.");
					
					targetIssuer = checkTargetIssuer(targetIssuer);
						
					// get cert for rest, make targetDist valid with amount
					byte[] newCert = issuer.emit(targetDigest, amount, signature, msg.getSignersPrimaryKeyFingerprint(), targetIssuer, extractCommentFromRequest(req));

					resp.setStatus(200);
					resp.getOutputStream().write(newCert);
				}
			}
		}
		catch(Exception ex)
		{
			// Aquire a ProcessingException for the thrown exception to be able
			// to return the correct response status and short text
			ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
			respondWithProcessingException(resp, pEx);
		}
	}

	private void doUpdateAI(HttpServletRequest req, HttpServletResponse resp) throws IOException
	{
		// update an AIs parameters (create if necessary)
		try
		{
			if (req.getParameterMap().isEmpty())
			{
				throw new MalformedRequestException("Illegal call without arguments");
			}
			else
			{
				// This attribute has to be set by SecureMessageRedirectionFilter that validates/decrypts and redirects messages
				MessageDecryptionVerificationHandler msg = null;
				
				if (req.getAttribute(SecureMessageRedirectionFilter.ATTR_SECURE_MESSAGE_HANDLER) != null)
				{
					msg = (MessageDecryptionVerificationHandler)req.getAttribute(SecureMessageRedirectionFilter.ATTR_SECURE_MESSAGE_HANDLER);

					if (!msg.isValid().booleanValue())
						throw new MalformedRequestException("Illegal call to /updateAI with non-valid signed message.");
					else if (msg.getSignersFingerprint() == null)
						throw new BadsignatureException ("Illegal call: cannot resolve signer.");
					else if (!msg.isClearText().booleanValue())
						throw new BadsignatureException ("Illegal call: signed requests to /updateAI must be clearsigned.");
					else if (!Boolean.TRUE.equals(msg.signerIsIssuersMasterKey()))
						throw new BadsignatureException ("Illegal call: signed requests to /updateAI must be clearsigned by the issuers master key.");
					// msg is valid and signed, proceed
				}
				
				
				String authorizedIssuerId = req.getParameter("A");
				String authorizedIssuerPublicKeyBlock = req.getParameter("KEY");
				
				String emissionLimitLowerBoundary = req.getParameter("L");
				String emissionLimitHigherBoundary = req.getParameter("H");
				
				if ((emissionLimitLowerBoundary != null || emissionLimitHigherBoundary != null) && (msg == null || !Boolean.TRUE.equals(msg.isValid())))
					throw new BadsignatureException ("Illegal call: requests to /updateAI changing emission bounds have be clearsigned by the issuers master key.");

				String targetIssuer = req.getParameter("K");

				// verify that the request was originally created for this issuer instance
				targetIssuer = checkTargetIssuer(targetIssuer);
				
				issuer.updateAuthorizedIssuer(authorizedIssuerId, authorizedIssuerPublicKeyBlock, emissionLimitLowerBoundary, emissionLimitHigherBoundary, msg, targetIssuer);

				resp.setStatus(200);
				resp.getOutputStream().write("OK".getBytes());
				resp.flushBuffer();
			}
		}
		catch(Exception ex)
		{
			// Aquire a ProcessingException for the thrown exception to be able
			// to return the correct response status and short text
			ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
			respondWithProcessingException(resp, pEx);
		}
	}
	
	private void doAction(HttpServletRequest req, HttpServletResponse resp) throws IOException
	{
		if (req.getParameter("B") != null && req.getParameter("D") != null && req.getParameter("F") != null && req.getParameter("C") != null)
		{
			doActionSplit(req, resp);
		}
		else if (
					(req.getParameter("B") != null && req.getParameter("D") != null)
				||	(req.getParameter("B") != null && req.getParameter("D") != null && req.getParameter("G") != null)
		)
		{
			doActionExchange(req, resp);
			
		}
		else if (
						(req.getParameter("B") != null && req.getParameter("KEY") != null)
					||	(req.getParameter("B") != null && req.getParameter("KEY") != null && req.getParameter("G") != null)
		)
		{
			doActionMergeObligation(req, resp);
		}
		else if (req.getParameter("D") != null && req.getParameter("KEYDATA") != null)
		{
			doActionProcessDraft(req, resp);

		}
		else
		{
			ProcessingException pEx = getProcessingExceptionMapper().getMappedException(new MalformedRequestException("Request did not match parameter scheme and will be ignored."));
			respondWithProcessingException(resp, pEx);
		}
	}

	private void doActionProcessDraft(HttpServletRequest req,
			HttpServletResponse resp) throws IOException {
		try
		{
			// Draft message -> Draft endorsement request -> obligation split request
			String draft = req.getParameter("D");
			String pubKey = req.getParameter("KEYDATA");
			
			if (draft == null || draft.length() == 0)
				throw new MalformedRequestException("Illegal value for parameter 'D'. Only a valid, signed draft will be accepted.");

			if (pubKey == null || pubKey.length() == 0)
				throw new MalformedRequestException("Illegal value for parameter 'D'. Only a valid, signed draft will be accepted.");

			MessageDecryptionVerificationHandler signedMessage = ServletMessageDecryptionVerificationFactory.getMessageHandler();
			
			// add the public key to use for verification
			signedMessage.addTemporaryPublicKey(pubKey);
			
			signedMessage.setMessage(draft.getBytes("UTF-8"));
			signedMessage.analyse();
			
			if (signedMessage.isValid() == null)
				throw new ProcessingException("The Decryption/Verification-Handler could not validate message. Request ignored.");
			else if (signedMessage.isValid().booleanValue()) 
			{
				// message could be processed and signature is valid
				// parse draft and start processing if all parameters can be found
				String messageContent = signedMessage.getClearText();
				
				// extract clear text from signedMessage
				if (messageContent != null && messageContent.length() > 0)
				{
					byte[] newCert = issuer.processDraftMessage(signedMessage);
					
					if (newCert != null)
					{
						resp.setStatus(200);
						resp.getOutputStream().write(newCert);
					}
					else
						throw new ConflictException("Draft endorsement could not be processed.");
				}
				else
					throw new MalformedRequestException("Illegal value for parameter 'D', empty message content.");
			}
			else
				throw new ProcessingException("The signature of the draft is not valid. Request ignored.");
		}
		catch (Exception ex)
		{
			// Aquire a ProcessingException for the thrown exception to be able
			// to return the correct response status and short text
			ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
			respondWithProcessingException(resp, pEx);
		}
	}

	private void doActionMergeObligation(HttpServletRequest req,
			HttpServletResponse resp) throws IOException {
		// obligation merge request
		try
		{
			String rand = req.getParameter("B");
			String key = req.getParameter("KEY");
			
			// check the rand parameter
			if (rand == null || rand.length() == 0)
				throw new MalformedRequestException("Invalid value submitted for parameter 'B'");
			
			// check the KEY parameter
			if (key == null || key.length() == 0)
				throw new MalformedRequestException("Invalid value submitted for parameter 'KEY'");
			else
				key = key.toUpperCase().replaceAll(" ","");
			
			log.trace("Obligation merge request: rand " + rand + " is going to be merged with account " + key + (req.getParameter("G") != null ? " and then redirected to: " + req.getParameter("G"):""));

			// check availability of key here? If we sell obligations to somebody whose key we can not
			// resolve, he won't be able to call them as we cannot verify the draft's signature
			
			if (req.getParameter("G") == null)
			{
				// no redirect required
				// get cert for rest, make targetDist valid with amount
				byte[] newCert = issuer.mergeObligation(RedCent.RETURN_RESULTING_CERT, rand, key, extractCommentFromRequest(req));

				if (newCert != null)
				{
					resp.setStatus(200);
					resp.getOutputStream().write(newCert);
				}
				else
					throw new ConflictException("Obligation merge could not be processed.");
			}
			else
			{
				String redirect = req.getParameter("G");
				
				if (redirect.endsWith("="))
				{
					try
					{
						byte[] result = issuer.mergeObligation(RedCent.RETURN_TRANSACTION_RECEIPT, rand, key, extractCommentFromRequest(req));
						resp.sendRedirect(redirect + new String(new URLCodec().encode(result)));
					}
					catch(NotFoundException nfe)
					{
						// In case the transaction fails, just follow the redirect *without* a transaction receipt
						resp.sendRedirect(redirect);
					}
				}
				else if (redirect.endsWith("?"))
				{
					try
					{
						byte[] result = issuer.mergeObligation(RedCent.RETURN_TRANSACTIONVALUE_STRING_AS_BYTEARRAY_UTF8, rand, key, extractCommentFromRequest(req));
						resp.sendRedirect(redirect + new String(result, "UTF-8"));
					}
					catch(NotFoundException nfe)
					{
						// In case the transaction fails, just follow the redirect with a transaction value of zero
						resp.sendRedirect(redirect + "0");
					}
				}
				else
				{
					issuer.mergeObligation(RedCent.RETURN_RESULTING_CERT, rand, key, extractCommentFromRequest(req));
					resp.sendRedirect(redirect);
				}
			}
		}
		catch (Exception ex)
		{
			// Aquire a ProcessingException for the thrown exception to be able
			// to return the correct response status and short text
			ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
			respondWithProcessingException(resp, pEx);
		}
	}

	private void doActionExchange(HttpServletRequest req,
			HttpServletResponse resp) throws IOException {
		// exchange request without additional redirect parameter G
		try
		{
			log.debug("Exchange request: rand is " + req.getParameter("B") + " and targetDigest is " + req.getParameter("D") + (req.getParameter("G") != null ? " and then redirect request to: " + req.getParameter("G"):""));
			
			String redirect = req.getParameter("G");
			
			if (redirect == null)
			{
				// normal request, just process and return certificate
				byte[] newCert = issuer.exchange(RedCent.RETURN_RESULTING_CERT, req.getParameter("B"), req.getParameter("D"), extractCommentFromRequest(req));

				resp.setStatus(200);
				resp.getOutputStream().write(newCert);
			}
			else
			{
				if (redirect.endsWith("="))
				{
					try
					{
						byte[] result = issuer.exchange(RedCent.RETURN_TRANSACTION_RECEIPT, req.getParameter("B"), req.getParameter("D"), extractCommentFromRequest(req));
						resp.sendRedirect(redirect + new String(new URLCodec().encode(result)));
					}
					catch(NotFoundException nfe)
					{
						// In case the transaction fails, just follow the redirect *without* a transaction receipt
						resp.sendRedirect(redirect);
					}
				}
				else if (redirect.endsWith("?"))
				{
					try
					{
						byte[] result = issuer.exchange(RedCent.RETURN_TRANSACTIONVALUE_STRING_AS_BYTEARRAY_UTF8, req.getParameter("B"), req.getParameter("D"), extractCommentFromRequest(req));
						resp.sendRedirect(redirect + new String(result, "UTF-8"));
					}
					catch(NotFoundException nfe)
					{
						// In case the transaction fails, just follow the redirect with a transaction value of zero
						resp.sendRedirect(redirect + "0");
					}
				}
				else
				{
					issuer.exchange(RedCent.RETURN_RESULTING_CERT, req.getParameter("B"), req.getParameter("D"), extractCommentFromRequest(req));
					resp.sendRedirect(redirect);
				}
			}
		}
		catch (Exception ex)
		{
			// Aquire a ProcessingException for the thrown exception to be able
			// to return the correct response status and short text
			ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
			respondWithProcessingException(resp, pEx);
		}
	}

	private void doActionSplit(HttpServletRequest req, HttpServletResponse resp)
			throws IOException {
		// split request
		try
		{
			String rand = req.getParameter("B");
			String targetDigest = req.getParameter("D");
			long amount = Long.parseLong(req.getParameter("F"));
			String restDigest = req.getParameter("C");
			
			if (amount <= 0)
				throw new MalformedRequestException("Parameter 'F' must not be less or equal to zero.");
			
			if (targetDigest.equals(restDigest))
				throw new MalformedRequestException("Parameters 'D' and 'C' must not be equal.");
			
			log.trace("Split request: rand to split is " +  rand
					+ " and first targetDigest is " + targetDigest
					+ " and first amount is " + amount
					+ " and rest digest is " + restDigest
			);

			// get cert for rest, make targetDist valid with amount
			byte[] newCert = issuer.split(rand, targetDigest, amount, restDigest, extractCommentFromRequest(req));
			
			resp.setStatus(200);
			resp.getOutputStream().write(newCert);
		}
		catch (Exception ex)
		{
			// Aquire a ProcessingException for the thrown exception to be able
			// to return the correct response status and short text
			ProcessingException pEx = getProcessingExceptionMapper().getMappedException(ex);
			respondWithProcessingException(resp, pEx);
		}
	}

	private String checkTargetIssuer(String targetIssuer) throws MalformedRequestException, ConflictException
	{
		if(targetIssuer == null)
			throw new MalformedRequestException("Fatal Error: No parameter 'K' submitted containing the issuer's fingerprint.");
		else
		{
			targetIssuer = targetIssuer.toUpperCase().replaceAll(" ","");
			String thisIssuer = issuer.getIssuerFingerPrintText().toUpperCase().replaceAll(" ", "");
			
			if (!targetIssuer.equals(thisIssuer))
				throw new MalformedRequestException("Fatal Error: The submitted issuers fingerprint("+targetIssuer+") does not match the fingerprint of this issuer ("+thisIssuer+").");
		}
		return targetIssuer;
	}

	protected void respondWithProcessingException(HttpServletResponse resp, ProcessingException mappedException) throws IOException
	{
		log.debug("Responding with ProcessingException: " + mappedException, mappedException);
		
		// is forwarded to the container's default error handler
		resp.sendError(mappedException.getStatus(), mappedException.getMessage());
	}

	protected ProcessingExceptionMapper defaultExceptionMapper;
	protected ProcessingExceptionMapper getProcessingExceptionMapper()
	{
		if (defaultExceptionMapper == null)
			defaultExceptionMapper = new DefaultProcessingExceptionMapper();
		
		return defaultExceptionMapper;
	}

	public String getServletInfo()
	{
		return "RedCent application gateway servlet.";
	}
	
	/**
	 * Output a parameterMap to Sting and correctly print arrays of String 
	 * @param map
	 * @return
	 */
	private String parameterMapToString(Map map) {
		StringBuffer buf = new StringBuffer("{");
		
		if (map != null) {
			Iterator keyIter = map.keySet().iterator();
			
			while (keyIter.hasNext()) {
				
				Object key = keyIter.next();
				Object value = map.get(key);
				
				buf.append(key.toString() + "=");

				if (value != null){
					if (value instanceof String[]) {
						buf.append(Arrays.asList(((String[])value)));
					} else
						buf.append(value);
				}
				
				if (keyIter.hasNext())
					buf.append(", ");
			}
			
		}
		
		buf.append("}");
		
		return buf.toString();
	}
	
	/**
	 * Extract comment parameter "I" from request and limit it to 256 chars before sending it 
	 * to the application core.
	 * 
	 * @param req
	 * @return
	 */
	private String extractCommentFromRequest(HttpServletRequest req) {

		if (req.getParameterMap().containsKey("I")) {
			String comment = req.getParameter("I");
			
			if (comment != null) {

				if(comment.length() > 256)
					comment = comment.substring(0, 256);
				
				comment = comment.replace('~', '_');
			}
			
			return comment;
		}
		else
			return null;
	}
}
