/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.servlet.crypto;

import java.io.FileInputStream;

import javax.servlet.ServletContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPPublicKeyRingCollection;
import org.bouncycastle.openpgp.PGPSecretKeyRingCollection;
import org.bouncycastle.openpgp.PGPUtil;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.context.ServletContextAware;
import org.springframework.web.context.support.WebApplicationContextUtils;

import de.nachtlicht_media.epoint.redcent.RedCent;
import de.nachtlicht_media.epoint.redcent.crypto.BouncycastleMessageDecryptionVerificationHandler;
import de.nachtlicht_media.epoint.redcent.interfaces.AuthorizedIssuerStore;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageDecryptionVerificationHandler;
import de.nachtlicht_media.epoint.redcent.servlet.RedCentServletConfig;

public class ServletMessageDecryptionVerificationFactory implements ServletContextAware
{
	private static final Log log = LogFactory.getLog(ServletMessageDecryptionVerificationFactory.class);
	private ServletContext context;
	
	@Autowired
	private RedCent redcent;

	/** 
	 * Don't allow instances of this class. Only calls to static methods are allowed.
	 *
	 */
	private ServletMessageDecryptionVerificationFactory()
	{
	}

	public void setServletContext(ServletContext context)
	{
		log.info("ServletContext " + context + " was injected.");
		this.context = context;
	}

	private static PGPPublicKeyRingCollection issuersPublicKey;
	private static PGPSecretKeyRingCollection issuersSecretKey;
	private static String issuersPrivateKeyPassword;
	private static PGPPublicKeyRingCollection trustedPublicKeys;
	private static PGPPublicKeyRing masterKey;
	private static AuthorizedIssuerStore authorizedIssuerStore;
	
	public void init() throws Exception
	{
		if (issuersPublicKey != null)
			throw new Exception("ServletMessageDecryptionVerificationFactory is already initialized!");
		
		issuersPublicKey = new PGPPublicKeyRingCollection(PGPUtil.getDecoderStream(new FileInputStream(context.getRealPath(context.getInitParameter("IssuersPublicKeyRingPath")))));
		issuersSecretKey = new PGPSecretKeyRingCollection(PGPUtil.getDecoderStream(new FileInputStream(context.getRealPath(context.getInitParameter("IssuersPrivateKeyRingPath")))));
		issuersPrivateKeyPassword = context.getInitParameter("IssuersPrivateKeyPassword");
		trustedPublicKeys = new PGPPublicKeyRingCollection(PGPUtil.getDecoderStream(new FileInputStream(context.getRealPath(context.getInitParameter("TrustedPublicKeyRingPath")))));

		String masterKeyId = context.getInitParameter("MasterKeyId").toLowerCase();
		
		if (masterKeyId.startsWith("0x"))
			masterKeyId = masterKeyId.substring(2);
		
		masterKey = trustedPublicKeys.getPublicKeyRing(Long.parseLong(masterKeyId, 16));
		if (masterKey == null)
			throw new IllegalArgumentException("Could not initialize ServletMessageDecryptionVerificationFactory: no trusted public key with id " + masterKeyId + " could be found as a master key!");
		
		authorizedIssuerStore = redcent;
	}

	public static synchronized MessageDecryptionVerificationHandler getMessageHandler() throws Exception
	{
		return new BouncycastleMessageDecryptionVerificationHandler(authorizedIssuerStore, issuersPublicKey, issuersSecretKey, issuersPrivateKeyPassword, masterKey, trustedPublicKeys);
	}
}
