/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.servlet.filter;

import java.io.IOException;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * This Filter is used to allow calls to /action, /emit and other side-effecting
 * URLs only via the /secureMessage gateway or via a secure transport
 * @author js
 */

public class DelayActionsFilter implements Filter
{
	private static final Log log = LogFactory.getLog(DelayActionsFilter.class);
	private FilterConfig filterConfig;
	
	private static final long MIN_FINAL_SLEEP_TIME = 250;
	
	private long preDelayRequestsToMinMillis = 0;
	private long delayRequestsToMinMillis = 1500l;
	
	public void init(FilterConfig filterConfig) throws ServletException
	{
		log.info("init()");
		this.filterConfig = filterConfig;
		
		if (this.filterConfig.getInitParameter("DelayToMinMillis") != null)
		{
			this.delayRequestsToMinMillis = Long.parseLong(this.filterConfig.getInitParameter("DelayToMinMillis"));
			
			if (delayRequestsToMinMillis < MIN_FINAL_SLEEP_TIME)
				log.warn("Setting 'DelayToMinMillis' to a value smaller than "+MIN_FINAL_SLEEP_TIME+" ms. makes no sense, because the final delay after processing the request is only performed if there are more than "+MIN_FINAL_SLEEP_TIME+" ms. to sleep left.");
		}
		
		if (this.filterConfig.getInitParameter("PreDelayToMinMillis") != null)
		{
			this.preDelayRequestsToMinMillis = Long.parseLong(this.filterConfig.getInitParameter("PreDelayToMinMillis"));
			
			if (this.preDelayRequestsToMinMillis >= delayRequestsToMinMillis)
				log.warn("Setting 'PreDelayToMinMillis' time to " + preDelayRequestsToMinMillis + " ms. makes 'DelayToMinMillis' useless.");
		}
	}

	public void destroy()
	{
		log.info("destroy()");
	}
	
	public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException,
	ServletException
	{
		// save startTime to local memory
		long startTime = System.currentTimeMillis();
		
		
		// If requests shall be delayed before entering the application realm, delay them now
		// this function maybe used to prevent hammering
		// Attention: this feature is only useful if the servlet container kills the request processing
		// after the session was closed by the remote client. Otherwise this feature would allow
		// to open more connections and flood the server...
		if (preDelayRequestsToMinMillis > 0)
		{
			try
			{
				Thread.sleep(preDelayRequestsToMinMillis);
			}
			catch (InterruptedException e)
			{
				log.warn("ActionDelayFilter was interrupted", e);
			}
		}

		// continue to process requests
		chain.doFilter(request,response);
		
		// calculate remaining sleepTime and send thread to sleep if
		// the remaining time is more than 250 ms. so that suspending
		// the thread is not taking more clock time than the amount
		// that is gained by pausing this thread -- and thereby preventing
		// the client from continuing to send requests.
		long sleeptime = delayRequestsToMinMillis - (System.currentTimeMillis() - startTime);
		if (sleeptime >= MIN_FINAL_SLEEP_TIME)
		{
			try
			{
				Thread.sleep(sleeptime);
			}
			catch (InterruptedException e)
			{
				log.warn("ActionDelayFilter was interrupted", e);
			}
		}
	}
}
