/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.servlet.filter;

import java.io.IOException;
import java.net.URLDecoder;
import java.net.URLEncoder;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.nachtlicht_media.epoint.redcent.crypto.BouncycastleUtil;
import de.nachtlicht_media.epoint.redcent.exception.InvalidPGPMessageException;
import de.nachtlicht_media.epoint.redcent.exception.RedcentProcessingException;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageDecryptionVerificationHandler;
import de.nachtlicht_media.epoint.redcent.servlet.crypto.ServletMessageDecryptionVerificationFactory;
import de.nachtlicht_media.epoint.redcent.servlet.exception.ProcessingException;

public class SecureMessageRedirectionFilter implements Filter
{
	private static final Log log = LogFactory.getLog(SecureMessageRedirectionFilter.class);
	public static final String ATTR_SECURE_MESSAGE_FLAG = "ATTR_SECURE_MESSAGE_FLAG";
	public static final String ATTR_SECURE_MESSAGE_HANDLER = "ATTR_SECURE_MESSAGE_HANDLER";
	
	private FilterConfig filterConfig;
	
	public void init(FilterConfig filterConfig) throws ServletException
	{
		log.info("init()");
		this.filterConfig = filterConfig;
	}

	public void destroy()
	{
		log.info("destroy()");
	}
	
	public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException
	{
		if (request.getParameterMap().isEmpty())
			throw new ProcessingException("Illegal call to " + filterConfig.getFilterName() + " with empty parameterMap.");
		else if (request.getAttribute(ATTR_SECURE_MESSAGE_FLAG) != null)
		{
			log.warn("Illegal call to " + filterConfig.getFilterName() + " with a request that was already processed.");
			throw new ProcessingException(404, "Illegal call to /secureMessage with target /secureMessage", "Illegal call to " + filterConfig.getFilterName() + " with a request that was already processed.");
		}
		else
		{
			request.setAttribute(ATTR_SECURE_MESSAGE_FLAG, ATTR_SECURE_MESSAGE_FLAG);
			if (request.getParameterMap().size() == 1)
			{
				String message = request.getParameter("M");
				
				if (message != null && message.length() > 0)
				{
					try
					{
						// Decode
						MessageDecryptionVerificationHandler msg = ServletMessageDecryptionVerificationFactory.getMessageHandler();
						
						byte[] messageBytes = message.getBytes("UTF-8");
						
						if (log.isTraceEnabled())
							log.trace("Processing incoming message: " + BouncycastleUtil.convertKeyFingerprintToHex(messageBytes));
						
						msg.setMessage(messageBytes);
						msg.analyse();
						
						if (msg.isValid() == null)
						{
							// if msg.isValid() is null there must have occured a processing exception that prevented
							// the msgProcessor from deciding whether the message is valid or not.
							
							// Do not process this request any more
							throw new ProcessingException("The Decryption/Verification-Handler could not validate message. Request ignored.");
						}
						else if(msg.isValid().booleanValue())
						{
							request.setAttribute(ATTR_SECURE_MESSAGE_HANDLER, msg);
							
							String redirection = msg.getClearText();
							// messages routed through this filter MUST contain a redirect call to e.g. /action?... or /emit?...
							if (redirection != null && redirection.length() > 0 && (redirection.startsWith("/") || redirection.startsWith("%2F")))
							{
								// Signed messages usually end with a linebreak that must not be forwarded
								if (redirection.endsWith("\n"))
									redirection = redirection.substring(0, redirection.length() - 1);
								
								// Correctly handle messages where the redirection is url-encoded.
								// Decode and then re-encode parameter values to url-encoding
								// to prevent e.g. loss of + chars in ascii public keys.
								if (redirection.startsWith("%2F")) {
									String rawUrl = URLDecoder.decode(redirection, "UTF-8");

									// Signed messages usually end with a linebreak that must not be forwarded
									if (rawUrl.endsWith("\n"))
										rawUrl = rawUrl.substring(0, rawUrl.length() - 1);
									
									// if the redirect contains a query part
									if (rawUrl.indexOf("?") != -1) {
										
										String query = rawUrl.substring(rawUrl.indexOf("?") + 1);
										String queryParts[] = query.split("&");
										
										redirection = rawUrl.substring(0, rawUrl.indexOf("?") + 1 );
										
										for (int i = 0; i < queryParts.length; i++) {
											String parameterRedir = queryParts[i];
											if (parameterRedir.indexOf("=") != -1) {
												parameterRedir = parameterRedir.substring(0, parameterRedir.indexOf("=") + 1) + URLEncoder.encode(parameterRedir.substring(parameterRedir.indexOf("=") + 1), "UTF8"); 
											} else
												parameterRedir = URLEncoder.encode(parameterRedir, "UTF8");
											
											redirection = redirection + parameterRedir;
											if (i < queryParts.length-1)
												redirection += "&";
										}
									}
								}
								
								// looks like a valid redirect and stop filtering for this call
								log.debug("Forwarding request to '" + redirection + "'");
								request.setCharacterEncoding("UTF8");
								request.getRequestDispatcher(redirection).forward(request, response);
							}
							else
								throw new ProcessingException("Malformed message: the encrypted/signed message MUST contain a valid redirection.");
						}
						else
							throw new ProcessingException("The Decryption/Verification-Handler returned an invalid message state. Request ignored.");
					}
					catch (InvalidPGPMessageException e)
					{
						e.printStackTrace();
						request.setAttribute(ATTR_SECURE_MESSAGE_HANDLER, null);
						ProcessingException pe = new ProcessingException(RedcentProcessingException.FORB_BAD_CERT, "Error decoding message", e.getMessage());
						throw pe;
					}
					catch(ProcessingException ex)
					{
						request.setAttribute(ATTR_SECURE_MESSAGE_HANDLER, null);
						throw ex;
					}
					catch (Exception e)
					{
						request.setAttribute(ATTR_SECURE_MESSAGE_HANDLER, null);
						ProcessingException pe = new ProcessingException(RedcentProcessingException.CONFLICT, "Error decoding message", e.getMessage());
						throw pe;
					}
				}
				else
					throw new ProcessingException("Illegal call to " + filterConfig.getFilterName() + " with empty message parameter.");
			}
		}
	}
}
