/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.sqlImplementation;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.List;

import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.annotation.Autowired;

import de.nachtlicht_media.epoint.redcent.AuthorizedIssuer;
import de.nachtlicht_media.epoint.redcent.EpointRecord;
import de.nachtlicht_media.epoint.redcent.RedCent;
import de.nachtlicht_media.epoint.redcent.exception.DatabaseErrorException;
import de.nachtlicht_media.epoint.redcent.exception.NotFoundException;
import de.nachtlicht_media.epoint.redcent.exception.RedcentProcessingException;
import de.nachtlicht_media.epoint.redcent.interfaces.EpointCertificateTypes;
import de.nachtlicht_media.epoint.redcent.sqlImplementation.RedCentSqlImplementation.StatsCollector;

public class EpointSQLDB
{
	@Autowired(required=false)
	StatsCollector statsCollector;
	
	private static final Log log = LogFactory.getLog(EpointSQLDB.class);

	public EpointSQLDB()
	{
	}

	public void verifyDatabase(Connection con) throws Exception
	{
		// verify that all serialnumbers do not have gaps
		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			// get number of assigned certificates
			pStm = con.prepareStatement("SELECT COUNT(serialnumber) from certificate");			
			rs = pStm.executeQuery();
			
			rs.next();
			int count = rs.getInt(1);
			closeStatementAndResultSet(pStm, rs);
			pStm = null;
			rs = null;
			
			if (count > 0)
			{
				// verify that that the last assigned certificate's serialnumber is equal to the total count of numbers
				// this together with the unique and NOT NULL-constraint of the database column guarantees consistent serial numbers
				pStm = con.prepareStatement("SELECT MAX(serialnumber) from certificate");			
				rs = pStm.executeQuery();
				
				rs.next();
				int max = rs.getInt(1);
				closeStatementAndResultSet(pStm, rs);
				pStm = null;
				rs = null;
				
				if (count == max)
					log.info("Verification of certificate database succeeded, the last assigned certificate has serial number " + max);
				else
					throw new Exception("The found serial numbers are inconsistent, the application may not proceed.");
			}
			else
				log.info("Verification of empty certificate database succeeded.");

			// get total amount of emitted ePoints from certificate store
			pStm = con.prepareStatement("SELECT SUM(value) FROM certificate WHERE isValid=?");
			pStm.setBoolean(1, true);
			rs = pStm.executeQuery();
			
			rs.next();
			long totalCertificateValue = rs.getLong(1);
			closeStatementAndResultSet(pStm, rs);
			pStm = null;
			rs = null;

			// verify that total amount from certificate store and total amount from balance match
			if (!new Long(totalCertificateValue).equals(getTotalAmount(con)))
				throw new Exception("The database's balance is not correct: " + totalCertificateValue + " ePoints are balanced, but the total value of valid certificates is: " + getTotalAmount(con));
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
	}
	
	public EpointRecord findValidRecordByID(Connection con, String id) throws SQLException
	{
		long startTime = System.currentTimeMillis();
		if (id != null)
		{
			PreparedStatement pStm = null;
			ResultSet rs = null;

			try
			{
				pStm = con.prepareStatement("SELECT * from certificate WHERE id=? AND isValid=?");			
				pStm.setString(1, id);
				pStm.setBoolean(2, true);
				rs = pStm.executeQuery();

				if(statsCollector != null)
					statsCollector.logStatRecord(new RedCentSqlImplementation.StatRecord(RedCentSqlImplementation.StatRecord.AC_FIND_VALID_RECORD, id, System.currentTimeMillis()-startTime));

				return createFromResultSet(rs);
			}
			finally
			{
				closeStatementAndResultSet(pStm, rs);
			}
		}
		
		return null;
	}
	
	public EpointRecord findLastObligationBeforeSerial(Connection con, String id, long serial) throws SQLException
	{
		long startTime = System.currentTimeMillis();
		if (id != null)
		{
			PreparedStatement pStm = null;
			ResultSet rs = null;

			try
			{
				String sqlString =
						"SELECT * FROM certificate WHERE serialnumber=(" +
							"SELECT max(serialnumber) FROM certificate WHERE (invalidatedid=?) AND certificateType=?" +
							" AND serialnumber < ?" +
						")";
				
				pStm = con.prepareStatement(sqlString);
				
				int parameterIndex = 1;
				pStm.setString(parameterIndex++, id);
				pStm.setString(parameterIndex++, EpointCertificateTypes.OBLIGATION_CERTIFICATE);
				pStm.setLong(parameterIndex++, serial);
				
				rs = pStm.executeQuery();

				if(statsCollector != null)
					statsCollector.logStatRecord(new RedCentSqlImplementation.StatRecord(RedCentSqlImplementation.StatRecord.AC_FIND_OBLIGATION_BEFORE, id, System.currentTimeMillis()-startTime));

				return createFromResultSet(rs);
			}
			finally
			{
				closeStatementAndResultSet(pStm, rs);
			}
		}
		
		return null;
	}
	
	public List<EpointRecord> findObligationHistoryByID (Connection con, String id, Long startSerial, Long endSerial) throws SQLException
	{
		long startTime = System.currentTimeMillis();
		if (id != null)
		{
			PreparedStatement pStm = null;
			ResultSet rs = null;

			try
			{
				String sqlString = "SELECT * FROM certificate WHERE (id=? OR invalidatedid=?) AND certificateType=?" +
						(startSerial != null ? " AND serialnumber >= ?" : "")+
						(endSerial != null ? " AND serialnumber <= ?" : "")+
						" ORDER BY serialnumber" +
						" LIMIT 500";
				
				pStm = con.prepareStatement(sqlString);
				
				int parameterIndex = 1;
				pStm.setString(parameterIndex++, id);
				pStm.setString(parameterIndex++, id);
				pStm.setString(parameterIndex++, EpointCertificateTypes.OBLIGATION_CERTIFICATE);
				
				if (startSerial != null)
					pStm.setLong(parameterIndex++, startSerial);
				
				if (endSerial != null)
					pStm.setLong(parameterIndex++, endSerial);
				
				rs = pStm.executeQuery();

				if(statsCollector != null)
					statsCollector.logStatRecord(new RedCentSqlImplementation.StatRecord(RedCentSqlImplementation.StatRecord.AC_FIND_OBLIGATION_HISTORY, id, System.currentTimeMillis()-startTime));

				return createListFromResultSet(rs);
			}
			finally
			{
				closeStatementAndResultSet(pStm, rs);
			}
		}
		
		return null;
	}
	
	
	
	public Collection findInvalidRecordsByID(Connection con, String id) throws SQLException
	{
		long startTime = System.currentTimeMillis();
		if (id != null)
		{
			PreparedStatement pStm = null;
			ResultSet rs = null;

			try
			{
				pStm = con.prepareStatement("SELECT * from certificate WHERE invalidatedid=? AND isValid=?");			
				pStm.setString(1, id);
				pStm.setBoolean(2, false);
				rs = pStm.executeQuery();

				if(statsCollector != null)
					statsCollector.logStatRecord(new RedCentSqlImplementation.StatRecord(RedCentSqlImplementation.StatRecord.AC_FIND_INVALID_RECORD, id, System.currentTimeMillis()-startTime));
				return createListFromResultSet(rs);
			}
			finally
			{
				closeStatementAndResultSet(pStm, rs);
			}
		}
		
		return null;
	}
	
	public EpointRecord findBySerial(Connection con, long serial) throws DatabaseErrorException
	{
		try
		{
			PreparedStatement pStm = null;
			ResultSet rs = null;

			try
			{
				pStm = con.prepareStatement("SELECT * from certificate WHERE serialNumber=?");			
				pStm.setLong(1, serial);
				rs = pStm.executeQuery();

				return createFromResultSet(rs);
			}
			finally
			{
				closeStatementAndResultSet(pStm, rs);
			}
		}
		catch(SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
	}
	public Long getLatestSerial(Connection con) throws DatabaseErrorException
	{
		try
		{
			Statement stm = null;
			ResultSet rs = null;

			try
			{
				stm = con.createStatement();
				rs = stm.executeQuery("SELECT MAX(serialNumber) FROM certificate");
				if (rs.next())
					return new Long(rs.getLong(1));
				return null;
			}
			finally
			{
				closeStatementAndResultSet(stm, rs);
			}
		}
		catch(SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
	}
	
//	private Long getBalanceInfosLastEntry(Connection con) throws DatabaseErrorException
//	{
//		try
//		{
//			Statement stm = null;
//			ResultSet rs = null;
//
//			try 
//			{
//				stm = con.createStatement();
//				rs = stm.executeQuery("SELECT MAX(serialnumber) FROM balance");
//
//				if (rs.next())
//					return new Long(rs.getLong(1));
//				
//				return null;
//			}
//			finally
//			{
//				closeStatementAndResultSet(stm, rs);
//			}
//		}
//		catch(SQLException ex)
//		{
//			throw new DatabaseErrorException(ex);
//		}
//	}
	
	public String getBalanceInfo(Connection con, String key) throws DatabaseErrorException
	{
		try
		{
			PreparedStatement stm = null;
			ResultSet rs = null;

			try 
			{
				stm = con.prepareStatement("SELECT authorizedBy, SUM(value) FROM balance "+(key != null ? "WHERE authorizedBy=?" : "")+" GROUP BY authorizedBy ORDER BY authorizedBy");
				
				if (key != null)
					stm.setString(1, key);
				
				rs = stm.executeQuery();

				StringBuffer buf = new StringBuffer();
				
				while (rs.next())
				{
					if(buf.length() != 0)
						buf.append("\r");
					
					buf.append(rs.getString(1));
					buf.append("\t");
					buf.append(rs.getLong(2));
				}
				
				return buf.toString();
			}
			finally
			{
				closeStatementAndResultSet(stm, rs);
			}
		}
		catch(SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
	}
	
	public Long getTotalAmount(Connection con) throws DatabaseErrorException
	{
		try
		{
			Statement stm = null;
			ResultSet rs = null;

			try 
			{
				stm = con.createStatement();
				rs = stm.executeQuery("SELECT SUM(value) FROM balance");
				if (rs.next())
					return new Long(rs.getLong(1));
				return null;
			}
			finally
			{
				closeStatementAndResultSet(stm, rs);
			}
		}
		catch(SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
	}
	
	public EpointRecord createRecord(Connection con, long value, String id, int issuingCause, String certificateType, String templateReference) throws SQLException
	{
		long startTime = System.currentTimeMillis();
		
		PreparedStatement pStm = null;
		ResultSet rs = null;
		
		try
		{
			// DERBY: derby does not support syntax 'RETURNING serialnumber'. Comment it out and execute select statement below instead
			//cert_serial_next()
			pStm = con.prepareStatement("INSERT INTO certificate (serialnumber, value, id, issuingCause, certificateType, templateReference) VALUES ((SELECT COALESCE((MAX(serialnumber)+1),1) FROM certificate),?,?,?,?,?) RETURNING serialnumber, txDate");
			pStm.setLong(1,value);
			pStm.setString(2,id);
			pStm.setShort(3,(short)issuingCause);
			pStm.setString(4, certificateType);
			pStm.setString(5, templateReference);

//			pStm.executeUpdate();
//			pStm.close();
			
//			pStm = con.prepareStatement("SELECT serialNumber FROM certificate WHERE id=?");
//			pStm.setString(1, id);

			ResultSet generated = pStm.executeQuery();
			if (generated.next())
			{
				if(statsCollector != null)
					statsCollector.logStatRecord(new RedCentSqlImplementation.StatRecord(RedCentSqlImplementation.StatRecord.AC_CREATE_RECORD, certificateType, System.currentTimeMillis()-startTime));
				return new EpointRecord(generated.getLong(1), issuingCause, id, value, null, generated.getTimestamp(2), false, certificateType, templateReference);
			}
			else
				throw new SQLException("Serial number could not be retrieved from INSERT");
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
	}
	
	public EpointRecord createFromResultSet(ResultSet rs) throws SQLException
	{
		if (rs != null && rs.next())
		{
			EpointRecord record = createRecordFromResultSetEntry(rs);

			if(rs.next())
				throw new SQLException("certificate' ID "+record+" is not unique!");
			else
				return record;
		}

		return null;
	}

	private EpointRecord createRecordFromResultSetEntry(ResultSet rs) throws SQLException
	{
		EpointRecord record = new EpointRecord(rs.getLong("serialNumber"), rs.getInt("issuingCause"), rs.getString("id"), rs.getLong("value"), rs.getBytes("cert"), rs.getTimestamp("txDate") ,rs.getBoolean("isValid"), rs.getString("certificateType"),rs.getString("templateReference"));
		return record;
	}

	public List<EpointRecord> createListFromResultSet(ResultSet rs) throws SQLException
	{
		ArrayList<EpointRecord> list = new ArrayList<EpointRecord>();
		
		while (rs != null && rs.next())
		{
			list.add(createRecordFromResultSetEntry(rs));
		}

		return list;
	}

	public EpointRecord validate(Connection con, EpointRecord newRec) throws SQLException
	{
		long startTime = System.currentTimeMillis();
		
		if (newRec != null)
		{
			PreparedStatement pStm = null;
			ResultSet rs = null;
			
			try
			{
				pStm = con.prepareStatement("UPDATE certificate SET cert=?, isValid=? WHERE serialNumber=?");
				pStm.setBytes(1, newRec.getCert());
				pStm.setBoolean(2, true);
				pStm.setLong(3, newRec.getSerialNumber());

				pStm.executeUpdate();
				newRec.setValid(true);

				if(statsCollector != null)
					statsCollector.logStatRecord(new RedCentSqlImplementation.StatRecord(RedCentSqlImplementation.StatRecord.AC_VALIDATE_RECORD, newRec.getCertificateType(), System.currentTimeMillis()-startTime));
			}
			finally
			{
				closeStatementAndResultSet(pStm, rs);
			}
		}
		
		return newRec;
	}

	public EpointRecord invalidate(Connection con, EpointRecord randRec) throws SQLException
	{
		if (randRec != null)
		{
			PreparedStatement pStm = null;
			ResultSet rs = null;
			
			try
			{
				pStm = con.prepareStatement("UPDATE certificate SET isValid=?, id=?, invalidatedId=? WHERE serialNumber=?");
				pStm.setBoolean(1, false);
				pStm.setString(2, null);
				pStm.setString(3, randRec.getId());
				pStm.setLong(4, randRec.getSerialNumber());
				pStm.executeUpdate();
				randRec.setValid(false);
			}
			finally
			{
				closeStatementAndResultSet(pStm, rs);
			}
		}

		return randRec;
	}

	private void closeStatementAndResultSet(Statement stm, ResultSet rs) throws SQLException
	{
		if (rs != null)
			rs.close();
		
		if(stm != null)
			stm.close();
	}

	public void beginTransaction(Connection con) throws SQLException
	{
		if (!con.getAutoCommit())
		{
			log.error("Transaction already in progress!");
			throw new IllegalArgumentException("Transaction already in progress!");
		}
		con.setAutoCommit(false);			
	}

	public void lockWrites(Connection con) throws SQLException
	{
		Statement stm = con.createStatement();
		stm.execute("LOCK TABLE certificate IN SHARE ROW EXCLUSIVE MODE");
// DERBY: Different locking syntax
//		stm.execute("LOCK TABLE certificate IN EXCLUSIVE MODE");
		stm.close();
	}

	public void setTransactionSerializable(Connection con) throws SQLException
	{
		if (con.getAutoCommit())
			beginTransaction(con);
		
		con.setTransactionIsolation(Connection.TRANSACTION_SERIALIZABLE);
	}
	
	public void balanceEmission(Connection con, EpointRecord newRec, String issuer, AuthorizedIssuer ai) throws SQLException
	{
		PreparedStatement pStm = null;
		
		try
		{
			pStm = con.prepareStatement("INSERT INTO balance (serialnumber,issuer,authorizedBy,value) VALUES (?,?,?,?)");
			pStm.setLong(1, newRec.getSerialNumber());
			pStm.setString(2, issuer);
			pStm.setString(3, ai.getFingerprint());
			pStm.setLong(4, newRec.getValue());
			pStm.execute();
			
			balanceAuthorizedIssuerEmission(con, newRec.getValue(), ai.getFingerprint());
		}
		finally
		{
			closeStatementAndResultSet(pStm, null);
		}
	}
	
	public void balanceInvalidation(Connection con, long serialnumber, long valueInvalidated, String issuer, AuthorizedIssuer ai) throws SQLException
	{
		PreparedStatement pStm = null;
		
		try
		{
			pStm = con.prepareStatement("INSERT INTO balance (serialnumber,issuer,authorizedBy,value) VALUES (?,?,?,?)");
			pStm.setLong(1, serialnumber);
			pStm.setString(2, issuer);
			pStm.setString(3, ai.getFingerprint());
			pStm.setLong(4, -1 * Math.abs(valueInvalidated));
			
			pStm.execute();
			
			balanceAuthorizedIssuerInvalidation(con, valueInvalidated, ai.getFingerprint());
		}
		finally
		{
			closeStatementAndResultSet(pStm, null);
		}
	}
	
	public void balanceAuthorizedIssuerEmission(Connection con, long valueEmitted, String authorizedIssuerId) throws SQLException
	{
		PreparedStatement pStm = null;
		
		try
		{
			AuthorizedIssuer ai = findAuthorizedIssuerByFingerprint(con, authorizedIssuerId);
			if (ai == null)
				throw new SQLException("No valid authorized isser found corresponding to the id " + authorizedIssuerId + " you submitted.");

			pStm = con.prepareStatement("UPDATE authorizedIssuer SET balance=? where fingerprint=?");
			pStm.setLong(1, ai.getBalance() + Math.abs(valueEmitted));
			pStm.setString(2, authorizedIssuerId);
			
			pStm.execute();
		}
		finally
		{
			closeStatementAndResultSet(pStm, null);
		}
	}

	public void balanceAuthorizedIssuerInvalidation(Connection con, long valueInvalidated, String authorizedIssuerId) throws SQLException
	{
		PreparedStatement pStm = null;
		
		try
		{
			AuthorizedIssuer ai = findAuthorizedIssuerByFingerprint(con, authorizedIssuerId);
			if (ai == null)
				throw new SQLException("No valid authorized isser found corresponding to the id " + authorizedIssuerId + " you submitted.");

			pStm = con.prepareStatement("UPDATE authorizedIssuer SET balance=? where fingerprint=?");
			pStm.setLong(1, ai.getBalance() - Math.abs(valueInvalidated));
			pStm.setString(2, authorizedIssuerId);
			
			pStm.execute();
		}
		finally
		{
			closeStatementAndResultSet(pStm, null);
		}
	}

	public AuthorizedIssuer findAuthorizedIssuerById(Connection con, String authorizedIssuerId) throws SQLException, NotFoundException
	{
		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			if (authorizedIssuerId != null && authorizedIssuerId.length() == 16 && (Hex.decodeHex(authorizedIssuerId.toLowerCase().toCharArray()) != null))
			{
				authorizedIssuerId = authorizedIssuerId.replaceAll(" ", "").trim().toUpperCase();
				
				pStm = con.prepareStatement("SELECT * from authorizedIssuer WHERE publicKeyIds like ?");			
				pStm.setString(1, "%" + authorizedIssuerId + "%");
				rs = pStm.executeQuery();

				return createAuthorizedIssuerFromResultSet(rs);
			} else
				log.warn("Encountered invalid lookup request for authorized issuer by Id '" + authorizedIssuerId + "'");
		}
		catch (DecoderException e) {
			throw new NotFoundException(e);
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}

		return null;
	}
	
	public AuthorizedIssuer findAuthorizedIssuerByFingerprint(Connection con, String authorizedIssuerId) throws SQLException
	{
		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			if (authorizedIssuerId != null)
			{
				authorizedIssuerId = authorizedIssuerId.replaceAll(" ", "").trim().toUpperCase();
				
				pStm = con.prepareStatement("SELECT * from authorizedIssuer WHERE fingerprint=?");			
				pStm.setString(1, authorizedIssuerId);
				rs = pStm.executeQuery();

				return createAuthorizedIssuerFromResultSet(rs);
			}
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}

		return null;
	}

	public AuthorizedIssuer createAuthorizedIssuer(Connection con, final String authorizedIssuerFingerprint, final String authorizedIssuerPublicKeyBlock, final String authorizedIssuerPublicKeyIds, Long lowerBound, Long upperBound, boolean shallBeNotary, String introducingNotaries, String authorization) throws SQLException {
		PreparedStatement pStm = null;

		try
		{
			if (authorizedIssuerFingerprint != null)
			{
				pStm = con.prepareStatement("INSERT INTO authorizedIssuer " +
						"(fingerprint,publicKey,publickeyids,emissionlimitlowerbound,emissionlimitupperbound," +
						"balance,introducingnotaryids,lastaction,lastactionsauthorization,lastactionslsn" +
						
						(
								shallBeNotary ? ",hasnotaryrights,notaryupdatesignaturetime,notarysincelsn" : ""
						) +
						
						") " +
						"VALUES " +
						"(?,?,?,?,?,?,?,?,?,(SELECT MAX(serialNumber) FROM certificate)" +
						(
								shallBeNotary ? ",?,?,(SELECT MAX(serialNumber) FROM certificate)" : ""
						) +
						")");

				pStm.setString(1, authorizedIssuerFingerprint.replaceAll(" ", "").trim().toUpperCase());
				pStm.setString(2, authorizedIssuerPublicKeyBlock);
				pStm.setString(3, authorizedIssuerPublicKeyIds);
				pStm.setObject(4, lowerBound);
				pStm.setObject(5, upperBound);
				pStm.setLong(6, 0);
				pStm.setString(7, introducingNotaries);
				pStm.setString(8, AuthorizedIssuer.Operations.createAuthorizedIssuer.name());
				pStm.setString(9, authorization);
				
				if (shallBeNotary) {
					pStm.setBoolean(10, shallBeNotary);
					pStm.setLong(11, System.currentTimeMillis());
				}
				
				pStm.execute();
				
				return findAuthorizedIssuerByFingerprint(con, authorizedIssuerFingerprint);
			}
		}
		finally
		{
			closeStatementAndResultSet(pStm, null);
		}

		return null;
	}
	
	public AuthorizedIssuer updateAuthorizedIssuersEmissionLimits(Connection con, AuthorizedIssuer ai, String signature) throws SQLException {
		
		PreparedStatement pStm = null;

		try
		{
			if (ai != null)
			{
				pStm = con.prepareStatement("UPDATE authorizedIssuer SET emissionlimitlowerbound=?, emissionlimitupperbound=?,lastaction=?,lastactionsauthorization=?,lastactionslsn=(SELECT MAX(serialNumber) FROM certificate) WHERE fingerprint=?");

				pStm.setObject(1, ai.getEmissionLimitLowerBound());
				pStm.setObject(2, ai.getEmissionLimitUpperBound());
				pStm.setString(3, AuthorizedIssuer.Operations.updateEmissionLimits.name());
				pStm.setString(4, signature);
				pStm.setString(5, ai.getFingerprint());

				pStm.execute();
				
				return findAuthorizedIssuerByFingerprint(con, ai.getFingerprint());
			}
		}
		finally
		{
			closeStatementAndResultSet(pStm, null);
		}

		return null;
	}
	
	public AuthorizedIssuer updateAuthorizedIssuersPublicKey(Connection con, AuthorizedIssuer ai) throws SQLException {
		
		PreparedStatement pStm = null;

		try
		{
			if (ai != null)
			{
				pStm = con.prepareStatement("UPDATE authorizedIssuer SET publicKey=?, publicKeyIds=?, lastaction=?,lastactionsauthorization=?,lastactionslsn=(SELECT MAX(serialNumber) FROM certificate) WHERE fingerprint=?");

				pStm.setString(1, ai.getPublicKey());
				pStm.setString(2, ai.getPublicKeyIdList());
				pStm.setString(3, AuthorizedIssuer.Operations.updatePublicKey.name());
				pStm.setString(4, "Public key block with positive certification");
				
				//TODO: update notary status
				
				pStm.setString(5, ai.getFingerprint());

				pStm.execute();
				
				return findAuthorizedIssuerByFingerprint(con, ai.getFingerprint());
			}
		}
		finally
		{
			closeStatementAndResultSet(pStm, null);
		}

		return null;
	}
	
	public AuthorizedIssuer updateAuthorizedIssuersNotaryStatus(Connection con, AuthorizedIssuer ai, boolean shallHaveNotaryRights) throws SQLException {
		
		PreparedStatement pStm = null;

		try
		{
			if (ai != null)
			{
				pStm = con.prepareStatement("UPDATE authorizedIssuer SET hasnotaryrights=?, notaryupdatesignaturetime=?, "+ (ai.getNotarySinceLSN() == null && shallHaveNotaryRights ? "notarysincelsn=(SELECT MAX(serialNumber) FROM certificate), " : "") +"lastaction=?, lastactionsauthorization=?,lastactionslsn=(SELECT MAX(serialNumber) FROM certificate) WHERE fingerprint=?");

				pStm.setBoolean(1, shallHaveNotaryRights);
				pStm.setLong(2, System.currentTimeMillis());
				pStm.setString(3, AuthorizedIssuer.Operations.updateNotaryStatus.name());
				pStm.setString(4, "Public key block with "+(shallHaveNotaryRights?"positive":"negative")+" certification");
				pStm.setString(5, ai.getFingerprint());

				pStm.execute();
				
				return findAuthorizedIssuerByFingerprint(con, ai.getFingerprint());
			}
		}
		finally
		{
			closeStatementAndResultSet(pStm, null);
		}

		return null;
	}
	
	private AuthorizedIssuer createAuthorizedIssuerFromResultSet(ResultSet rs) throws SQLException
	{
		
		if (rs != null && rs.next())
		{
			AuthorizedIssuer record = createAuthorizedIssuerFromResultSetEntry(rs);

			if(rs.next())
				throw new SQLException("AuthorizedIssuer's ID in "+record+" is not unique!");
			else
				return record;
		}

		return null;
	}

	private AuthorizedIssuer createAuthorizedIssuerFromResultSetEntry(ResultSet rs) throws SQLException {
		return new AuthorizedIssuer(
				rs.getString("fingerprint"), 
				rs.getString("publicKey"), 
				rs.getString("publicKeyIds"), 
				(Long)rs.getObject("emissionLimitLowerBound"), 
				(Long)rs.getObject("emissionLimitUpperBound"), 
				rs.getLong("balance"), 
				rs.getBoolean("hasNotaryRights"),
				(Long)rs.getObject("notarySinceLSN"), 
				rs.getString("lastAction"), 
				rs.getString("lastActionsAuthorization"), 
				(Long)rs.getObject("lastActionsLSN")
			);
	}
}
