/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.sqlImplementation;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.Connection;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;

import javax.sql.DataSource;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.codec.binary.Hex;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.bouncycastle.openpgp.PGPPublicKey;
import org.bouncycastle.openpgp.PGPPublicKeyRing;
import org.bouncycastle.openpgp.PGPSignature;
import org.springframework.beans.factory.annotation.Autowired;

import de.nachtlicht_media.epoint.redcent.AuthorizedIssuer;
import de.nachtlicht_media.epoint.redcent.EpointRecord;
import de.nachtlicht_media.epoint.redcent.RedCent;
import de.nachtlicht_media.epoint.redcent.crypto.BouncycastleUtil;
import de.nachtlicht_media.epoint.redcent.exception.BadsignatureException;
import de.nachtlicht_media.epoint.redcent.exception.ConflictException;
import de.nachtlicht_media.epoint.redcent.exception.DatabaseErrorException;
import de.nachtlicht_media.epoint.redcent.exception.ForbiddenExistingException;
import de.nachtlicht_media.epoint.redcent.exception.ForbiddenSelfException;
import de.nachtlicht_media.epoint.redcent.exception.InvalidTransactionError;
import de.nachtlicht_media.epoint.redcent.exception.MalformedRequestException;
import de.nachtlicht_media.epoint.redcent.exception.NotFoundException;
import de.nachtlicht_media.epoint.redcent.exception.RedcentProcessingException;
import de.nachtlicht_media.epoint.redcent.interfaces.EpointCertificateTypes;
import de.nachtlicht_media.epoint.redcent.interfaces.MessageDecryptionVerificationHandler;
import de.nachtlicht_media.epoint.redcent.interfaces.RedCentConfig;
import de.nachtlicht_media.epoint.redcent.messaging.StringDataMap;
import de.nachtlicht_media.epoint.redcent.util.Splitter;
import de.nachtlicht_media.epoint.redcent.util.StringUtils;

/**
 * The main RedCent implementation based on a postgresql database.
 * @author js
 *
 */
public class RedCentSqlImplementation extends RedCent
{
	public static class StatRecord
	{
		public static final String AC_APP_EXCHANGE = "APP_EXCHANGE";

		public static final String AC_CREATE_RECORD = "CREATE_RECORD";
		public static final String AC_FIND_VALID_RECORD = "FIND_VALID_RECORD";
		public static final String AC_FIND_INVALID_RECORD = "FIND_INVALID_RECORD";
		public static final String AC_FIND_OBLIGATION_HISTORY = "AC_FIND_LEDGER_HISTORY";
		public static final String AC_FIND_OBLIGATION_BEFORE = "AC_FIND_OBLIGATION_BEFORE";
		public static final String AC_VALIDATE_RECORD = "VALIDATE_RECORD";
		public static final String AC_COMMIT = "COMMIT";
		public static final String AC_LOCK = "LOCK";
		
		public String action;
		public String type;
		public long duration;
		
		public StatRecord(String action, String type, long duration)
		{
			super();
			this.action = action;
			this.type = type;
			this.duration = duration;
		}
		
		public String getData()
		{
			return action+";"+type+";"+duration;
		}
	}
		
	@Autowired(required=false)
	private StatsCollector statsCollector;
	public static class StatsCollector extends ArrayList
	{
		public void logStatRecord(RedCentSqlImplementation.StatRecord rec)
		{
			if (rec != null)
				add(rec);
		}
	}

	private static final Log log = LogFactory.getLog(RedCentSqlImplementation.class);
	
	public RedCentSqlImplementation(RedCentConfig config)
	{
		super(config);
	}
	
	@Autowired
	private DataSource  dataSource  = null;	
	
	@Autowired
	private EpointSQLDB epointDB;
	protected EpointSQLDB getEpointDB() throws DatabaseErrorException
	{
		return epointDB;
	}
	
	protected ArrayList getDBStats() throws DatabaseErrorException
	{
		return statsCollector;
	}
	
	public boolean testDataSource() throws Exception
	{
		if (dataSource != null)
		{
			Connection c = getConnection();
			
			if (c == null)
				return false;
			else
			{
				log.info("Datasource bound, proceeding to test database integrity.");
				c.close();
			}
		} else
			return false;
		
		log.info("Testing EpointSQLDB...");
		Connection con = null;
		try
		{
			con = getConnection();
			getEpointDB().beginTransaction(con);
			getEpointDB().lockWrites(con);
			
			getEpointDB().verifyDatabase(con);
		}
		catch (SQLException ex)
		{
			throw rollbackAndReturnCause(con, new DatabaseErrorException(ex));
		}
		finally
		{
			closeConnection(con);
		}
		log.info("Testing EpointSQLDB succeeded.");
		
		return true;
	}
	
	public byte[] getServerStatus() throws DatabaseErrorException, ConflictException
	{
		Connection con = null;
		try
		{
			log.trace("Getting connection");
			con = getConnection();
			log.trace("got connection " + con);
			
			getEpointDB().beginTransaction(con);
			getEpointDB().setTransactionSerializable(con);
			
			// extract data from session
			StringDataMap data = new StringDataMap();
			data.put("LSN", getPaddedSerialNumber(getEpointDB().getLatestSerial(con)));
			data.put("SUM", getEpointDB().getTotalAmount(con).toString());
	
			// all transactional data is gathered, now add runtime information
			con.commit();
			
			data.put("NUM", (Runtime.getRuntime().freeMemory() / 1024) + " kb");
			data.put("DEN", (Runtime.getRuntime().maxMemory() / 1000) + " kb");
			data.put("PER", new Double((double)Runtime.getRuntime().freeMemory() / (double)Runtime.getRuntime().maxMemory() * 100.0d));
			
			return signMessage(getMessageHandler().createMessage(EpointCertificateTypes.SERVER_STATUS, data));
		}
		catch (SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
		finally
		{
			closeConnection(con);
		}
	}
	
	public byte[] getExchangeRateInformation() throws DatabaseErrorException, ConflictException
	{
		Connection con = null;
		try
		{
			log.trace("Getting connection");
			con = getConnection();
			log.trace("got connection " + con);
			
			getEpointDB().beginTransaction(con);
			getEpointDB().setTransactionSerializable(con);
			
			// extract data from session
			StringDataMap data = new StringDataMap();
			data.put("LSN", getPaddedSerialNumber(getEpointDB().getLatestSerial(con)));
			data.put("SUM", getEpointDB().getTotalAmount(con).toString());
			data.put("LXRNSN", getPaddedSerialNumber(new Long(StringUtils.parsePrimitiveLongFromString(getDocumentStore().getMostRecentDocumentId(DOCUMENT_REALM_EXCHANGE_RATE_NOTIFICATIONS)))));
			
			// all transactional data is gathered, now add runtime information
			con.commit();
			
			data.put("NUM", (Runtime.getRuntime().freeMemory() / 1024) + " kb");
			data.put("DEN", (Runtime.getRuntime().maxMemory() / 1000) + " kb");
			data.put("PER", new Double((double)Runtime.getRuntime().freeMemory() / (double)Runtime.getRuntime().maxMemory() * 100.0d));
			
			return signMessage(getMessageHandler().createMessage(EpointCertificateTypes.EXCHANGE_RATE_INFORMATION, data));
		}
		catch (SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
		catch (Exception ex)
		{
			throw new ConflictException(ex);
		}
		finally
		{
			closeConnection(con);
		}
	}

	public byte[] getServerBalance(final String key) throws DatabaseErrorException, ConflictException
	{
		Connection con = null;
		try
		{
			log.trace("Getting connection");
			con = getConnection();
			log.trace("got connection " + con);

			getEpointDB().beginTransaction(con);
			getEpointDB().setTransactionSerializable(con);

			// extract data from session
			StringDataMap data = new StringDataMap();
			data.put("LSN", getPaddedSerialNumber(getEpointDB().getLatestSerial(con)));
			data.put("SUM", getEpointDB().getTotalAmount(con).toString());
			data.put("BALANCE", getEpointDB().getBalanceInfo(con, key).toString());
			data.put("KEY", key!=null?key:"all authorized keys");

			// all transactional data is gathered, now add runtime information
			con.commit();

			byte[] sMsg = signMessage(getMessageHandler().createMessage(EpointCertificateTypes.SERVER_BALANCE, data));

			return sMsg;
		}
		catch (SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
		finally
		{
			closeConnection(con);
		}
	}
	
	public byte[] getSignedLedger(final String key, final Long start, final Long end) throws RedcentProcessingException
	{
		Connection con = null;
		try
		{
			log.trace("Getting connection");
			con = getConnection();
			log.trace("got connection " + con);

			getEpointDB().beginTransaction(con);
			getEpointDB().setTransactionSerializable(con);

			if (key != null) {
				
				String searchKey = key.trim().replaceAll(" ", "").toUpperCase();
				
				EpointRecord currentObligation = findValidRecordByID(searchKey);

				if (currentObligation != null) {

					List<EpointRecord> obligationHistory = getEpointDB().findObligationHistoryByID(con, searchKey, start, end);

					if (obligationHistory != null && obligationHistory.size() > 0) {

						EpointRecord openingRecord = getEpointDB().findLastObligationBeforeSerial(con, searchKey, obligationHistory.get(0).getSerialNumber());
						
						// extract data from session
						StringDataMap data = new StringDataMap();
						data.put("ID", Splitter.group(searchKey, 4, 5));
						data.put("ISSUER", getIssuerFingerPrintText());

						data.put("START_DATE", obligationHistory.get(0).getTxDate());
						data.put("END_DATE", obligationHistory.get(obligationHistory.size() - 1).getTxDate());

						long openingBalance = openingRecord != null ? openingRecord.getValue() : 0;
						data.put("OPENING_BALANCE", openingBalance);
						data.put("CLOSING_BALANCE", obligationHistory.get(obligationHistory.size() - 1).getValue());
						data.put("COSN", getPaddedSerialNumber(currentObligation.getSerialNumber()));

						data.put("DATE", new java.util.Date());
						data.put("LSN", getPaddedSerialNumber(getEpointDB().getLatestSerial(con)));

						StringBuffer ledgerBuffer = new StringBuffer();

						final String separator = "    ";
						
						SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss z");
						formatter.setTimeZone(TimeZone.getTimeZone("UTC"));
						
						Iterator<EpointRecord> iter = obligationHistory.iterator();
						
						EpointRecord previousRecord = openingRecord;
						
						while(iter.hasNext()) {
							
							EpointRecord rec = iter.next();
							Long txValue = rec.getValue() - (previousRecord != null ? previousRecord.getValue() : 0);
							
							// TODO: replace with TX-date
							ledgerBuffer.append(formatter.format(rec.getTxDate()) + separator + getPaddedSerialNumber(rec.getSerialNumber()) + separator + (txValue < 0 ? txValue : "+" + txValue) + separator + /*getLeftPaddedNumber(*/rec.getValue()/*, ' ')*/ + "\n");
							
							previousRecord = rec;
						}

						data.put("LEDGER", ledgerBuffer.toString());

						// all transactional data is gathered, now add runtime information
						con.commit();

						byte[] sMsg = signMessage(getMessageHandler().createMessage(EpointCertificateTypes.LEDGER, data));

						return sMsg;
					} else {
						con.rollback();
						return null;
					}
				} else { 
					con.rollback();
					return null;
				}
			} else {
				con.rollback();
				return null;
			}
		}
		catch (SQLException ex)
		{
			throw rollbackAndReturnCause(con, new DatabaseErrorException(ex));
		}
		catch (RedcentProcessingException ex)
		{
			throw rollbackAndReturnCause(con, ex);
		}
		finally
		{
			closeConnection(con);
		}
	}

	public byte[] exchange(int resultType, String rand, String targetID, String userNotes) throws RedcentProcessingException
	{
		long startTime = System.currentTimeMillis();
		
		String existingID;
		try {
			existingID = calcIdFromSecret(rand);
		} catch (NoSuchAlgorithmException ex) {
			throw new ConflictException(ex);
		}
		targetID = targetID.toUpperCase();
		
		if (existingID.equalsIgnoreCase(targetID))
			throw new ForbiddenSelfException("You must not exchange an ePoint rand with itself.");
		
		targetID = checkTargetDigest("Target", targetID);
		
		checkResultType(resultType);
		
		Connection con = null;
		try
		{
			con = getConnection();
			getEpointDB().beginTransaction(con);
			long lockStartTime = System.currentTimeMillis();
			getEpointDB().lockWrites(con);
			if (statsCollector != null)
				statsCollector.logStatRecord(new StatRecord(StatRecord.AC_LOCK, "EXCHANGE", System.currentTimeMillis() - lockStartTime));


			EpointRecord randRec = getEpointDB().findValidRecordByID(con, existingID);

			if (randRec != null)
			{
				long amount = randRec.getValue();

				EpointRecord existingTargetRec = getEpointDB().findValidRecordByID(con, targetID);
				Collection existingInvalidTargetRecs = getEpointDB().findInvalidRecordsByID(con, targetID);

				if (existingTargetRec != null)
				{
					// this is a merge operation
					amount += existingTargetRec.getValue();
					getEpointDB().invalidate(con, existingTargetRec);
				}
				else
				{
					// this is a plain exchange operation. Prevent users from assigning ePoints to already
					// spent rands.
					
					if (existingInvalidTargetRecs.size() > 0)
					{
						con.rollback();
						throw new ForbiddenExistingException("An invalidated ePoint rand with digest " + targetID + " already exists.");
					}
				}

				EpointRecord newRec = createEpointRecord(con, amount, targetID, EpointRecord.EXCHANGE, EpointCertificateTypes.EPOINT_CERTIFICATE);

				String message = createCertMessage(amount, newRec, ((existingTargetRec != null)?"#"+getPaddedSerialNumber(new Long(existingTargetRec.getSerialNumber()))+"\n":"") + "+" + rand, userNotes);

				byte[] newCert = signMessage(message);

				log.trace(newCert);
				newRec.setCert(newCert);
				newRec = getEpointDB().validate(con, newRec);
				randRec = getEpointDB().invalidate(con, randRec);

				long commitStartTime = System.currentTimeMillis();
				con.commit();
				if (statsCollector != null)
					statsCollector.logStatRecord(new StatRecord(StatRecord.AC_COMMIT, "EXCHANGE", System.currentTimeMillis() - commitStartTime));
				
				logTransaction(newRec);
				
				if (statsCollector != null)
					statsCollector.logStatRecord(new StatRecord(StatRecord.AC_APP_EXCHANGE, (existingTargetRec==null?"EXCHANGE":"MERGE"), System.currentTimeMillis() - startTime));

				if (resultType == RedCent.RETURN_RESULTING_CERT)
				{
					return newCert;
				}
				else if (resultType == RedCent.RETURN_TRANSACTION_RECEIPT)
				{
					return signBinaryContent(createTransactionReceipt(randRec.getValue(), targetID), false);
				}
				else if (resultType == RedCent.RETURN_TRANSACTIONVALUE_STRING_AS_BYTEARRAY_UTF8)
				{
					return String.valueOf(randRec.getValue()).getBytes("UTF-8");
				}
				else
				{
					// This else clause should be unreachable as checkResultType should abort the
					// operation prior database access if the the return type is illegal
					return newCert;
				}
			}
			else
				throw new NotFoundException("No valid ePoint certificate for submitted rand could be found.");
		}
		catch (SQLException ex)
		{
			throw rollbackAndReturnCause(con, new DatabaseErrorException(ex));
		}
		catch (UnsupportedEncodingException ex)
		{
			throw rollbackAndReturnCause(con, new ConflictException(ex));
		}
		finally
		{
			closeConnection(con);
		}
	}

	private void checkResultType(int resultType) throws MalformedRequestException
	{
		if (		resultType != RedCent.RETURN_RESULTING_CERT 
				&& 	resultType != RedCent.RETURN_TRANSACTION_RECEIPT
				&&	resultType != RedCent.RETURN_TRANSACTIONVALUE_STRING_AS_BYTEARRAY_UTF8)
		{
			throw new MalformedRequestException("Illegal return type ("+resultType+") specified for command.");
		}
	}

	protected void logTransaction(EpointRecord newRec)
	{
		try
		{
			if (xLog != null)
			{
				if (xLog.isInfoEnabled())
					xLog.info(getXLogConverter().convert(newRec));
			}
			else
				log.warn("Transaction logger not initialized!");
		}
		catch (Exception e)
		{
			log.error("Could not write to transaction log.", e);
		}
	}

	public byte[] emit(String targetID, long amount, String signature, String authorizedBy, String targetIssuer, String userNotes) throws RedcentProcessingException
	{
		if (amount <= 0)
			throw new MalformedRequestException("Amount to emit must not be less or equal to zero.");
		
		targetID = checkTargetDigest("Target", targetID);

		// verify that emission was originally intended for this issuer (and not another one, one is holding emission rights with, too)
		targetIssuer = checkTargetIssuer(targetIssuer);
		
		Connection con = null;
		try
		{
			targetID = targetID.toUpperCase();
			
			con = getConnection();
			getEpointDB().beginTransaction(con);
			getEpointDB().lockWrites(con);
			
			EpointRecord existingValidTargetRec = getEpointDB().findValidRecordByID(con,targetID);
			Collection existingInvalidTargetRecs = getEpointDB().findInvalidRecordsByID(con,targetID);

			// check for invalid records with this ID, too
			if (existingValidTargetRec != null || existingInvalidTargetRecs.size() > 0)
				throw new ForbiddenExistingException("You must not emit ePoint with targetDigest already existing");
			
			// Verify the AI
			AuthorizedIssuer ai = getEpointDB().findAuthorizedIssuerByFingerprint(con,authorizedBy);
			if (ai == null)
				throw new NotFoundException("No valid issuer to emit epoints for exists corresponding to the id '"+authorizedBy+"' you submitted.");
			else {
				// verify that this emission won't violate the AI's emission limits
				ai.checkEmitEpointAmount(amount);
			}

			EpointRecord newRec = createEpointRecord(con, amount, targetID, EpointRecord.EMIT, EpointCertificateTypes.EPOINT_CERTIFICATE);

			String message = createCertMessage(amount, newRec, signature, userNotes);

			byte[] newCert = signMessage(message);

			log.trace(newCert);
			newRec.setCert(newCert);
			newRec = getEpointDB().validate(con,newRec);
			getEpointDB().balanceEmission(con, newRec, getIssuerFingerPrintText().replaceAll(" ","") ,ai);
			
			con.commit();
			
			logTransaction(newRec);
			
			return newCert;
		}
		catch (SQLException ex)
		{
			throw rollbackAndReturnCause(con, new DatabaseErrorException(ex));
		}
		catch (RedcentProcessingException ex)
		{
			throw rollbackAndReturnCause(con, ex);
		}
		finally
		{
			closeConnection(con);
		}
	}

	public byte[] invalidate(String rand, String signature, String authorizedIssuerId, String targetIssuer, String userNotes) throws RedcentProcessingException
	{
		// verify that invalidation was originally intended for this issuer (and not for another one, one is holding invalidation rights with, too)
		targetIssuer = checkTargetIssuer(targetIssuer);

		Connection con = null;
		try
		{
			String idToInvalidate = calcIdFromSecret(rand);
			
			con = getConnection();
			getEpointDB().beginTransaction(con);
			getEpointDB().lockWrites(con);
			
			EpointRecord existingValidRecord = getEpointDB().findValidRecordByID(con,idToInvalidate);
			if (existingValidRecord == null)
				throw new NotFoundException("No valid certificate exists corresponding to the rand you submitted.");

			// Verify the AI
			AuthorizedIssuer ai = getEpointDB().findAuthorizedIssuerByFingerprint(con,authorizedIssuerId);
			if (ai == null)
				throw new NotFoundException("No valid issuer to destroy epoints for exists corresponding to the id you submitted.");
			else {
				// verify that this invalidation won't violate the AI's emission limits
				ai.checkDestroyEpointAmount(existingValidRecord.getValue());
			}
			
			getEpointDB().invalidate(con, existingValidRecord);

			// attempt to create the new record (optionally containing a signature)
			EpointRecord newRec = createEpointRecord(con, existingValidRecord.getValue(), idToInvalidate, EpointRecord.INVALIDATION, EpointCertificateTypes.INVALIDATION_CERTIFICATE);
			String message = createInvalidationMessage(newRec, idToInvalidate, "+" + rand + (signature != null ? "\n" + signature : ""), userNotes);

			byte[] newCert = signMessage(message);

			log.trace(newCert);
			newRec.setCert(newCert);

			// make the new record valid to correctly invalidate the former record
			newRec = getEpointDB().validate(con,newRec);

			// and invalidate it
			newRec = getEpointDB().invalidate(con,newRec);
			getEpointDB().balanceInvalidation(con, newRec.getSerialNumber(), existingValidRecord.getValue(), getIssuerFingerPrintText().replaceAll(" ",""), ai);

			con.commit();

			logTransaction(newRec);

			return newCert;
		}
		catch (SQLException ex)
		{
			throw rollbackAndReturnCause(con, new DatabaseErrorException(ex));
		}
		catch (RedcentProcessingException ex)
		{
			throw rollbackAndReturnCause(con, ex);
		}
		catch (NoSuchAlgorithmException ex)
		{
			throw new ConflictException(ex);
		}
		finally
		{
			closeConnection(con);
		}
	}

	public byte[] split(String rand, String targetID, long amount, String restID, String userNotes) throws RedcentProcessingException
	{
		if (amount <= 0)
			throw new MalformedRequestException("Amount to split must not be less or equal to zero.");
		
		targetID = targetID.toUpperCase();

		targetID = checkTargetDigest("Target", targetID);

		restID = checkTargetDigest("Rest", restID);
		
		Connection con = null;
		try
		{
			con = getConnection();
			getEpointDB().beginTransaction(con);
			getEpointDB().lockWrites(con);
			
			EpointRecord randRec = getEpointDB().findValidRecordByID(con, calcIdFromSecret(rand));

			EpointRecord existingValidTargetRec = getEpointDB().findValidRecordByID(con,targetID);
			Collection existingInvalidTargetRecs = getEpointDB().findInvalidRecordsByID(con,targetID);

			// check for invalid records with this ID, too
			if (existingValidTargetRec != null || existingInvalidTargetRecs.size() > 0)
				throw new ForbiddenExistingException("You must not split an ePoint using a targetDigest is already existing or spent.");
			
			if (randRec != null)
			{
				if(randRec.getValue() > amount)
				{
					EpointRecord targetRec = createEpointRecord(con, amount, targetID, EpointRecord.SPLIT, EpointCertificateTypes.EPOINT_CERTIFICATE);
					EpointRecord restRec = createEpointRecord(con, randRec.getValue() - amount, restID, EpointRecord.SPLIT, EpointCertificateTypes.EPOINT_CERTIFICATE);

					String targetMessage = createCertMessage(targetRec.getValue(), targetRec, "+" + rand + "\n -" + restRec.getId(), userNotes);
					byte[] targetCert = signMessage(targetMessage);
					targetRec.setCert(targetCert);

					String restMessage = createCertMessage(restRec.getValue(), restRec, "+" + rand + "\n -" + targetRec.getId(), userNotes);
					byte[] restCert = signMessage(restMessage);
					restRec.setCert(restCert);

					getEpointDB().validate(con, targetRec);
					getEpointDB().validate(con, restRec);
					getEpointDB().invalidate(con, randRec);

					con.commit();
					
					logTransaction(targetRec);
					logTransaction(restRec);
					
					return restCert;
				}
				else
					throw new InvalidTransactionError("Can't split record into parts that are more valuable than the record itself! Nice try!");
			}
			else
				throw new NotFoundException("No valid certificate to split exists corresponding to the rand you submitted.");
		}
		catch (SQLException ex)
		{
			throw rollbackAndReturnCause(con, new DatabaseErrorException(ex));
		}
		catch (RedcentProcessingException ex)
		{
			throw rollbackAndReturnCause(con, ex);
		}
		catch (NoSuchAlgorithmException ex)
		{
			throw new ConflictException(ex);
		}
		finally
		{
			closeConnection(con);
		}
	}

	public byte[] mergeObligation(int resultType, String rand, String key, String userNotes) throws RedcentProcessingException
	{
		if (!ENABLE_OBLIGATION_AND_DRAFT_PROCESSING)
			throw new RedcentProcessingException("This build of RedCent does not allow Obligation and Draft processing.")
			{
				public int getType()
				{
					return FORBIDDEN;
				}
			};
			
		checkResultType(resultType);
		
		Connection con = null;
		try
		{
			key = checkKeyFingerprint(key);

			con = getConnection();
			getEpointDB().beginTransaction(con);
			getEpointDB().lockWrites(con);
			
			EpointRecord randRec = getEpointDB().findValidRecordByID(con, calcIdFromSecret(rand));
			EpointRecord obligationRec = getEpointDB().findValidRecordByID(con, key);

			if (randRec == null)
				throw new NotFoundException("No valid certificate exists corresponding to the rand you submitted.");
			
			long difference = randRec.getValue();
			
			// calculate denomination of resulting obligations certificate
			long value = randRec.getValue();

			if (obligationRec != null)
				value += obligationRec.getValue();
			// value is now the resulting denomination

			// invalidate any possibly existing obligations certificate with the same key and
			// invalidate the rand sent in
			getEpointDB().invalidate(con, randRec);
			if (obligationRec != null)
				getEpointDB().invalidate(con, obligationRec);

			// create the resulting obligations certificate
			EpointRecord targetRec = createEpointRecord(con, value, key, EpointRecord.OBLIGATIONS_MERGE, EpointCertificateTypes.OBLIGATION_CERTIFICATE);
			String targetMessage = createObligationCertMessage(value, targetRec, difference,(obligationRec != null ? getPaddedSerialNumber(obligationRec.getSerialNumber()) + "\n" : "") + "+" + rand, userNotes);
			byte[] targetCert = signMessage(targetMessage);
			targetRec.setCert(targetCert);


			// make the resulting obligations certificate valid
			getEpointDB().validate(con, targetRec);

			con.commit();
			
			logTransaction(targetRec);

			if (resultType == RedCent.RETURN_RESULTING_CERT)
			{
				return targetCert;
			}
			else if (resultType == RedCent.RETURN_TRANSACTION_RECEIPT)
			{
				// #22 add a zero byte to the key's fingerprint to make clients able to distinguish between an ePoint and an obligation's transaction receipt
				return signBinaryContent(createTransactionReceipt(randRec.getValue(), key + "00"), false);
			}
			else if (resultType == RedCent.RETURN_TRANSACTIONVALUE_STRING_AS_BYTEARRAY_UTF8)
			{
				return String.valueOf(randRec.getValue()).getBytes("UTF-8");
			}
			else
				return targetCert;
		}
		catch (SQLException ex)
		{
			throw rollbackAndReturnCause(con, new DatabaseErrorException(ex));
		}
		catch (RedcentProcessingException ex)
		{
			throw rollbackAndReturnCause(con, ex);
		}
		catch (UnsupportedEncodingException e)
		{
			throw new ConflictException("Error converting transaction value to String", e);
		}
		catch (NoSuchAlgorithmException ex)
		{
			throw new ConflictException(ex);
		}
		finally
		{
			closeConnection(con);
		}
	}

	
	public byte[] processDraftMessage(MessageDecryptionVerificationHandler signedMessage) throws RedcentProcessingException
	{
		if (!ENABLE_OBLIGATION_AND_DRAFT_PROCESSING)
			throw new RedcentProcessingException("This build of RedCent does not allow Obligation and Draft processing.")
			{
				public int getType()
				{
					return FORBIDDEN;
				}
			};
			
		if (signedMessage != null)
		{
			String clearText = signedMessage.getClearText();
			String signer = signedMessage.getSignersFingerprint();
			
			if (signedMessage.isClearText().booleanValue())
			{
				Map data = getMessageHandler().parseMessage(clearText);
				
				if (data != null && data.size() != 0)
				{
					String targetDigest = ((String)data.get("ID")).toUpperCase();
					long amount = Long.parseLong((String)data.get("VALUE"));

					String drawer = ((String)data.get("DRAWER")).toUpperCase();
					String drawee = ((String)data.get("DRAWEE")).toUpperCase();
					
					if (drawer == null || drawee == null)
						throw new MalformedRequestException("A draft has to contain drawer and drawee values!");
					
					// verify that this draft was issued for this issuer, making this issuer the drawee
					if(!getIssuerFingerPrintText().replaceAll(" ","").equals(drawee.replaceAll(" ","")))
						throw new MalformedRequestException("The draft you submitted was pulled for issuer " + drawee + " and not the issuer you submitted it to: " + getIssuerFingerPrintText());

					// verify that drawer and signer are equivalent or signer is a subkey of drawer
					if (signer.equals(drawer))
						log.debug("The draft's drawer and the signer are equal ("+drawer+"), proceeding splitObligation.");
					else if (drawer.equals(signedMessage.getSignersPrimaryKeyFingerprint())) {
						// make sure that signer is a subkey of drawer, otherwise throw Exception
						log.debug("The draft's drawer ("+drawer+") and the signer ("+signer+") are not equal, a subkey of the masterkey was used to sign the draft, proceeding splitObligation.");
					} else
						throw new MalformedRequestException("The draft's drawer ("+drawer+") and the signer ("+signer+") are not equal and a relation between the two cannot be resolved. Draft rejected.");

					return splitObligation(targetDigest, amount, drawer, signedMessage.getMessage(), null);
				}
				else
					throw new MalformedRequestException("Draft message could not be parsed");
			}
			else
				throw new BadsignatureException("Drafts have to be clearsigned.");
		}
		else
			throw new MalformedRequestException("Illegal call to processDraftMessage with empty argument");
	}

	public byte[] splitObligation(String targetID, final long amount, String key, byte[] signedMessage, String userNotes) throws RedcentProcessingException
	{
		if (!ENABLE_OBLIGATION_AND_DRAFT_PROCESSING)
			throw new RedcentProcessingException("This build of RedCent does not allow Obligation and Draft processing.")
			{
				public int getType()
				{
					return FORBIDDEN;
				}
			};
			
		if (amount < 0)
			throw new MalformedRequestException("You must not assign an ePoint with a negative amount");
		
		targetID = targetID.toUpperCase();

		targetID = checkTargetDigest("Target", targetID);
		
		key = key.toUpperCase();
		
		Connection con = null;
		try
		{
			String signedDraft = new String(signedMessage, "UTF-8");	
			
			con = getConnection();
			getEpointDB().beginTransaction(con);
			getEpointDB().lockWrites(con);
	
			
			EpointRecord existingValidObligationRecord = getEpointDB().findValidRecordByID(con,key);
	
			if (existingValidObligationRecord != null && existingValidObligationRecord.getValue() >= amount)
			{
				EpointRecord existingValidTargetRec = getEpointDB().findValidRecordByID(con,targetID);
				Collection existingInvalidTargetRecs = getEpointDB().findInvalidRecordsByID(con,targetID);
	
				// check for invalid records with this ID, too
				if (existingInvalidTargetRecs.size() > 0)
					throw new ForbiddenExistingException("You must not emit an ePoint with an ID already spent");
				
				if (existingValidTargetRec != null)
				{
					if (isTransActionVulnerableToReplayAttacks(existingValidTargetRec))
					{
						//TODO: if it exists another invalid draft that has the same MD but which is different from the current
						// one, the draft's signer must be accused of issuing drafts twice
						throw new ForbiddenExistingException("An obligations split operation with the same draft was excercised before. Drafts must not be executed more than once.");
					}
					else
					{
						// it exists already a normal ePoint record with the same MD. In this case the the new record that 
						// is split off the obligation must be merged with the existing record.
					}
				}
				
				// create additionalAmount and notes, if an existing record with the targetDigest exists and shall
				// be merged
				long additionalMergeAmount = existingValidTargetRec != null ? existingValidTargetRec.getValue() : 0l;
				String issuerAdditionalNotes = getPaddedSerialNumber(existingValidObligationRecord.getSerialNumber()) + (existingValidTargetRec != null ? ("\n+" + targetID) : "")  + "\n" + signedDraft;
				
				if(existingValidTargetRec != null)
					getEpointDB().invalidate(con, existingValidTargetRec);
				
				EpointRecord newRec = createEpointRecord(con, amount + additionalMergeAmount, targetID, EpointRecord.OBLIGATIONS_SPLIT, EpointCertificateTypes.EPOINT_CERTIFICATE);
	
				String message = createCertMessage(amount, newRec, issuerAdditionalNotes, userNotes);
	
				byte[] newCert = signMessage(message);
	
				log.trace(newCert);
				newRec.setCert(newCert);
				newRec = getEpointDB().validate(con,newRec);
	
				// decrease obligation by amount
				EpointRecord existingObligation = getEpointDB().invalidate(con, existingValidObligationRecord);
				
				long value = existingObligation.getValue() - amount;
				
				EpointRecord newObligation = createEpointRecord(con, value, existingObligation.getId(), EpointRecord.OBLIGATIONS_SPLIT, EpointCertificateTypes.OBLIGATION_CERTIFICATE);
				
				String targetMessage = createObligationCertMessage(value, newObligation, -1l*amount,getPaddedSerialNumber(existingValidObligationRecord.getSerialNumber()) + "\n -" + targetID + "\n" + signedDraft, userNotes);
				byte[] targetCert = signMessage(targetMessage);
				newObligation.setCert(targetCert);
				
				getEpointDB().validate(con, newObligation);
				
				con.commit();
	
				logTransaction(newRec);
				logTransaction(newObligation);
				
				return newCert;
			}
			else
			{
				//TODO: punish key KEY for assigning unbacked drafts
				con.rollback();
				throw new NotFoundException("NO valid obligation with amount >= " + amount + " could be found for " + key);
			}
		}
		catch (SQLException ex)
		{
			throw rollbackAndReturnCause(con, new DatabaseErrorException(ex));
		}
		catch (RedcentProcessingException ex)
		{
			throw rollbackAndReturnCause(con, ex);
		}
		catch (UnsupportedEncodingException e)
		{
			throw new ConflictException("Error converting transaction value to String", e);
		}
		finally
		{
			closeConnection(con);
		}
	}

	/**
	 * This methods returns 
	 */
	public EpointRecord findValidRecordByID(String id) throws NotFoundException, DatabaseErrorException
	{
		/*
		 * No transactions are needed inside this method since no modifications to the database
		 * are performed.
		 */
		Connection con = null;
		try
		{
			if (id != null)
			{
				id = id.toUpperCase();
				
				con = getConnection();
				EpointRecord result = getEpointDB().findValidRecordByID(con , id);

				if (result != null)
					return result;
			}
			throw new NotFoundException("Could not find any certificate matching " + id);
		}
		catch (SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
		finally
		{
			closeConnection(con);
		}
	}

	protected Connection getConnection() throws SQLException, DatabaseErrorException
	{
		Connection con = dataSource.getConnection();
		
		if (!con.isClosed())
			return con;
		else
			return dataSource.getConnection();
	}

	protected void closeConnection(Connection con) throws DatabaseErrorException
	{
		try
		{
			if (con != null && !con.isClosed()) {
				con.close();
			}
		}
		catch (SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
	}
	
	protected void rollback(Connection con) throws RedcentProcessingException
	{
		try
		{
			if (con != null)
				con.rollback();
		}
		catch (SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
	}
	
	protected RedcentProcessingException rollbackAndReturnCause(Connection con, RedcentProcessingException cause)
	{
		try
		{
			if (con != null)
				con.rollback();
			
			if (cause != null)
				return cause;
		}
		catch (SQLException ex)
		{
			log.error("Error rolling back transaction!", ex);
			if (cause != null)
				return cause;
			else
				return new DatabaseErrorException(ex);
		}
		
		if (cause == null)
			throw new IllegalArgumentException("rollbackAndThrow with empty cause!");
		
		return cause;
	}

	protected EpointRecord createEpointRecord(Connection con, long amount, String targetID, int issuingCause, String recordType) throws DatabaseErrorException, ConflictException, SQLException
	{
		return getEpointDB().createRecord(con, amount, targetID, issuingCause, recordType, getTemplateManager().getHashValueForCurrentVersion(recordType));
	}
	
	public EpointRecord findBySerial(long serial) throws NotFoundException, DatabaseErrorException
	{
		Connection con = null;
		try
		{
			con = getConnection();
			EpointSQLDB db = getEpointDB();
			EpointRecord result = db.findBySerial(con, serial);

			if (result != null)
				return result;

			throw new NotFoundException("A valid certificate for serial " + serial + " could not be found");
		}
		catch (SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
		finally
		{
			closeConnection(con);
		}
	}

	@Override
	public AuthorizedIssuer findAuthorizedIssuerByFingerprint(String authorizedIssuerFingerprint) throws NotFoundException, DatabaseErrorException
	{
		log.debug("findAuthorizedIssuerByFingerprint: " + authorizedIssuerFingerprint);
		
		Connection con = null;
		try
		{
			con = getConnection();
			return getEpointDB().findAuthorizedIssuerByFingerprint(con, authorizedIssuerFingerprint);
		}
		catch (SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
		finally
		{
			closeConnection(con);
		}
	}

	@Override
	public AuthorizedIssuer findAuthorizedIssuerByKeyID(long keyID) throws NotFoundException, DatabaseErrorException
	{
		String authorizedIssuerId = BouncycastleUtil.convertKeyIDToHex(keyID);
		
		log.debug("findAuthorizedIssuerById: " + authorizedIssuerId);
		
		Connection con = null;
		try
		{
			con = getConnection();
			return getEpointDB().findAuthorizedIssuerById(con, authorizedIssuerId);
		}
		catch (SQLException ex)
		{
			throw new DatabaseErrorException(ex);
		}
		finally
		{
			closeConnection(con);
		}
	}

	@Override
	public AuthorizedIssuer updateAuthorizedIssuer(final String authorizedIssuerFingerPrint, final String authorizedIssuerPublicKeyBlock, final String emissionLimitLowerBound, final String emissionLimitUpperBound, final MessageDecryptionVerificationHandler _signedMessage, final String _targetIssuer) throws RedcentProcessingException {

		// verify that request was originally intended for this issuer
		checkTargetIssuer(_targetIssuer);
		
		Connection con = null;
		try
		{
			con = getConnection();
			getEpointDB().beginTransaction(con);
			getEpointDB().lockWrites(con);
			
			boolean aiWasCreated = false;
			AuthorizedIssuer ai = getEpointDB().findAuthorizedIssuerByFingerprint(con, authorizedIssuerFingerPrint);
			PGPPublicKeyRing submittedRing = null;
			
			// AI does not exist, insert it into DB
			if (ai == null) {
				
				log.debug("No AI with fingerprint " + authorizedIssuerFingerPrint + " present in DB. Attempting to insert it.");
				
				// Verify that this request is a valid insertion request for the specified AI's key
				if (authorizedIssuerPublicKeyBlock == null || authorizedIssuerPublicKeyBlock.trim().length() == 0)
					throw new MalformedRequestException("Cannot create Authorized Issuer without a valid public key block!");
				
				log.debug("Attempting to read public key block from armored storage:\n" + authorizedIssuerPublicKeyBlock);
				log.debug("Public key block has Digest " + BouncycastleUtil.calcDigest(authorizedIssuerPublicKeyBlock));
				
				submittedRing = BouncycastleUtil.readArmoredPublicKeyRing(authorizedIssuerPublicKeyBlock.getBytes("UTF-8"));

				try{
					log.debug("Submitted key ring" + BouncycastleUtil.dumpPGPPublicKeyRing(submittedRing));
					} catch (Throwable t) {
						t.printStackTrace();
					}
				
				//verify that there is a signing key somewhere
				if (!BouncycastleUtil.keyContainsValidSigningKey(submittedRing))
					throw new MalformedRequestException("The submitted public key block does not contain a key suitable for signing.");

				//verify that authorizedIssuerId is the master key's id of submittedRing
				PGPPublicKey submittedPrimaryKey = BouncycastleUtil.getPrimaryKey(submittedRing);
				
				try{
				log.debug("Extracted primary key " + BouncycastleUtil.dumpPGPPublicKey(submittedPrimaryKey));
				} catch (Throwable t) {}
				
				if (!authorizedIssuerFingerPrint.equals(BouncycastleUtil.convertKeyFingerprintToHex(submittedPrimaryKey.getFingerprint())))
						throw new MalformedRequestException("The submitted public key block's master key does not match the selected AI's master key, '"+BouncycastleUtil.convertKeyFingerprintToHex(submittedPrimaryKey.getFingerprint())+" (Key-ID: "+submittedPrimaryKey.getKeyID()+")' <> '"+authorizedIssuerFingerPrint+"'");

				Map<AuthorizedIssuer, PGPSignature> notarySignatures = null;
				
				if (appConfig.getNumberOfNotarySignaturesRequiredForDefaultAIRegistration() != null) {
					notarySignatures = BouncycastleUtil.getNotarySignatures(getIssuerFingerPrint(), this, submittedPrimaryKey);
				} else
					log.debug("Not parsing notary signatures from submitted key, registration by notaries is disabled.");
				
				String introducingNotaries = null;
				String introductionAuthorization = null;
				
				if (_signedMessage != null && _signedMessage.isValid() && _signedMessage.signerIsIssuersMasterKey()) {
					
					log.debug("AI is created because of issuer master's signed request.");
					introducingNotaries = _signedMessage.getSignersPrimaryKeyFingerprint();
					introductionAuthorization = "Request signed by issuer master";

				} else if (notarySignatures != null && notarySignatures.size() >= appConfig.getNumberOfNotarySignaturesRequiredForDefaultAIRegistration()) {
					
					log.debug("AI is created because of signing notaries: " + introducingNotaries);
					introducingNotaries = BouncycastleUtil.getFingerprintList(notarySignatures.keySet());
					introductionAuthorization = "Public key block with positive certifications";

				} else
					throw new MalformedRequestException("The submitted request was neither properly signed by the issuer master nor has the submitted key been signed by the required number of notaries.");
				
				ai = getEpointDB().createAuthorizedIssuer(con, authorizedIssuerFingerPrint, authorizedIssuerPublicKeyBlock, BouncycastleUtil.getKeyIDList(submittedRing), appConfig.getDefaultEmissionLimitLowerBound(), appConfig.getDefaultEmissionLimitUpperBound(), BouncycastleUtil.hasKeyNotaryRights(getIssuerFingerPrint(), appConfig.getMasterKeyId(), submittedPrimaryKey), introducingNotaries, introductionAuthorization);
				aiWasCreated = true;
				log.debug("Inserted AI with fingerprint " + authorizedIssuerFingerPrint + ".");
			}
			
			// Update for an existing AI with a possibly new public key block:
			if (!aiWasCreated && (authorizedIssuerPublicKeyBlock != null || authorizedIssuerPublicKeyBlock.trim().length() > 0)) {

				// Only perform update if submitted block and AI's key block from db differ.
				if (!authorizedIssuerPublicKeyBlock.equals(ai.getPublicKey())) {

					submittedRing = BouncycastleUtil.readArmoredPublicKeyRing(authorizedIssuerPublicKeyBlock.getBytes("UTF-8"));
					PGPPublicKeyRing existingRing = BouncycastleUtil.readArmoredPublicKeyRing(ai.getPublicKey().getBytes("UTF-8"));

					//verify that authorizedIssuerId is the master key's id of submittedRing
					PGPPublicKey submittedMasterKey = BouncycastleUtil.getPrimaryKey(submittedRing);
					if (!ai.getFingerprint().equals(BouncycastleUtil.convertKeyFingerprintToHex(submittedMasterKey.getFingerprint())))
							throw new MalformedRequestException("The submitted public key block's primary key does not match the selected AI's primary key, '"+BouncycastleUtil.convertKeyFingerprintToHex(submittedMasterKey.getFingerprint())+"' <> '"+ai.getFingerprint()+"'");
					
					//verify that there is a signing key somewhere
					if (!BouncycastleUtil.keyContainsValidSigningKey(submittedRing))
						throw new MalformedRequestException("The submitted public key block does not contain a key suitable for signing.");

					//TODO: If not signed by master, check for key removals
					//TODO: perform compare of existing AI ring and submittedRing to ensure submittedRing is newer
					// submittedRing has to contain all currently known publicKeys and maybe some more?
					//TODO: verify that this public is newer than the one we have in our db
					log.warn("Updating AIs public key without checking for missing/changes signatures/subkeys!");

					PGPPublicKeyRing mergedRing = BouncycastleUtil.mergePublicKeys(existingRing, submittedRing);
					
					String mergedKeyBlock = BouncycastleUtil.writeArmoredPublicKeyRing(mergedRing);
					
					ai.setPublicKey(mergedKeyBlock);
					ai.setPublicKeyIdList(BouncycastleUtil.getKeyIDList(mergedRing));
					ai = getEpointDB().updateAuthorizedIssuersPublicKey(con, ai);
					
					//
					//	Update Notary Key Block
					//
					//verify notary signatures, care for revocations, too
					boolean submittedKeyHasNotaryRights = BouncycastleUtil.hasKeyNotaryRights(getIssuerFingerPrint(), appConfig.getMasterKeyId(), BouncycastleUtil.getPrimaryKey(submittedRing));
					//grant or revoke notary rights accordingly
					if (submittedKeyHasNotaryRights && !ai.hasNotaryRights()) {
						// grant notary rights
						ai = getEpointDB().updateAuthorizedIssuersNotaryStatus(con, ai, true);
					} else if (!submittedKeyHasNotaryRights && ai.hasNotaryRights()) {
						// revoke notary rights
						ai = getEpointDB().updateAuthorizedIssuersNotaryStatus(con, ai, false);
					}
					
				} else
					log.info("Ignoring unchanged public key for AI " + ai.getFingerprint());
			}
			
			// This is a priviledged update containing a signature
			if (_signedMessage != null && _signedMessage.isValid() && _signedMessage.signerIsIssuersMasterKey()) {
				// verify that the message was signed by the issuer's master key
				
				boolean saveToDB = false;
				
				if (emissionLimitLowerBound != null) {
					try {
						Long valueToSet = emissionLimitLowerBound.trim().length() > 0 ? Long.parseLong(emissionLimitLowerBound) : null;
						ai.setEmissionLimitLowerBound(valueToSet);
						saveToDB = true;
					}catch(NumberFormatException nf) {
						throw new MalformedRequestException("Unable to parse emissionLimitLowerBound '"+emissionLimitLowerBound+"' to a value of type long.");
					}
				}

				if (emissionLimitUpperBound != null) {
					try {
						Long valueToSet = emissionLimitUpperBound.trim().length() > 0 ? Long.parseLong(emissionLimitUpperBound) : null;
						ai.setEmissionLimitUpperBound(valueToSet);
						saveToDB = true;
					}catch(NumberFormatException nf) {
						throw new MalformedRequestException("Unable to parse emissionLimitHigherBound '"+emissionLimitUpperBound+"' to a value of type long.");
					}
				}
				
				if(saveToDB)
					ai = getEpointDB().updateAuthorizedIssuersEmissionLimits(con, ai, new String(_signedMessage.getMessage(), "UTF8"));
			}
			
			con.commit();
			return ai;
		}
		catch (SQLException ex)
		{
			throw rollbackAndReturnCause(con, new DatabaseErrorException(ex));
		}
		catch (RedcentProcessingException ex)
		{
			throw rollbackAndReturnCause(con, ex);
		}
		catch (UnsupportedEncodingException e)
		{
			throw rollbackAndReturnCause(con, new ConflictException(e));
		}
		catch (IOException e)
		{
			throw rollbackAndReturnCause(con, new ConflictException(e));
		}
		catch (Exception e)
		{
			throw rollbackAndReturnCause(con, new ConflictException(e));
		}
		finally
		{
			closeConnection(con);
		}
	}
	
	private String calcIdFromSecret(String rand) throws NoSuchAlgorithmException
	{
		if (rand == null)
			return null;
		
		// Calculate a certificate's ID from a given rand
		log.trace("Secret: " + rand);
		byte[] code = Base64.decodeBase64(rand.getBytes());
		MessageDigest md = MessageDigest.getInstance("SHA");
		String id = new String(Hex.encodeHex(md.digest(code))).toUpperCase();
		
		return id;
	}
}
