/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.sqlImplementation;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;

import javax.annotation.PostConstruct;
import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.annotation.Autowired;

import de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore;

/**
 * @author js
 *
 */
public class SqlDBStoringDocumentStore implements DocumentStore
{
	private static final Log log = LogFactory.getLog(SqlDBStoringDocumentStore.class);
	
	@Autowired
	private DataSource dataSource  = null;	

	/**
	 * Method just selects count of documents in table to verify that the table exists.
	 */
	protected void ensureArchiveStorageExists() throws Exception
	{
		log.info("Ensuring archive storage exists  with datasource " + dataSource);
		
		PreparedStatement stm = null;
		ResultSet rs = null;
		Connection con = null;
		try
		{
			con = dataSource.getConnection();
			stm = con.prepareStatement("SELECT COUNT(documentId) FROM document");
			rs = stm.executeQuery();
			
			// do not parse ResultSet, the selected result does not matter. The only thing to verify is that
			// the query did run well.
		}
		finally
		{
			if (rs != null)
				rs.close();
			if (stm != null)
				stm.close();
			if (con != null)
				con.close();
		}
	}
	
	/**
	 * @see de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore#getDocument(java.lang.String, java.lang.String)
	 */
	public byte[] getDocument(String realm, String documentId) throws Exception
	{
		PreparedStatement stm = null;
		ResultSet rs = null;
		Connection con = null;
		try
		{
				con = dataSource.getConnection();
				// update
				stm = con.prepareStatement("SELECT data FROM document WHERE realm=? AND documentId=?");
				
				stm.setString(1, realm);
				stm.setString(2, documentId);
				
				rs = stm.executeQuery();
				
				if (rs.next())
				{
					byte data[] = rs.getBytes(1);
					
					if (rs.next())
						throw new Exception("Error: document store returns more than one document for unique identifier '"+realm + " / " + documentId +"'");
					
					return data;
				}
				else
					return null;
		}
		finally
		{
			if (rs != null)
				rs.close();
			if (stm != null)
				stm.close();
			if (con != null)
				con.close();
		}
	}

	/**
	 * @see de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore#storeOrUpdateDocument(java.lang.String, java.lang.String, byte[])
	 */
	public void storeOrUpdateDocument(String realm, String documentId, byte[] data) throws Exception
	{
		PreparedStatement stm = null;
		ResultSet rs = null;
		Connection con = null;
		try
		{
			con = dataSource.getConnection();

			if (getDocument(realm, documentId) != null)
			{
				// update
				stm = con.prepareStatement("UPDATE document SET data=? WHERE realm=? AND documentId=?");
				
				stm.setBytes(1, data);
				stm.setString(2, realm);
				stm.setString(3, documentId);
				
				stm.execute();
			}
			else
			{
				// insert
				stm = con.prepareStatement("INSERT INTO document (realm, documentId, data) VALUES (?,?,?)");
				
				stm.setString(1, realm);
				stm.setString(2, documentId);
				stm.setBytes(3, data);
				
				stm.execute();
			}
			
		}
		finally
		{
			if (rs != null)
				rs.close();
			if (stm != null)
				stm.close();
			if (con != null)
				con.close();
		}
	}

	/* (non-Javadoc)
	 * @see de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore#init()
	 */
	@PostConstruct
	public void init() throws Exception
	{
		ensureArchiveStorageExists();
	}

	/* (non-Javadoc)
	 * @see de.nachtlicht_media.epoint.redcent.interfaces.DocumentStore#getMostRecentDocumentId(java.lang.String)
	 */
	public String getMostRecentDocumentId(String realm) throws Exception
	{

		PreparedStatement stm = null;
		ResultSet rs = null;
		Connection con = null;
		try
		{
			con = dataSource.getConnection();
			// update
			stm = con.prepareStatement("SELECT documentId FROM document WHERE realm=? AND documentIndex=(SELECT MAX(documentIndex) FROM document WHERE realm=?)");

			stm.setString(1, realm);
			stm.setString(2, realm);

			rs = stm.executeQuery();

			if (rs.next())
				return rs.getString(1);

			return null;
		}
		finally
		{
			if (rs != null)
				rs.close();
			if (stm != null)
				stm.close();
			if (con != null)
				con.close();
		}
	}
}
