/*******************************************************************************
 * Copyright (c) 2006-2010 Janis Schuller, nachtlicht-media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package de.nachtlicht_media.epoint.redcent.sqlImplementation;

import java.io.ByteArrayInputStream;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;

import javax.sql.DataSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.annotation.Autowired;

import de.nachtlicht_media.epoint.redcent.interfaces.RedCentConfig;
import de.nachtlicht_media.epoint.redcent.messaging.DefaultTemplateManager;

/**
 * A TemplateManager implementation that uses an SQL-backend to store archived certificates.
 * @author js
 *
 */
public class SqlDbStoringTemplateManager extends DefaultTemplateManager
{
	private static final Log log = LogFactory.getLog(SqlDbStoringTemplateManager.class);

	@Autowired
	private DataSource dataSource  = null;		
	
	public SqlDbStoringTemplateManager()
	{
		log.info("Created SqlDbStoringTemplateManager with issuer " + this.issuer + " and dataSource " + dataSource);
	}
	
	/**
	 * Method just selects count of templates in table to verify that the table exists.
	 * @see de.nachtlicht_media.epoint.redcent.messaging.DefaultTemplateManager#ensureArchiveStorageExists()
	 */
	protected void ensureArchiveStorageExists() throws Exception
	{
		PreparedStatement stm = null;
		ResultSet rs = null;
		Connection con = null;
		try
		{
			con = dataSource.getConnection();
			stm = con.prepareStatement("SELECT COUNT(id) FROM template");
			rs = stm.executeQuery();
			
			// do not parse ResultSet, the selected result does not matter. The only thing to verify is that
			// the query did run well.
		}
		finally
		{
			if (rs != null)
				rs.close();
			if (stm != null)
				stm.close();
			if (con != null)
				con.close();
		}
	}
	
	protected void readArchivedTemplatesToCache() throws Exception
	{
		PreparedStatement stm = null;
		ResultSet rs = null;
		Connection con = null;
		try
		{
			con = dataSource.getConnection();
			stm = con.prepareStatement("SELECT id, template FROM template");
			rs = stm.executeQuery();
			
			while(rs.next())
			{
				// Read record id (digest) and templateData from ResultSet
				String id = rs.getString(1);
				byte[] templateData = rs.getBytes(2);
				
				// Create the application internal String representation from the binary data
				String template = readTemplate(new ByteArrayInputStream(templateData));
				
				String digest = calcDigest(template);
				
				// If the given id(digest) matches the recalculated digest, add template to cache
				if (id.equalsIgnoreCase(digest))
					templateCache.put(digest, template);
				else
					throw new Exception("Ignored archived template object " + id + ": Persisted id does not match content's hash value '"+digest+"'");
			}
		}
		finally
		{
			if (rs != null)
				rs.close();
			if (stm != null)
				stm.close();
			if (con != null)
				con.close();
		}
	}
	
	protected void persistTemplateVersion(String digest, String template) throws Exception
	{
		PreparedStatement stm = null;
		ResultSet rs = null;
		Connection con = null;
		try
		{
			con = dataSource.getConnection();
			stm = con.prepareStatement("INSERT INTO template (id, template) VALUES (?,?)");
			
			stm.setString(1, digest);
			stm.setBytes(2, template.getBytes("UTF-8"));
			
			stm.execute();
		}
		finally
		{
			if (rs != null)
				rs.close();
			if (stm != null)
				stm.close();
			if (con != null)
				con.close();
		}
	}
}
