/*******************************************************************************
 * Copyright (c) 2006-2007 Janis Schuller, Nachtlicht-Media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package hu.vems.epoint.walcent;

import java.io.*;
import java.security.*;
import java.sql.SQLException;
import java.util.*;

import org.apache.commons.codec.*;
import org.apache.commons.codec.binary.*;
import org.apache.commons.logging.*;
import org.bouncycastle.bcpg.*;
import org.bouncycastle.openpgp.*;

import hu.vems.epoint.walcent.crypto.*;
import hu.vems.epoint.walcent.interfaces.*;
import hu.vems.epoint.walcent.messaging.*;
import hu.vems.epoint.walcent.util.*;
import de.nachtlicht_media.epoint.redcent.*;
import de.nachtlicht_media.epoint.redcent.crypto.*;
import de.nachtlicht_media.epoint.redcent.exception.*;
import de.nachtlicht_media.epoint.redcent.interfaces.*;
import de.nachtlicht_media.epoint.redcent.messaging.*;
import de.nachtlicht_media.epoint.redcent.util.*;
import de.nachtlicht_media.epoint.redcent.util.StringUtils;

/**
 * WalCent is the commmon super-class for all implementations of the market application. It is instaciated once,
 * for example by a servlet-application and handles all ePoint transactions and signing processes.
 */
public abstract class WalCent extends EpointService
{

	private static final Log log = LogFactory.getLog(WalCent.class);
	protected Log xLog;
	
	public static final String DOCUMENT_REALM_ISSUED_DRAFTS = "DOCUMENT_REALM_ISSUED_DRAFTS";
	public static final String DOCUMENT_REALM_OFFER_CERTIFICATES = "DOCUMENT_REALM_OFFER_CERTIFICATES";
	
	public WalCent(RedCentConfig config)
	{
		super(config);
	}
	
	public String getMarketsPublicKeyRingPath()
	{
		return getIssuersPublicKeyRingPath();
	}

	public String getMarketsPrivateKeyRingPath()
	{
		return getIssuersPrivateKeyRingPath();
	}
	
	/**
	 * @return the password required to extract the markets private key from his private keyring
	 */
	protected String getMarketsKeyPassword()
	{
		return "test";
	}

	/**
	 * Dependency injection style configuration function for MarketURL.
	 * @return The webapplication's URL on the
	 * external interface. Used to replace variable $ISSUER_URL in templates.
	 */
	public String getMarketURL()
	{
		return "http://localhost:8080/walcent";
	}
	
	/**
	 * Dependency injection style configuration function for MarketName.
	 * @return The market's name, used to replace variable $ISSUER_NAME in templates.
	 */
	public String getMarketName()
	{
		return "ePoint test market";
	}
	
	/**
	 * Dependency injection style configuration function for SSL_KEY.
	 * @return The market's ssl key, used to replace variable $SSL_KEY in templates.
	 */
	public String getSSLKey()
	{
		return "N/A";
	}
	
	

	public String getMarketsEmailAddress()
	{
		return "test@host.com";
	}
	
	/**
	 * 
	 * @return the number of digits ePoint certificate's serial numbers are padded to.
	 */
	public int getFixedSerialNumberlength()
	{
		return 16;
	}
	
	

	public void init() throws Exception
	{
		testDataSource();
		
		String xLogName = StringUtils.replace(StringUtils.replace(getMarketURL(), "/" , "_"), ":", "_") + ".TransactionLog";
		xLog = LogFactory.getLog(xLogName);

		BouncycastleMessageDecryptionVerificationFactory.init(
				null,
				getMasterKeyId(),
				new PGPPublicKeyRingCollection(PGPUtil.getDecoderStream(new FileInputStream(getMarketsPublicKeyRingPath()))), 
				new PGPSecretKeyRingCollection(PGPUtil.getDecoderStream(new FileInputStream(getMarketsPrivateKeyRingPath()))), 
				getMarketsKeyPassword(), new PGPPublicKeyRingCollection(PGPUtil.getDecoderStream(new FileInputStream(getTrustedPublicKeyRingPath()))));
		
//		initMessageSignatureCreator();
		
//		initMessageEncryptor();
		
		initTemplateManager();
		
//		initDocumentStore();
		
		initServiceContract();
		
		log.info("Initialized market at " + getMarketURL());
		log.info("Initialized transaction logger '"+xLogName+"'");
		xLog.info("INIT " + getMarketURL());
	}

	public void destroy() throws Exception
	{
		
	}
	
	protected void initServiceContract() throws ConflictException
	{
		// verify that the service contract located at /$WEBROOT/service is up to date
		// if not update it from the current template
		try
		{
			String currentServiceContractTemplate = getTemplateManager().getTemplate(getTemplateManager().getHashValueForCurrentVersion(EpointCertificateTypes.SERVER_SERVICE_CONTRACT));
		
			byte[] deployedServiceContractData = getDocumentStore().getDocument(DOCUMENT_REALM_ISSUER_PERSISTENT_DOCUMENT, EpointCertificateTypes.SERVER_SERVICE_CONTRACT);
		
			MessageDecryptionVerificationHandler messageSignatureChecker = null;

			boolean replaceServiceContract = false;

			if (deployedServiceContractData != null)
			{
				String deployedServiceContract = getTemplateManager().readTemplate(new ByteArrayInputStream(deployedServiceContractData));
				messageSignatureChecker =  BouncycastleMessageDecryptionVerificationFactory.getMessageDecryptionVerificationHandler();
				messageSignatureChecker.setMessage(deployedServiceContract.getBytes("UTF-8"));

				try{
					messageSignatureChecker.analyse();
				}catch (InvalidPGPMessageException e) {
					replaceServiceContract = true;
				}
			}

			if (	messageSignatureChecker != null 
					&& 	messageSignatureChecker.isValid() != null && messageSignatureChecker.isValid().booleanValue() 
					&& 	messageSignatureChecker.isClearText() != null && messageSignatureChecker.isClearText().booleanValue())
			{
				// Message is properly signed and in cleartext, check contents
				log.info("Found a service contract at webroot, comparing with a current version from the template archive...");
				
				String deployedServiceContractClearText = messageSignatureChecker.getClearText();
//				deployedServiceContractClearText = deployedServiceContractClearText.substring(deployedServiceContractClearText.indexOf("\n\n") + 2);
				
				if (getTemplateManager().calcDigest(createServerServiceContractContents()).equals(getTemplateManager().calcDigest(deployedServiceContractClearText)))
				{
					log.info("... service contract is up to date and will not be touched.");
					replaceServiceContract = false;
				}
				else
				{
					log.info("... service contract is not up to date and will be replaced.");
					log.debug("currentServiceContractTemplate is '"+currentServiceContractTemplate+"'");
					log.debug("deployedServiceContractTemplate is '"+deployedServiceContractClearText+"'");
					replaceServiceContract = true;
				}
			}
			else
				replaceServiceContract = true;
			
			if (replaceServiceContract)
			{
				// The deployed service contract is not signed or not cleartext or nonexisting...
				// get a signed current version and write it to the documentStore
				log.info("Updating service contract with a current version from the template archive...");
				getDocumentStore().storeOrUpdateDocument(DOCUMENT_REALM_ISSUER_PERSISTENT_DOCUMENT, EpointCertificateTypes.SERVER_SERVICE_CONTRACT, createSignedServerServiceContract());
			}
		}
		catch(Exception ex)
		{
			throw new ConflictException(ex);
		}
	}

	private byte[] createSignedServerServiceContract() throws DatabaseErrorException, ConflictException
	{
		return signMessage(createServerServiceContractContents());
	}

	private String createServerServiceContractContents() throws ConflictException
	{
		StringDataMap data = new StringDataMap();
		
		data.put("ISSUER_URL", getMarketURL());
		data.put("ISSUER_NAME", getMarketName());
		data.put("ISSUER_EMAIL", getMarketsEmailAddress());
		data.put("ISSUER_FINGERPRINT", getMarketFingerPrintText());
		data.put("SSL_KEY", getSSLKey());

		return getMessageHandler().createMessage(EpointCertificateTypes.SERVER_SERVICE_CONTRACT, data);
	}

	/**
	 * @return TemplateManager
	 */
	public TemplateManager getTemplateManager() throws ConflictException
	{
		try
		{
			if (templateManager == null)
				initTemplateManager();
		}
		catch(Exception ex)
		{
			throw new ConflictException(ex);
		}
		
		return templateManager;
	}
	
	protected TemplateManager templateManager;
	protected void initTemplateManager() throws Exception
	{
		final String templateDir = getTemplateRealPath();
		log.info("Template dir is: " + templateDir);
		
		templateManager = new DefaultTemplateManager()
		{
			protected String getTemplatePath()
			{
				return templateDir;
			}
		};
		
		templateManager.init(this);
	}
	
	private MessageHandler messageHandler = null;
	protected MessageHandler getMessageHandler()
	{
		if (messageHandler == null)
			messageHandler = new DefaultMessageHandler(this);
		
		return messageHandler;
	}
	
/*	private EpointRecordConverter xLogConverter = null;
	protected EpointRecordConverter getXLogConverter()
	{
		if (xLogConverter == null)
			xLogConverter = new EpointRecordToXLogConverter();
		
		return xLogConverter;
	}*/
	
	public String getMarketFingerPrintText() throws ConflictException
	{
		return getMessageSignatureCreator().getIssuerFingerPrintText();
	}

	/* All abstract methods beneath have to be implemented by specific implementations, for example an SQL-based solution*/

//	public abstract boolean testDataSource() throws Exception;
//	public abstract byte[] getServerStatus() throws Exception;
//	public abstract byte[] getServerBalance(String key) throws Exception;

//	public abstract byte[] getExchangeRateInformation() throws Exception;

	
	public byte[] getDeployedServerServiceContract() throws Exception
	{
		return getDocumentStore().getDocument(DOCUMENT_REALM_ISSUER_PERSISTENT_DOCUMENT, EpointCertificateTypes.SERVER_SERVICE_CONTRACT);
	}
	
	public abstract byte[] processDepositMessage(MessageDecryptionVerificationHandler signedMessage) throws Exception;//WalcentProcessingException, SQLException, UnsupportedEncodingException;
	public abstract byte[] processAccountInfo(byte[] UserFPR) throws Exception;//WalcentProcessingException, SQLException, UnsupportedEncodingException;
	public abstract byte[] withdraw(byte[] UserFPR, byte[] IssuerFPR, long amount, String targetMD, String reqText) throws Exception;
	public abstract byte[] offer(byte[] UserFPR, byte[] IssuerAFPR, byte[] IssuerBFPR, long RA, long RB, long VA, long VB, long SN, String reqText) throws Exception;
	public abstract byte[] search(byte[] IssuerAFPR, byte[] IssuerBFPR, long RA, long RB, long VA, long VB) throws Exception;
}
