/*******************************************************************************
 * Copyright (c) 2006-2007 Janis Schuller, Nachtlicht-Media
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package hu.vems.epoint.walcent.servlet.filter;

import java.io.*;

import javax.servlet.*;

import org.apache.commons.logging.*;

import hu.vems.epoint.walcent.servlet.crypto.*;

import de.nachtlicht_media.epoint.redcent.exception.*;
import de.nachtlicht_media.epoint.redcent.interfaces.*;
import de.nachtlicht_media.epoint.redcent.servlet.exception.*;
import hu.vems.epoint.walcent.util.FPRUtils;

public class SecureMessageRedirectionFilter implements Filter
{
	private static final Log log = LogFactory.getLog(SecureMessageRedirectionFilter.class);
	public static final String ATTR_SECURE_MESSAGE_FLAG = "ATTR_SECURE_MESSAGE_FLAG";
	public static final String ATTR_SECURE_MESSAGE_HANDLER = "ATTR_SECURE_MESSAGE_HANDLER";
	
	private FilterConfig filterConfig;

	public void init(FilterConfig filterConfig) throws ServletException
	{
		log.info("init()");
		this.filterConfig = filterConfig;
	}

	public void destroy()
	{
		log.info("destroy()");
	}
	
	public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException
	{	
		if (request.getParameterMap().isEmpty())
			throw new ProcessingException("Illegal call to " + filterConfig.getFilterName() + " with empty parameterMap.");
		else if (request.getAttribute(ATTR_SECURE_MESSAGE_FLAG) != null)
		{
			log.warn("Illegal call to " + filterConfig.getFilterName() + " with a request that was already processed.");
			throw new ProcessingException(404, "Illegal call to /secureMessage with target /secureMessage", "Illegal call to " + filterConfig.getFilterName() + " with a request that was already processed.");
		}
		else
		{
			request.setAttribute(ATTR_SECURE_MESSAGE_FLAG, ATTR_SECURE_MESSAGE_FLAG);
			//if (request.getParameterMap().size() == 1)
			{
				String message = request.getParameter("M");
				
				if (message != null && message.length() > 0)
				{
					try
					{
						// Decode
						MessageDecryptionVerificationHandler msg = MarketServletMessageDecryptionVerificationFactory.getMessageHandler();
						
//						byte[] UserFPR=FPRUtils.parseFPR(request.getParameter("U"));
						String pubKey=request.getParameter("KEYDATA");
						if (/*UserFPR!=null || */pubKey!=null)
						{
							msg.addTemporaryPublicKey(pubKey);
						}
						msg.setMessage(message.getBytes("UTF-8"));
						msg.analyse();
						
						if (msg.isValid() == null)
						{
							// if msg.isValid() there must have occured a processing exception that prevented
							// the msgProcessor from deciding whether the message is valid or not.
							
							// Do not process this request any more
							throw new ProcessingException("The Decryption/Verification-Handler could not validate message. Request ignored.");
						}
						else if (msg.isValid().booleanValue())
						{
							request.setAttribute(ATTR_SECURE_MESSAGE_HANDLER, msg);
							
							String redirection = msg.getClearText();
							// messages routed through this filter MUST contain a redirect call to e.g. /action?... or /emit?...
							if (redirection != null && redirection.length() > 0 && redirection.startsWith("/"))
							{
								// Signed messages usually end with a linebreak that must not be forwarded
								if (redirection.endsWith("\n"))
									redirection = redirection.substring(0, redirection.length() - 1);
									
								// looks like a valid redirect and stop filtering for this call
								log.debug("Forwarding request to '" + redirection + "'");
								request.setCharacterEncoding("UTF8");
								request.getRequestDispatcher(redirection).forward(request, response);
							}
							else
								throw new ProcessingException("Malformed message: the encrypted/signed message MUST contain a valid redirection.");
						}
						else
							throw new ProcessingException("The Decryption/Verification-Handler returned an invalid message state. Request ignored.");
					}
					catch (InvalidPGPMessageException e)
					{
						request.setAttribute(ATTR_SECURE_MESSAGE_HANDLER, null);
						ProcessingException pe = new ProcessingException(RedcentProcessingException.FORB_BAD_CERT, "Error decoding message", e.getMessage());
						throw pe;
					}
					catch(ProcessingException ex)
					{
						request.setAttribute(ATTR_SECURE_MESSAGE_HANDLER, null);
						throw ex;//new ProcessingException(ex);
					}
					catch (Exception e)
					{
						request.setAttribute(ATTR_SECURE_MESSAGE_HANDLER, null);
						ProcessingException pe = new ProcessingException(RedcentProcessingException.CONFLICT, "Error decoding message", e.getMessage());
						throw pe;
					}
				}
				else
					throw new ProcessingException("Illegal call to " + filterConfig.getFilterName() + " with empty message parameter.");
			}
		}
	}
}
