/*******************************************************************************
 * Copyright (c) 2009 Felfldi Zsolt
 * 
 * The software and the associated documentation are licensed under the 
 * GNU GENERAL PUBLIC LICENSE, Version 3, 29 June 2007. For a detailed
 * license description, see LICENSE.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *******************************************************************************/
package hu.vems.epoint.walcent.sqlImplementation;

import java.sql.*;
import java.util.*;

import org.apache.commons.logging.*;

import hu.vems.epoint.walcent.*;
import hu.vems.epoint.walcent.sqlImplementation.WalCentSqlImplementation.*;
import hu.vems.epoint.walcent.util.FPRUtils;

import de.nachtlicht_media.epoint.redcent.*;
import de.nachtlicht_media.epoint.redcent.exception.*;

public class MarketSQLDB
{
	private boolean collectStats = false;
	StatsCollector statsCollector;
	
	public void setStatsCollector(StatsCollector statsCollector)
	{
		this.statsCollector = statsCollector;
		this.collectStats = (statsCollector != null);
	}
	
	private static final Log log = LogFactory.getLog(MarketSQLDB.class);

	public MarketSQLDB()
	{
	}

	public void verifyDatabase(Connection con) throws Exception
	{
//TODO: verification		
		
		// verify that all serialnumbers do not have gaps
		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			// get number of assigned certificates
			pStm = con.prepareStatement("SELECT COUNT(serialnumber) from certificate");			
			rs = pStm.executeQuery();
			
			rs.next();
			int count = rs.getInt(1);
			closeStatementAndResultSet(pStm, rs);
			pStm = null;
			rs = null;
			
			if (count > 0)
			{
				// verify that that the last assigned certificate's serialnumber is equal to the total count of numbers
				// this together with the unique and NOT NULL-constraint of the database column guarantees consistent serial numbers
				pStm = con.prepareStatement("SELECT MAX(serialnumber) from certificate");			
				rs = pStm.executeQuery();
				
				rs.next();
				int max = rs.getInt(1);
				closeStatementAndResultSet(pStm, rs);
				pStm = null;
				rs = null;
				
				if (count == max)
					log.info("Verification of certificate database succeeded, the last assigned certificate has serial number " + max);
				else
					throw new Exception("The found serial numbers are inconsistent, the application may not proceed.");
			}
			else
				log.info("Verification of empty certificate database succeeded.");

			// get total amount of emitted ePoints from certificate store
			pStm = con.prepareStatement("SELECT SUM(value) FROM certificate WHERE isValid=?");
			pStm.setBoolean(1, true);
			rs = pStm.executeQuery();
			
			rs.next();
			long totalCertificateValue = rs.getLong(1);
			closeStatementAndResultSet(pStm, rs);
			pStm = null;
			rs = null;

			// verify that total amount from certificate store and total amount from balance match
//			if (!new Long(totalCertificateValue).equals(getTotalAmount(con)))
//				throw new Exception("The database's balance is not correct: " + totalCertificateValue + " ePoints are balanced, but the total value of valid certificates is: " + getTotalAmount(con));
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
	}
	
	public long findUserAccRecord(Connection con, byte[] UserFPR, byte[] IssuerFPR) throws SQLException
	{
//		long startTime = System.currentTimeMillis();

		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("SELECT Amount from UserAcc WHERE UserFPR=? AND IssuerFPR=?");			
			pStm.setBytes(1, UserFPR);
			pStm.setBytes(2, IssuerFPR);
			rs = pStm.executeQuery();

//			if(collectStats)
//				statsCollector.logStatRecord(new WalCentSqlImplementation.StatRecord(WalCentSqlImplementation.StatRecord.AC_FIND_USERACC_RECORD, FPRUtils.ToString(UserFPR)+" / "+FPRUtils.ToString(IssuerFPR) , System.currentTimeMillis()-startTime));

			if (rs.next()) {
				long amount=rs.getLong(1);
				if (amount==0) setUserAccRecord(con, UserFPR, IssuerFPR, 0, false);
				return amount;
			} else return 0;
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
			
	}
	
	public void setUserAccRecord(Connection con, byte[] UserFPR, byte[] IssuerFPR, long amount, boolean newRecord) throws SQLException
	{
//		long startTime = System.currentTimeMillis();

		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			if (amount!=0) {
				if (newRecord)
					pStm = con.prepareStatement("INSERT INTO UserAcc (Amount, UserFPR, IssuerFPR) VALUES (?,?,?)");
				else
					pStm = con.prepareStatement("UPDATE UserAcc SET Amount=? WHERE UserFPR=? AND IssuerFPR=?");
				
				pStm.setLong(1, amount);
				pStm.setBytes(2, UserFPR);
				pStm.setBytes(3, IssuerFPR);
//				rs = pStm.executeQuery();
				pStm.executeUpdate();
			}
			else if (!newRecord)
			{
				pStm = con.prepareStatement("DELETE FROM UserAcc WHERE UserFPR=? AND IssuerFPR=?");
				pStm.setBytes(1, UserFPR);
				pStm.setBytes(2, IssuerFPR);
//				rs = pStm.executeQuery();
				pStm.executeUpdate();
			}
			

//			if(collectStats)
//				statsCollector.logStatRecord(new WalCentSqlImplementation.StatRecord(WalCentSqlImplementation.StatRecord.AC_SET_USERACC_RECORD, FPRUtils.ToString(UserFPR)+" / "+FPRUtils.ToString(IssuerFPR)+" / "+ , System.currentTimeMillis()-startTime));

		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
			
	}

	public boolean findOCert(Connection con, byte[] IssuerFPR, long OSN) throws SQLException
	{
//		long startTime = System.currentTimeMillis();

		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("SELECT SN from OCert WHERE IssuerFPR=? AND SN=?");			
			pStm.setBytes(1, IssuerFPR);
			pStm.setLong(2, OSN);
			rs = pStm.executeQuery();

			return rs.next();
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
			
	}
	
	public void deactivateOffer(Connection con, long SN, long nextSN) throws SQLException
	{
//		long startTime = System.currentTimeMillis();

		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("UPDATE Offer SET active=false, nextSN=? WHERE SN=?");			
			pStm.setLong(1, nextSN);
			pStm.setLong(2, SN);
			pStm.executeUpdate();
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
			
	}
	
	public void updateOfferExchanged(Connection con, OfferRecord offer) throws SQLException
	{
//		long startTime = System.currentTimeMillis();

		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("UPDATE Offer SET active=?, ExchangedA=? WHERE SN=?");			
			pStm.setBoolean(1, offer.exchanged<offer.limit);
			pStm.setLong(2, offer.exchanged);
			pStm.setLong(3, offer.SN);
			pStm.executeUpdate();
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
			
	}
	
	public void addOffer(Connection con, OfferRecord offer) throws SQLException
	{
//		long startTime = System.currentTimeMillis();

		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("INSERT INTO Offer (SN, prevSN, nextSN, active, UserFPR, IA, IB, RA, RB, logR, LimitA, ExchangedA, atomic, ReqText, ReqHash) VALUES (?,?,-1,true,?,?,?,?,?,?,?,0,?,?,?)");			
			pStm.setLong(1, offer.SN);
			pStm.setLong(2, offer.prevSN);
			pStm.setBytes(3, offer.UserFPR);
			pStm.setBytes(4, offer.IssuerAFPR);
			pStm.setBytes(5, offer.IssuerBFPR);
			pStm.setLong(6, offer.RA);
			pStm.setLong(7, offer.RB);
			pStm.setLong(8, offer.logR);
			pStm.setLong(9, offer.limit);
			pStm.setBoolean(10, offer.atomic);
			pStm.setBytes(11, offer.reqText);
			pStm.setBytes(12, offer.reqHash);
			pStm.executeUpdate();
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
			
	}
	
	public OfferRecord findBestOffer(Connection con, byte[] IA, byte[] IB, long VA, long VB) throws SQLException
	{

		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{

			if (VB==-1) {
//				pStm = con.prepareStatement("SELECT SN, UserFPR, IA, IB, RA, RB, logR, LimitA, ExchangedA, AvailA, atomic from OfferSearch WHERE IA=? AND IB=? AND LimitA>ExchangedA AND AvailA>0 AND (atomic=false OR LeftA<=?) ORDER BY logR DESC LIMIT 1");			
				pStm = con.prepareStatement("SELECT SN, UserFPR, IA, IB, RA, RB, logR, LimitA, ExchangedA, AvailA, atomic from OfferSearch WHERE IA=? AND IB=? AND AvailA>0 AND (atomic=false OR (LimitA-ExchangedA)<=?) ORDER BY logR DESC LIMIT 1");			
				pStm.setBytes(1, IA);
				pStm.setBytes(2, IB);
				pStm.setLong(3, VA);
			} else {
				pStm = con.prepareStatement("SELECT SN, UserFPR, IA, IB, RA, RB, logR, LimitA, ExchangedA, AvailA, atomic from OfferSearch WHERE IA=? AND IB=? AND AvailA>0 AND (atomic=false OR (LimitA-ExchangedA)*RB<=?*RA) ORDER BY logR DESC LIMIT 1");			
				pStm.setBytes(1, IA);
				pStm.setBytes(2, IB);
				pStm.setLong(3, VB);
			}
			
			rs = pStm.executeQuery();
			
			if (!rs.next()) return null;

			return getRSOfferInfo(rs);
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
	}
	
	OfferRecord getRSOfferInfo(ResultSet rs) throws SQLException
	{
		
		OfferRecord offer = new OfferRecord();
		
		offer.SN=rs.getLong(1);
		offer.UserFPR=rs.getBytes(2);
		offer.IssuerAFPR=rs.getBytes(3);
		offer.IssuerBFPR=rs.getBytes(4);
		offer.RA=rs.getLong(5);
		offer.RB=rs.getLong(6);
		offer.logR=rs.getLong(7);
		offer.limit=rs.getLong(8);
		offer.exchanged=rs.getLong(9);
		offer.avail=rs.getLong(10);
		offer.atomic=rs.getBoolean(11);
		
		if ((offer.limit!=-1) && (offer.limit-offer.exchanged<offer.avail)) offer.avail=offer.limit-offer.exchanged;
		
		return offer;

	}
	
	

	public byte[] findBestIMCurrency(Connection con, byte[] IA, byte[] IB) throws SQLException
	{

		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("SELECT T1.IB, (T1.lr+T2.lr) AS lrs FROM (SELECT IB, MAX(logR) AS lr FROM OfferSearchAvailNA WHERE IA=? GROUP BY IB) T1 JOIN (SELECT IA, MAX(logR) AS lr FROM OfferSearchAvailNA WHERE IB=? GROUP BY IA) T2 ON T1.IB=T2.IA ORDER BY lrs DESC LIMIT 1");			
			pStm.setBytes(1, IA);
			pStm.setBytes(2, IB);
			
			rs = pStm.executeQuery();
			
			if (!rs.next()) return null;
			return rs.getBytes(1);
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
		
	}

	public void addOCert(Connection con, byte[] IssuerFPR, long OSN) throws SQLException
	{
//		long startTime = System.currentTimeMillis();

		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("INSERT INTO OCert (IssuerFPR, SN) VALUES (?,?)");			
			pStm.setBytes(1, IssuerFPR);
			pStm.setLong(2, OSN);
			pStm.executeUpdate();
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
			
	}


	public OfferRecord getSearchOfferInfo(Connection con, long SN) throws SQLException
	{

		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{

			pStm = con.prepareStatement("SELECT SN, UserFPR, IA, IB, RA, RB, logR, LimitA, ExchangedA, AvailA, atomic from OfferSearch WHERE SN=?");			
			pStm.setLong(1, SN);
			
			rs = pStm.executeQuery();
			
			if (!rs.next()) return null;

			return getRSOfferInfo(rs);
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
	}
	
	public OfferRecord getFullOfferInfo(Connection con, long SN) throws SQLException
	{
		OfferRecord offer = new OfferRecord();

		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("SELECT prevSN, nextSN, active, UserFPR, IA, IB, RA, RB, logR, LimitA, ExchangedA, atomic, ReqText, ReqHash from Offer WHERE SN=?");			
			pStm.setLong(1, SN);
			rs = pStm.executeQuery();
			
			rs.next();

			offer.SN=SN;
			offer.prevSN=rs.getLong(1);
			offer.nextSN=rs.getLong(2);
			offer.active=rs.getBoolean(3);
			offer.UserFPR=rs.getBytes(4);
			offer.IssuerAFPR=rs.getBytes(5);
			offer.IssuerBFPR=rs.getBytes(6);
			offer.RA=rs.getLong(7);
			offer.RB=rs.getLong(8);
			offer.logR=rs.getLong(9);
			offer.limit=rs.getLong(10);
			offer.exchanged=rs.getLong(11);
			offer.atomic=rs.getBoolean(12);
			offer.reqText=rs.getBytes(13);
			offer.reqHash=rs.getBytes(14);
			
			return offer;
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
		
	}
	
	public String getAccountInfo(Connection con, byte[] UserFPR, long ValueMultiplier) throws SQLException
	{
		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("SELECT IssuerFPR, Amount from UserAcc WHERE UserFPR=?");			
			pStm.setBytes(1, UserFPR);
			rs = pStm.executeQuery();
			
			String info = "";

			while (rs.next()) {
				byte[] IssuerFPR=rs.getBytes(1);
				long amount=rs.getLong(2);
				info = info + FPRUtils.ToString(IssuerFPR)+"   "+String.valueOf(amount/ValueMultiplier)+"\n";
			}
			
			return info;
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
		
	}
	
	public long getDraftSN(Connection con) throws SQLException
	{
		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("SELECT COALESCE((MAX(SN)+1),1) FROM Draft");			
			rs = pStm.executeQuery();
			if (rs.next()) return rs.getLong(1); else return 1;
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
		
	}

	public long getOfferSN(Connection con) throws SQLException
	{
		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("SELECT COALESCE((MAX(SN)+1),1) FROM Offer");			
			rs = pStm.executeQuery();
			if (rs.next()) return rs.getLong(1); else return 1;
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
		
	}

	public byte[] findDraftHash(Connection con, byte[] reqHash) throws SQLException
	{
		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("SELECT CertText FROM Draft WHERE ReqHash=?");			
			pStm.setBytes(1, reqHash);
			rs = pStm.executeQuery();
			
			if (rs.next()) return rs.getBytes(1); else return null;
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
		
	}

	public long findOfferHash(Connection con, byte[] reqHash) throws SQLException
	{
		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("SELECT SN FROM Offer WHERE ReqHash=?");			
			pStm.setBytes(1, reqHash);
			rs = pStm.executeQuery();
			
			if (rs.next()) return rs.getLong(1); else return -1;
			
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
		
	}

	public void storeDraft(Connection con, long DSN, byte[] reqHash, byte[] cert) throws SQLException
	{
		PreparedStatement pStm = null;
		ResultSet rs = null;

		try
		{
			pStm = con.prepareStatement("INSERT INTO Draft (SN, ReqHash, CertText) VALUES (?,?,?)");			
			pStm.setLong(1, DSN);
			pStm.setBytes(2, reqHash);
			pStm.setBytes(3, cert);
			pStm.executeUpdate();
		}
		finally
		{
			closeStatementAndResultSet(pStm, rs);
		}
		
	}
	
//---------------------	
	
	private void closeStatementAndResultSet(Statement stm, ResultSet rs) throws SQLException
	{
		if (rs != null)
			rs.close();
		
		if(stm != null)
			stm.close();
	}

	public void beginTransaction(Connection con) throws SQLException
	{
		if (!con.getAutoCommit())
		{
			log.error("Transaction already in progress!");
			throw new IllegalArgumentException("Transaction already in progress!");
		}
		con.setAutoCommit(false);			
	}

	public void lockWrites(Connection con, String table) throws SQLException
	{
// TODO:  lock		
		Statement stm = con.createStatement();
		stm.execute("LOCK TABLE "+table+" IN SHARE ROW EXCLUSIVE MODE");
		stm.close();
	}

	public void setTransactionSerializable(Connection con) throws SQLException
	{
		if (con.getAutoCommit())
			beginTransaction(con);
		
		con.setTransactionIsolation(Connection.TRANSACTION_SERIALIZABLE);
	}
	
	
}
